const char VERSION[] = "1.0 (Shatsky, kolja@sai.msu.ru, 18.01.2001)";
/**
	@file cnt2asc.cpp
	@brief	Reading the MASS binary count-file *.cnt and converting the count 
		data into an ASCII table. A standalone program.
	
	\b Usage
	\verbatim
	./cnt2asc <MASS_count_file_name> [address [Nbasetime]]
	or
	./cnt2asc [-h] to get a short help
	\endverbatim
	
	The program opens the \em MASS_count_file_name and seeks the position \em
	address. By default, the beginning of file is sought. Then the \em
	Nbasetime records (1 by default) are read and printed in standard output as
	an ASCII table with the number of columns equal to the number of channel
	buffers (active MASS channels when writing the count-file) and row number
	equal to the number of counts received in each channel during \em Nbasetime
	base times.

	The \em address is accepted either in decimal or in hexadecimal format
	(preceded with "0x" in latter case). It is normally given in the last field
	of the base-time scintillation index record (format of this address is
	"\@<hex_offset>"). This address is converted into the long integer number
	and used as an \em address for ioc::read() command to read data from the
	count-file.

	It is assumed that the count-file *.cnt is written with the same or
	compatible version of iocount.cpp module as used here. 
	
	The number of count buffers and the length of buffers (checked to be the
	same for all buffers of the same base-time record) are read from the
	descriptor to which the \em addr points. The constraint for several
	base-time records read in one turn (\em Nbasetime>1) is that they all have
	the same buffers' number and the length of any buffer must not exceed the
	one of the first buffer. If either of these conditions are violated, the
	reading is stopped with a message to stderr.

	@return
	On success, the long integer address of the \em next base-time record is 
	returned; on error -1 is returned and the message (normally, from the
	IOCOUNT module) is printed in standard error output. The change of buffer's
	number or length is not accounted as error; the address of this differing
	record is returned.

	Compilation:
	\verbatim
	g++ -Wall -o cnt2asc cnt2asc.cpp nrutil.o iocount.o
	\endverbatim	
	@author	N.Shatsky (SAI,2002)
	@version	1.01
*/

#include <iostream>
#include "iocount.h"
#include "nrutil.h"
using namespace std;

int main(int argc, char * argv[]) 
{


	FILE * f = 0;
	if (argc>=2) f=fopen(argv[1],"rb") ;

	if (!f || (argc<2) || !strcmp(argv[1],"-h")) {
		cout << "MASS count files converter version "<<VERSION<<". Usage:\n";
		cout << "./cnt2asc <MASS_count_file_name> [address [Nbasetime]]\n" ;
		cout << "	or\n";
		cout << "./cnt2asc [-h] to get this help\n";
		exit(0) ;
	}
	const int maxndata = 100000;	// maximal feasible buffer length
	const int maxnbuf = 10 ;	// maximal feasible buffers number
	int ndata = 0 ;
	int nbuf = 1 ;	// to be read from the first buffer descriptor
	count_t ** data = 0 ;
	
	char * endptr ;
	long address = argc<3 ? 0 : strtol(argv[2], &endptr, 0) ;
	int nbasetime = argc<4 ? 1 : strtol(argv[3], &endptr, 0) ;
	if (nbasetime<1) {
		cerr << "Invalid base times number\n";
		exit(-1);
	}
	long next = 0 ;
	for (int b=-1; b<nbasetime; b++) {
		ioc::read(f, (count_t**)data, &ndata, &nbuf, address) ;
		next = ftell(f) ;
		if (int err = nr::erget()) {
			/* first error must be the error of differing buffer parameters */
			if ((ndata>0) && (ndata<=maxndata) && 
				(nbuf>0) && (nbuf<=maxnbuf) && (err==nr::EROFL)) {
				nr::erreset() ;
				if (b==-1) {
					/* Ok: allocate the buffers: */
					data = new (count_t*)[nbuf] ;
					for (int ib=0; ib<nbuf; ib++) 
						data[ib] = new count_t[ndata] ;
				} else {
					/* the buffer parameters are changed for the second time:*/
					address = next ;
					cerr << "The next record is different: Nbuf="<<nbuf
					<<" Ndata=" <<ndata<<'\n';
					break ;	// normal exit				
				}	
			} else {
				cerr << nr::ermessage() << '\n' ;
				exit (-1);
			}
		} else {
			address = next ;
			for (int i=0; i<ndata; i++) {
				for (int b=0 ; b<nbuf; b++) {
					cout << data[b][i] << ' ';
				}
				cout << '\n' ;
			}
		}
	}
	fclose(f) ;

	/* free the buffers: */
	if ((nbuf>0) && (ndata>0)) {
		for (int ib=0; ib<nbuf; ib++) 
			delete[] data ;
		delete[] data;
	}
	if (nr::erget()) {
		cerr << nr::ermessage() ;
	}
	return address ;
}
