#include <pthread.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "exchange.h"
#include "rserrno.h"
#include "names.h"

extern int		errno;

using namespace std;

namespace exch {
	uchar	*msgdata = NULL;
	int		id_mcua = -1;
	FILE*	srv_file = NULL;
	string	_lastError = "";
	int		_lastErrCode = 0;
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
}

int exch::twoUcharToUint(uchar low, uchar high) {
	__u16 i16 = low + (high << 8);
	return int(i16);
}

int exch::twoUcharToSint(uchar low, uchar high) {
	__s16 i16 = low + (high << 8);
	return int(i16);
}

int exch::micro_wait(int mks) {
	static struct timeval moms;
	struct timeval mome;
	int mksd = 0;
	int secd = 0;
	if( mks == 0 ) {
		gettimeofday(&moms, NULL);
		return 0;
	}
	while( mksd < mks ) {
		gettimeofday(&mome, NULL);
		secd = mome.tv_sec - moms.tv_sec;
		mksd = mome.tv_usec - moms.tv_usec;
		if( (secd < 0) || (secd > 1) ) return -1;
		if( secd > 0 ) mksd += 1000*1000;
	}
	return 0;
}

void exch::terminate_program(void) {
//	jdebug("Terminate programm");
	if(id_mcua > 0) close(id_mcua);
	if(msgdata) free(msgdata);
}

void exch::init(bool& withDevice) {
	char vers[13];
//	jdebug("start of initThreads:");

	srv_file = NULL;
	_lastError = "";
	_lastErrCode = 0;

	if(withDevice) {
		micro_wait(0);	// init
		if( (id_mcua=open(devFileName, O_RDWR | O_NONBLOCK)) < 0 ) {
			int errcode = errno;
			if(errcode == ENODEV) fprintf(stdout,"No RS485 driver detected\n");
			else fprintf(stdout,"%s\n",rs_geterror(errcode));
			fprintf(stdout,"Start without device? (Y/N): ");
			char yesno;
			scanf( "%c",&yesno );
			if( (yesno != 'Y') && (yesno != 'y') ) exit(3);
			else {
				withDevice = false;
				return;
			}
		}
		int verscode = ioctl( id_mcua, RS_GET_VERS, vers );
		if( verscode < 0 ) {	// old driver, sorry...
			fprintf(stderr,"Unsupported rs485 driver version...\n");
			exit(3);
		}
		fprintf(stderr,"                   rs485 driver ver.%s", vers);
		bool rtl = RS485_RTL_CODE(verscode);
		if(rtl) fprintf(stderr,"-rtl\n");
		else fprintf(stderr,"\n");
		int size_data_fifo = ioctl( id_mcua, RS_GET_DATAFIFO_SIZE );
		if( size_data_fifo <= CACHE_MAXSIZE ) {
			fprintf(stderr,"DATA_FIFO_SIZE = %d < CACHE_MAXSIZE = %d... It's very strange...\n",
						size_data_fifo, CACHE_MAXSIZE);
			exit(3);
		} else {
			msgdata = (unsigned char *) malloc(CACHE_MAXSIZE*sizeof(uchar));
		}
	}
	atexit( terminate_program );
}

void exch::setSrvFile(FILE* f) {
	srv_file = f;
}

int exch::transaction(const uchar *sendbuf, int nbytes, uchar *recivbuf,
							int extraSignal, bool verifySignal,
							bool saveToLog) {
	pthread_mutex_lock(&mutex);

	uchar tmpbuf[CACHE_MAXSIZE];
	if(nbytes > CACHE_MAXSIZE) return ERSFORMAT;
	memcpy(tmpbuf,sendbuf,nbytes);

	int att = MAX_RETRY_COMMAND;
	int retval;
	while(att--) {
		if( saveToLog && srv_file ) {
			fprintf(srv_file,"->");
			for(int i=0; i<nbytes; i++) fprintf(srv_file," %X",tmpbuf[i]);
		}
		micro_wait(WRITE_WAIT_INTERVAL);	// 500ms
		retval = write(id_mcua,tmpbuf,nbytes);
		micro_wait(0);	// nachalo novogo ozhidaniya.
		if(saveToLog && srv_file) fprintf(srv_file,"   <-");
		if(retval < 0) {
			int errcode = errno;
			if(saveToLog && srv_file) fprintf(srv_file," ERROR: %s\n",
									rs_geterror(errcode));
			ioctl(id_mcua,RS_RESET);
			if(att) continue;
			else {
//				return -1*errcode;	// == errcode  
				retval = -1*errcode;	// == errcode  
				goto RETURN;
			}
		}
		if( retval < CACHE_MAXSIZE ) {
			if(saveToLog && srv_file) {
				for(int i=1; i<retval; i++)
					fprintf(srv_file," %X",tmpbuf[i]);
				fprintf(srv_file,"\n");
			}
			if(recivbuf) memcpy(recivbuf,tmpbuf+1,--retval);
			goto RETURN;
		}
		if(saveToLog && srv_file) fprintf(srv_file," %s\n",rs_getsignal(retval) );
		if(!verifySignal) goto RETURN;;
		if( (retval == SignalACY) || (retval == SignalACW)
				|| (retval == SignalACK) || (retval == extraSignal) ) goto RETURN;
	}
	retval = -1*retval;	// == errcode  
RETURN:
	pthread_mutex_unlock(&mutex);
	return retval;
}

int exch::send(const uchar* buf, int nbytes,
						int signalSuccess1,
						int signalSuccess2,
						int extraSignal,
						bool saveToLog) {
	int retval = transaction(buf, nbytes, NULL, extraSignal,true,saveToLog);
	if( retval < 0 ) return retval;	// exchange error
	if( retval < CACHE_MAXSIZE ) return -1*ERSUNEXPDATA;	//   ... 
	if( (retval != signalSuccess1) && (retval != signalSuccess2)
			&& (retval != extraSignal) ) {
		if(srv_file) fprintf(srv_file,"ERROR: %s\n",
									rs_getsignaltxt(retval));
		return -1*retval;	// == errcode  
	}
	return retval;
}

int exch::sendCommand(uchar module, uchar command,
							int signalSuccess1,
							int signalSuccess2,
							int extraSignal) {
	uchar buf[2];
	buf[0]=module;
	buf[1]=command;
	int retval = send(buf,2,signalSuccess1,signalSuccess2,extraSignal);
	if( retval < 0 ) return retval;
	return SUCCESS;
}

int exch::sendByte(uchar module, uchar command, uchar data) {
	uchar buf[3];
	buf[0]=module;
	buf[1]=command;
	buf[2]=data;
	int retval = send(buf,3);
	if( retval < 0 ) return retval;
	return SUCCESS;
}

int exch::sendWord(uchar module, uchar command, int data) {
//	  ,  data  low  high 
	uchar buf[4];
	buf[0]=module;
	buf[1]=command;
	buf[2]=low(data);
	buf[3]=high(data);

	int retval = send(buf,4);
	if( retval < 0 ) return retval;
	return SUCCESS;
}

int exch::sendData(const uchar* data, int nbytes) {
	int retval = send(data, nbytes);
	if( retval < 0 ) return retval;
	return SUCCESS;
}

int exch::sendReset(uchar module, uchar command) {
#ifdef __COUNTER_BUTTON__
//	 SUCCESS,    DSC,    (<0)
	return sendCommand(module, command, -1, -1, SignalDSC);
#else
//	 SUCCESS,    ACY,    (<0)
	return sendCommand(module, command, SignalACY, -1, -1);
#endif
}

int exch::sendTest(uchar module, uchar command) {
//	 SUCCESS,    ACY ili ACW,    (<0)
	uchar buf[2];
	buf[0]=module;
	buf[1]=command;
	return send(buf, 2, SignalACY, SignalACW, -1, false);	// no save to log-file
}

int exch::askData(uchar module, uchar command, uchar* buf, int nbytes, bool saveToLog) {
//       data
	uchar sendbuf[2];
	uchar recivbuf[CACHE_MAXSIZE];

	if(nbytes > CACHE_MAXSIZE) return ERSFORMAT;
	sendbuf[0]=module;
	sendbuf[1]=command;
	int retval = transaction(sendbuf, 2, recivbuf, -1, true, saveToLog);
	if( retval < 0 ) return retval;
	if( retval > CACHE_MAXSIZE ) return -1*retval; // errcode  
	if(retval != nbytes) return -1*ERSINVALIDATA;
	memcpy(buf,recivbuf,nbytes);
	return SUCCESS;
}

int exch::askByte(uchar module, uchar command, uchar& data, bool saveToLog) {
//       data
	const int nbytes = 1;
	uchar recivbuf[1];
	int retval = askData(module, command, recivbuf, nbytes, saveToLog);	
	if( retval != SUCCESS ) return retval;
	data = recivbuf[0];
	return SUCCESS;
}

int exch::askWord(uchar module, uchar command, int& data, bool saveToLog) {
//  2      data  integer
	const int nbytes = 2;
	uchar recivbuf[2];
	int retval = askData(module,command,recivbuf,nbytes, saveToLog);	
	if( retval != SUCCESS ) return retval;
//	data = int(recivbuf[0] + (recivbuf[1] << 8));
	data = twoUcharToSint(recivbuf[0],recivbuf[1]);
	return SUCCESS;
}

int exch::getLastErrCode(void) {
	return _lastErrCode;
}

const char* exch::getLastError(void) {
	return _lastError.c_str();
}

int exch::isErrorFatal(int errcode) {
	switch(errcode) {
		case ERSNODEVICE	: return 1;
		case ERSNOPORT		: return 1;
		case ERSNODRIVER	: return 1; 
		case ERSNOINTERFACE	: return 1; 
		case ERSSHUTDOWN	: return 1;
		case ERSNOCONVREPLY	: return 1;
		default				: return 0;
	}
}

int exch::isNoReply(int errcode) {
	if(errcode == SignalDSC) return 1;
	else return 0;
}
const char* exch::getErrTxt(int errcode) {
//	errcode > 0 !!!
//	errTxt     (errcode < ERSBASE)
//	   (errcode >= ERSBASE)

	_lastErrCode = errcode;
	const char* errText;
	if( errcode < ERSBASE ) errText = rs_getsignaltxt(errcode);
			//    errno ...
	else errText = rs_geterror(errcode);
	if(srv_file) fprintf(srv_file,"Error: %s\n", errText);
	return errText;
}

void exch::setLastErrorText(const char* errText) {
	_lastError = errText;
}

string exch::talk(const char* text) {
	const int len=6;
	uchar sendbuf[len];
	uint itmp[len];
	uchar recivbuf[CACHE_MAXSIZE];
	int nbytes = sscanf(text,"%X %X %X %X %X %X",
		&itmp[0],&itmp[1],&itmp[2],&itmp[3],&itmp[4],&itmp[5]);
	if(nbytes <= 0) return string("ERROR");
	for(int i=0; i<nbytes; i++)  sendbuf[i] = (uchar)itmp[i];
	int retval = transaction(sendbuf, nbytes, recivbuf,-1,false);
	if(retval < 0) {
		return string(rs_geterror(-1*retval));
	}
	if(retval > CACHE_MAXSIZE) {
		return string(rs_getsignal(retval));
	}
	int lenght = retval*3+1;
	char* response = new char[lenght];	// + '\0'

	for(int i=0; i<retval; i++) {
		sprintf(response+i*3,"%02X ",recivbuf[i]);
	}
	string response_str = string(response);
	delete[] response;
	return response_str;
}

int exch::readPacket(uchar **p_data) {
	int count_byte = read(id_mcua, msgdata, CACHE_MAXSIZE);
	*p_data = msgdata;
	return count_byte;
}

void exch::resetBlockCounter(void) {
	ioctl(id_mcua, RS_BLNUMBER_RESET);
}
