/*--------------------------------------------------------------------------*/
/** @file exchange.h
@brief  Low-level exchange implementation of the program with the device driver.

                  MASS project: TURBINA module file
                       header file for exchange.cpp

Implements the commands sending, data and signals receiving from the device via
the driver inserted into OS kernel.


 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/

#ifndef EXCHANGE_H
#define EXCHANGE_H

#include <stdio.h>
#include <string>
#include "mcua.h"

typedef unsigned char uchar;

#define HEAD_SIZE	5	// size of header of block (bytes)

#define DEVICE_WAIT_INTERVAL		5		// 5 millisec
#define WRITE_WAIT_INTERVAL			500		// 500 usec
#define APERTURE_DELAY				50		// 50 millisec
#define USER_TRACK_STEP				1000	// 1000 millisec
#define MAX_RETRY_COMMAND			3

#define low(i)	uchar( i & 0x00ff)
#define high(i)	uchar( (i & 0xff00) >> 8 )

namespace exch {
	int		twoUcharToUint(uchar low, uchar high);	// to unsigned int
	int		twoUcharToSint(uchar low, uchar high);	// to signed int
	int		micro_wait(int usec);
	void	terminate_program(void);
	void	init(bool& withDevice);
	void	setSrvFile(FILE* f);

	int		transaction(const uchar *sendbuf, int nbytes,
							uchar *recbuf,
							int extraSignal=-1,
							bool verifySignal=true,
							bool saveToLog=true);
	int		send(const uchar* buf, int nbytes,
							int signalSuccess1=SignalACY,
							int signalSuccess2=SignalACW,
							int extraSignal=-1,
							bool saveToLog=true);
	int		sendCommand(uchar module, uchar command,
							int signalSuccess1=SignalACY,
							int signalSuccess2=SignalACW,
							int extraSignal=-1);
	int		sendByte(uchar module, uchar command, uchar data);
	int		sendWord(uchar module, uchar command, int data);
	int		sendData(const uchar* data, int nbytes);
	int		sendReset(uchar module, uchar command);
	int		sendTest(uchar module, uchar command);
	int		askData(uchar module, uchar command, uchar* buf, int nbytes, bool saveToLog=true);
	int		askByte(uchar module, uchar command, uchar& data, bool saveToLog=true);
	int		askWord(uchar module, uchar command, int& data, bool saveToLog=true);
	int		getLastErrCode(void);
	const char* getLastError(void);
	int		isErrorFatal(int errcode);
	int		isNoReply(int errcode);
	const char*	getErrTxt(int errcode);
	void	setLastErrorText(const char* errText);
	std::string	talk(const char* text);
	int		readPacket(uchar **data);
	void	resetBlockCounter(void);
}

#endif
