#include <math.h>
#include <qapplication.h>
#include "names.h"
#include "graph.h"
#include "tools.h"
#include "cfg.h"
#include "messages.h"
#include "summary.h"
#include "nrutil.h"
#include "atmos.hxx"
#include "tree.h"

extern CFGTree* tree;
extern OutControl outcontrol;
int wDevX( double x, Scale* s ) {
	return int( rint( (x-s->xmin)*s->sx) );
}

int wDevY( double y, Scale* s ) {
	return int( rint( (y-s->ymin)*s->sy) );
}

QPoint wDev( double x, double y, Scale* s ) {
	return QPoint( wDevX(x,s),wDevY(y,s) );
}

double devWX( int x, Scale* s ) {
	return( (double)x / s->sx + s->xmin );
}

double devWY( int y, Scale* s ) {
	return( (double)y / s->sy + s->ymin );
}

/**
	@brief	Compute the ticks origin/spacing and axis limits for graphics
	@param	min	On enter: data minimum; on exit: axis origin (\a min - lmargin)
	@param	max	On enter: data maximum; on exit: axis end (\a max + rmargin)
	@param	expand	range expansion factor (relative "margins" size)
	@param	n	desired ("optimal") number of ticks
	@param	tick0	resulting coordinate (numeric value) of a first tick
	@param	dtick	resulting tick spacing
	@return	Number of resulted tick marks (in range ~ \a n/1.5..n*1.5)
	
	The utility acceps the range \a min..max of data for a graphic display
	(either abscissa or ordinate, no matters) and  adjusts the axis limits and
	tick marks origin and spacing to have \e about \a n ticks per axis. The
	ticks label values (i.e. coordinates) are adjusted to be "rounded", i.e. to
	contain the least possible number of significant digits.

	The axis limits are expanded with respect to the data range \a min..max to
	have some "margins" each of the size not less than \a expand*(max-min).
	Also, if only "a little" is needed in expansion to include a new tick mark
	at the exact axis limit, the latter is adjusted correspondingly.
	
	\e Example:
	
	Given the range \a min=22.01, \a max=22.87, \a expand=0.050, the
	utility returns:
	
	\a min = 21.967, \a max = 22.913 (recommended axis limit), \a tick0 = 22, 
	\a dtick = 0.1, returned value (number of ticks) is 10.
	
	This means that the ticks are to be placed at:
	
	22 22.1 22.2 22.3 22.4 22.5 22.6 22.7 22.8 22.9 .
*/
int graphticks(double& min, double& max, double expand, int n,
	double& tick0, double& dtick) {
	const double crit = 0.8;	// maximal distance of a marginal tick from 
						// a range limit in [dtick]
	if (min>=max) return 0;

	// expansion size: 
	double esize = expand * (max-min);

	// null-approximation of dtick:
	double d0 = (max-min+2*esize)/(n-1);
	// ... and its order:
	double order = floor(log10(d0));
	// convert dtick to the range 1..10 and take the nearest integer value:
	double d1 = rint(d0*pow(10.,-order));
	// change to the nearest "standard" ticks distance:
	if (d1>7)
		d1 = 10;
	else
		if (d1<8 && d1>3)
			d1 = 5;
		else
			if (d1==3)
				d1=2.5;

	// the number of resulted ticks:
	int N1 = int((max-min+2*esize)/d1/pow(10.,order)) ;
	// double or half it if out of range:
	const double sqrt2 = sqrt(2.);
	if (N1<n/sqrt2) {
		if (d1==2.5)
			d1 = 1;
		else
			d1 /= 2 ;
	} else 
		if (N1>n*sqrt2) {
			if (d1==2)
				d1 = 5;
			else
				d1 *= 2;
		}

	// convert back to the original scale:		
	dtick = d1 * pow(10.,order);

	// First tick position accounting the range expansion:
	tick0 = int(ceil((min-esize)/dtick))*dtick;
	// expand more than by esize if the first tick is "too far" from min:
	if ((tick0-min)/dtick > crit)
		// expand a little more to include a new tick:
		min = tick0 = tick0-dtick;
	else
		min -= esize;	// expand as ordered by "expand"

	// number of ticks now accounting the range expansion:
	N1 = int((max+esize-tick0)/dtick)+1 ;

	// last tick position accounting the range expansion:
	double tickl = tick0 + (N1-1)*dtick;
	// expand again more than by esize if the last tick is "too far" from max:
	if ((max-tickl)/dtick > crit) {
		// expand a little more to include a new tick:
		max = tickl = tickl+dtick;		
		N1 ++ ;
	} else
		max += esize;	// expand as ordered by "expand"

	return N1 ;
}

GraphWindow::GraphWindow( QWidget* parent,const char* name )
			: QWidget( parent,name ) {
	empty = true;
	_showLut = false;
	_reDrawLut = true;
	_minLutVal = 0;
	_maxLutVal = 0;

	isTimerEventLocked = false;

	setCaption( "Graph" );

	loadConst( true );

	connect( tree, SIGNAL( changed() ),
				this,SLOT( change() ) );
	timer = startTimer( 1000 ); // 1 sec
}

GraphWindow::~GraphWindow( void ) {
	killTimer( timer );
}

void GraphWindow::timerEvent( QTimerEvent *e ) {
	if( isTimerEventLocked ) return;
	isTimerEventLocked = true;

	if( outcontrol.isGraphReady() ) {
		bool reDraw = outcontrol.reDraw();
		outcontrol.resetGraphReady();
		drawAll( reDraw );
	}

	isTimerEventLocked = false;
}

void GraphWindow::loadConst( bool first ) {
	if( isLeafChanged( cfgFile,"Preferences",
			"Graphic window settings",
			"WindowGeometry" ) ) {
		const char* _geometry;
		leafDataToStr(&_geometry, first, cfgFile,
				"Preferences",
				"Graphic window settings",
				"WindowGeometry" );
		if( !strcasecmp( _geometry,"AUTO" ) ) {
			_winX = qApp->desktop()->width()/2;
			_winY = 0;
			_winWidth = qApp->desktop()->width()/2 - 10;
			_winHeight = qApp->desktop()->height() - 50;
		} else {
			int key = sscanf( _geometry,"%d%d%d%d", &_winX, &_winY,
				&_winWidth, &_winHeight );
			if( key != 4 ) {
				QString warn =
					"GraphWindow: Invalid geometry \nWill be modified.";
				showStrongMessage( warn );
				_winX = qApp->desktop()->width()/2;
				_winY = 0;
				_winWidth = qApp->desktop()->width()/2 - 10;
				_winHeight = qApp->desktop()->height() - 50;
			}
		}
		setGeometry( _winX+10, _winY+68, _winWidth, _winHeight );
//	Pochemu +10 i +68 NE ZNAYU !!!!!
	}

	bool redraw = false;

	if( isLeafChanged( cfgFile, "Preferences",
			"Graphic window settings",
			"Background" ) ) {
		const char* backgr;
		leafDataToStr(&backgr, first, cfgFile, "Preferences",
				"Graphic window settings",
				"Background" );
		_background = backgr;
		setBackgroundColor( _background );
		redraw = true;
	}
	if( isLeafChanged( cfgFile, "Preferences",
			"Graphic window settings",
			"AxesColor" ) ) {
		const char* axesCol;
		leafDataToStr(&axesCol, first, cfgFile, "Preferences",
				"Graphic window settings",
				"AxesColor" );
		_axesColor = axesCol;
		redraw = true;
	}
	if( isLeafChanged( cfgFile, "Preferences",
			"Graphic window settings",
			"FontSize" ) ) {
		leafDataToInt(_fontSize, first, cfgFile, "Preferences",
				"Graphic window settings",
				"FontSize" );
		_font = QFont( "Courier",_fontSize );
		redraw = true;
	}
	if( isLeafChanged( cfgFile, "Preferences",
			"Graphic window settings",
			"PointSize" ) ) {
		leafDataToInt(_pointSize, first, cfgFile, "Preferences",
			"Graphic window settings",
			"PointSize" );
		redraw = true;
	}
	if( redraw && !empty ) drawAll( true );
}

void GraphWindow::change( void ) {
	loadConst( false );
}

const double minFform = 0.0001;
const double maxFform = 10000;
const int G_FORM = -1;

int getPrec( double val ) {

	double val_abs = fabs( val );
	if( (val_abs<minFform)
		|| (val_abs>=maxFform) ) return G_FORM; // g-format
	int prec = 0;
	if( val_abs < 1.0 ) prec = 3;
	else if( val_abs < 10.0 ) prec = 2;
	else if( val_abs < 100.0 ) prec = 1;
	else prec = 0;

	return prec;
}

void toStr( QString& txt, double val, int prec ) {

	char form = 'f';
	if( prec == G_FORM ) form = 'g';
	txt.setNum( val,form,prec );
}

void GraphWindow::drawX( QPainter *p,
			Scale *s,
			int optimNTicks,
			bool top,
			const char* name ) {

	QPen oldpen = p->pen();
	p->setPen( _axesColor );

	int wl = p->window().left();
	int wr = p->window().right();
	int wt = p->window().top();
	int wb = p->window().bottom();

	int tw = p->fontMetrics().width( "wwwwwww" )+10; 
	int th = p->fontMetrics().height();
	QPoint p1 = p->xFormDev( QPoint(0,0) ); 
	QPoint p2 = p->xFormDev( QPoint(0,
//						int(rint(0+0.7*(double)th)) ); 
						int(rint(0+0.7*th))) ); 
	int swy = abs( p1.y() - p2.y() );

	if( top ) p->drawLine( wl,wt,wr,wt );
	else p->drawLine( wl,wb,wr,wb );

	double tick0=0, dtick=0;
	double min = s->xmin;
	double max = s->xmax;

	int nTicks = graphticks( min,max,0,
				optimNTicks,tick0,dtick );

	int prec = getPrec( tick0 + dtick*(nTicks-1) );
	double tick = tick0;
	for( int i=0; i<nTicks; i++ ) {
		int ix = wDevX( tick,s );
		if( (ix < wl) || (ix > wr) ) {
			tick += dtick;
			continue;
		}
		p->setViewXForm( true );
		QPoint b;
		if( top ) {
			p->drawLine( ix,wt,ix,wt-swy );
			b = p->xForm( QPoint(ix,wt) );
		}else {
			p->drawLine( ix,wb,ix,wb+swy );
			b = p->xForm( QPoint(ix,wb) );
		}
		p->setViewXForm( false );
		QString txt; 
		toStr( txt, tick, prec );
		tw = p->fontMetrics().width( txt );
		if( top ) p->drawText( b.x()-tw/2,b.y()-th/2,txt );
		else p->drawText( b.x()-tw/2,b.y()+th,txt );
		tick += dtick;
	}

// Draw Name:
	p->setViewXForm( true );
	QPoint b;
	if( top ) b = p->xForm( QPoint(wr,wt) );
	else b = p->xForm( QPoint(wr,wb) );
	p->setViewXForm( false );
	tw = 3 * p->fontMetrics().width( name );
	if( top ) p->drawText( b.x()-tw/2,
					int(rint(b.y()-1.5*th)), name );
	else p->drawText( b.x()-tw/2,b.y()+2*th,name );

	p->setViewXForm( true );
	p->setPen( oldpen );
}

void GraphWindow::drawX( QPainter *p ) {

	QPen oldpen = p->pen();
	p->setPen( _axesColor );

	int wl = p->window().left();
	int wr = p->window().right();
	int wb = p->window().bottom();

	p->drawLine( wl,wb,wr,wb );
	p->setPen( oldpen );
}

void GraphWindow::drawY( QPainter *p, Scale *s,
			int optimNTicks,
			bool left,
			bool drawNull,
			const char* name,
			bool logY ) {
	QPen oldpen = p->pen();
	p->setPen( _axesColor );

	int wl = p->window().left();
	int wr = p->window().right();
	int wt = p->window().top();
	int wb = p->window().bottom();

	int tw = p->fontMetrics().width( "wwwwwww" )+10; 
	int th = p->fontMetrics().height();
	int th2 = th/3;

	QPoint p1 = p->xFormDev( QPoint(0,0) ); 
	QPoint p2 = p->xFormDev( QPoint(0+int(rint(0.7*th)),0) ); 
	int swx = abs( p1.x() - p2.x() );

	p->setViewXForm( true );
	if( left ) p->drawLine( wl,wt,wl,wb );
	else p->drawLine( wr,wt,wr,wb );

	if( drawNull ) {
		int x0 = wDevX( 0,s );
		p->drawLine( x0, 0, x0, wt );
	}

	double tick0=0, dtick=0;
	double min = s->ymin;
	double max = s->ymax;

	int nTicks = graphticks( min,max,0,
			optimNTicks,tick0,dtick );

	int prec = getPrec( tick0 + dtick*(nTicks-1) );
	double tick = tick0;
	for( int i=0; i<nTicks; i++ ) {
		int iy = wDevY( tick,s );
		if( (iy < wb) || (iy > wt) ) {
			tick += dtick;
			continue;
		}
		p->setViewXForm( true );
		if( left ) p->drawLine( wl,iy,wl+swx,iy );
		else p->drawLine( wr,iy,wr-swx,iy );
		QPoint b = p->xForm( QPoint(wl,iy) );
		p->setViewXForm( false );
		QString txt;
		if( left ) {
			if( logY )
				toStr( txt, fromLogScale(tick), prec );
//				toStr( txt, pow(2,tick), prec );
			else toStr( txt, tick, prec );
			tw = p->fontMetrics().width( txt );
			p->drawText( b.x()-tw,b.y()+th2,txt );
		}
		tick += dtick;
	}
	if( left ) {
		p->setViewXForm( true );
		QPoint b = p->xForm( QPoint(wl,wt) );
		p->setViewXForm( false );
		tw = p->fontMetrics().width( "  " );
		p->drawText( b.x()+tw,b.y()+int(rint(1.8*th)),name );
	}
	p->setViewXForm( true );
	p->setPen( oldpen );
}

void GraphWindow::setLutColor( QPainter* p, double val, double maxVal ) {
	double lutVal = 255;
	if( maxVal != 0 ) lutVal = val / maxVal * 255;
	QColor lutColor = outcontrol.getLutColor( lutVal );
	p->setPen( lutColor );
	p->setBrush( lutColor );
}

void GraphWindow::drawCurve( GraphSet* gs,
				QPainter* bp,
				Scale* s,
				int wt,
				int wr,
				bool reDraw ) {

	bp->setWindow( 0,wt,wr+1,-1-wt );
	if( !gs ) return;
	if( !gs->nPoint() ) return;


	QColor color( gs->color() );
	bp->setBrush( color );
	bp->setPen( color );


	QPoint es1 = bp->xFormDev( QPoint(10,10) ); 
	QPoint es2 = bp->xFormDev( QPoint(10+_pointSize,
					10+_pointSize) ); 
	int ax = abs( es1.x() - es2.x() );
	int ay = abs( es1.y() - es2.y() );

	if( ay == 0 ) ax = 1;

	if( gs->isScan() ) {
		int x0 = wDevX( 0,s );
		bp->drawLine( x0, 0, x0, wt );
	} else if( reDraw && !gs->isProfile() 
			&& (gs->refer() != sum::NODATA) ) {
		int yRef = wDevY( gs->refer(),s );
		bp->drawLine( 0, yRef, wr, yRef );
	}

	int startPointNumber = gs->nPointGraph();
	if( reDraw ) startPointNumber = 0;
	if( gs->isProfile() ) {
		double value = gs->refer();
//		printf( "graph: profile gs->refer() = %g\n",value );
		if( value != _maxLutVal ) {
			_maxLutVal = value;
			_reDrawLut = true;
		}
		_minLutVal = 0;
// Dlia profile perestarivaem posledniuyu i predposledniuyu tochki,
// t.k. posledniaya risuyetsia do UT(current), a nado
// do UT(current) + ( UT(current+1)-UT(current) )/2
		int startProfile = startPointNumber > 0 ?
					startPointNumber-1 : 0;
		for( int i=startProfile; i<gs->nPoint(); i++ ) {
			int nCN2 = (int)sum::getdata( "NCN2" )[i];
			if( nCN2 == sum::NODATA ) continue;

			double* zCN2 = (double*)sum::getdata( "ZCN2",i );
			double* cn2 = (double*)sum::getdata( "CN2",i );

			if( !outcontrol.isPercent() ) {
				atm::interpolate( outcontrol.zSmooth(),
					outcontrol.cn2Smooth(),
					outcontrol.nZsmooth(),
					zCN2, cn2, nCN2 );
				cn2 = outcontrol.cn2Smooth();
				zCN2 = outcontrol.zSmooth();
				nCN2 = outcontrol.nZsmooth();
			}

			if( nr::erget() == nr::ERNOD ) {
				QString mess = nr::ermessage().c_str();
				debug( "!!!!!!!!!!!!!!!!!!! %s",mess.data() );
				nr::erreset();
			}

			int ixl = i, ixr = i;
			if( i > 0 ) ixl = i-1;
			if( i < gs->nPoint()-1 ) ixr = i+1;

			const double maxTimeInterval = 1.0; // 1 hour
			double xDataL = gs->xData()[i];
			if( (sum::getdata("NCN2")[ixl] != sum::NODATA)
				&& (fabs(gs->xData()[ixl]-gs->xData()[i]) < maxTimeInterval) )
				xDataL = gs->xData()[ixl];

			double xDataR = gs->xData()[i];
			if( (sum::getdata("NCN2")[ixr] != sum::NODATA)
				&& (fabs(gs->xData()[ixr]-gs->xData()[i]) < maxTimeInterval) )
				xDataR = gs->xData()[ixr];

			int axL = abs( wDevX( gs->xData()[i],s )
				 - wDevX( xDataL,s ) ) / 2;
			int axR = abs( wDevX( gs->xData()[i],s )
				 - wDevX( xDataR,s ) ) / 2;

			if( !outcontrol.isPercent() ) {
				ay = int( wt / nCN2 / 2 );
			}
			for( int j=0; j<nCN2; j++ ) {
				if( outcontrol.isPercent() ) {
					double maxVal = gs->refer();
					if( !maxVal ) maxVal = 1;
					ay = abs( int( wt / nCN2 * cn2[j] / maxVal ) );
				} else {
					setLutColor( bp, cn2[j], gs->refer() );
				}
				int x = wDevX( gs->xData()[i],s ) - axL;
				int y = wDevY( toLogScale(zCN2[j]),s ) - ay;
				int w = axL+axR+2;
				int h = 2*ay;
				bp->drawRect( x,y,w,h );
			}
		}
	} else {
		for( int i=startPointNumber; i<gs->nPoint(); i++ ) {
			if( gs->yData()[i] != sum::NODATA ) {
				double y = gs->yData()[i];
				if( y < gs->yMin() ) y = gs->yMin();
				else if( y > gs->yMax() ) y = gs->yMax();
//				printf( "draw curve: yMin=%f yMax=%f\n",
//					gs->yMin(), gs->yMax() );
				bp->drawEllipse( wDevX( gs->xData()[i],s ) - ax,
					wDevY( y,s ) - ay,
					2*ax, 2*ay );
			}
		}
	}
	gs->setNPointGraph( gs->nPoint() );	
}

void GraphWindow::drawViewPort(	GraphSet* gs,
			QPainter *bp,
			int left,
			int top,
			int w, // width
			int h, // height
			bool reDraw,
			bool drawTop,
			bool drawBottom ) {

	if( !gs ) return;
	h = (int)( (double)h * gs->quantum() );
	bp->setViewport( left, top, w, h );

	int optimNTicksX = w / ( 2*bp->fontMetrics().width("ww.w") ); 
	int optimNTicksY = h / ( 2*bp->fontMetrics().height() );
	if( optimNTicksY < 2 ) optimNTicksY = 2;

	int wr = 1000;
	int wt = 1000;

	bool changed = gs->searchMinMax();

	reDraw = reDraw || changed; // if changed => reDraw
	if( reDraw ) {
		int port_top = top-_border_y/2;
		int port_height = h+_border_y;//+_border_y/2;
		if( drawTop ) {
			port_top -= _view_marg;
			port_height += _view_marg;
		}
		if( drawBottom ) {
			port_height += _view_marg+_border_y;
		}
		bp->setViewXForm( false );
		bp->fillRect(0,
			port_top,
			width(),
			port_height,
			_background);
		bp->setViewXForm( true );
	}
	Scale s( gs->xMin(), gs->xMax(), gs->yMin(), gs->yMax() );
	if( gs->isProfile() ) {
		s.ymin = toLogScale(s.ymin) - 0.5;
		s.ymax = toLogScale(s.ymax) + 0.5;
	}

	if( s.xmax == s.xmin ) {
		s.xmax++;
		s.xmin--;
	}
	if( s.ymax == s.ymin ) {
		s.ymax++;
		s.ymin--;
	}
	double tick0=0, dtick=0;
	const double expand = 0.05;
	graphticks( s.xmin,s.xmax,expand,optimNTicksX,tick0,dtick );
	graphticks( s.ymin,s.ymax,0,optimNTicksY,tick0,dtick );

	s.sx = wr / ( s.xmax - s.xmin ); 
	s.sy = wt / ( s.ymax - s.ymin );

	int dh = bp->fontMetrics().height();
	bp->setClipRect( left,top-dh/2,w,h+dh );

	bp->setClipping( true );
	drawCurve( gs, bp, &s, wt, wr, reDraw );
	bp->setClipping( false );

	if( reDraw ) {
		const char* nameX;
		if( gs->isScan() ) nameX = "px";
		else nameX = "UT,h";
		if( drawTop ) drawX( bp,&s,optimNTicksX,true,nameX );
		if( drawBottom ) drawX( bp,&s,optimNTicksX,false,nameX );
		else drawX( bp );
		drawY( bp,&s,optimNTicksY,true,gs->isScan(),
				gs->pname(),gs->isProfile() );
		drawY( bp,&s,optimNTicksY,false,false,
				"",gs->isProfile() );
	}
}

void GraphWindow::drawLut( QPainter *bp,
			int left,
			int top,
			int w, //width
			int h, //height
			double minVal,
			double maxVal ) {


	bp->setViewport( left, top, w, h );
	int wt = 1000;
	int wb = 0;
	int wr = 1000;
	bp->setWindow( 0,wt,wr+1,wb-1-wt );

	int port_top = top;
	int port_height = h+_border_y;
	bp->setViewXForm( false );
	bp->fillRect(0,
		port_top,
		width(),
		port_height,
		_background);
	bp->setViewXForm( true );

	const int nStep = 250;
	const int step_width = 4;
	int step_left = 0;
	int step_height = wt / 2;


	for( int i=0; i<nStep; i++ ) {
		setLutColor( bp, double(i), double(nStep-1) );
		bp->drawRect( step_left, step_height,
					step_width+1, step_height );
		step_left += step_width;
	}

	bp->setPen( _axesColor );
	int th = bp->fontMetrics().height();
	QString txt;

	bp->setViewXForm( true );
	QPoint b = bp->xForm( QPoint(wr,step_height) );
	bp->setViewXForm( false );
	txt.setNum( maxVal );
	int tw = bp->fontMetrics().width( txt );
	bp->drawText( b.x()-tw,b.y()+th,txt );

	bp->setViewXForm( true );
	b = bp->xForm( QPoint(0,step_height) );
	bp->setViewXForm( false );
	txt.setNum( minVal );
	tw = 3 * bp->fontMetrics().width( txt );
	bp->drawText( b.x(),b.y()+th,txt );
}

void GraphWindow::drawAll( bool reDraw ) {
	if( outcontrol.nGraph() == 0 ) {
		empty = true;
		return;
	}
//	reDraw = true; // T.k. vsio ravno repaint delat'...
	empty = false;
	_showLut = outcontrol.drawProfile() && !outcontrol.isPercent();
	QPainter bufferpainter;
	bufferpainter.begin( &_buffer );
	if( reDraw ) _buffer.fill( _background );
	bufferpainter.setFont( _font );

	const int border_x = 20;

	_border_y = bufferpainter.fontMetrics().height();
	int view_left = bufferpainter.fontMetrics().width( "wwwwwww" )+10;
	_view_marg = int( (double)bufferpainter.fontMetrics()
			.height()*2.5 );
	int view_top = _view_marg;
	int lut_height = 0;
	if( _showLut ) lut_height = _border_y;
	int view_height = height() - 2*_view_marg - lut_height -
			_border_y*(outcontrol.nGraph()-1) -
			_border_y*(int)_showLut;
	int view_width = width()-view_left-border_x;

	for( int i=0; i<outcontrol.nGraph(); i++ ) {
		bool x_top = (i == 0); // x-ax.
		bool x_bottom = (i == outcontrol.nGraph()-1 ); // x-axis

		drawViewPort( outcontrol.graphSet(i),
			&bufferpainter,
			view_left,
			view_top,
			view_width,
			view_height,
			reDraw,
			x_top,
			x_bottom );

		view_top += int( (double)view_height*outcontrol.graphSet(i)->quantum()
			 + _border_y );
	}

//	if( _showLut ) {
	if( _showLut && _reDrawLut ) {
		view_top += _view_marg - _border_y;
		drawLut( &bufferpainter,
			view_left,
			view_top,
			view_width,
			lut_height,
			_minLutVal,
			_maxLutVal );
	}
	bufferpainter.end();
 	repaint();
}

void GraphWindow::paintEvent( QPaintEvent* e ) {

	bitBlt( this, 0, 0, &_buffer );
}

void GraphWindow::resizeEvent( QResizeEvent* e ) {
	_buffer.resize( e->size() );
	_buffer.fill( _background );
	if( !empty ) drawAll( true );
}
