/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            atmos.cpp                                     */
/*                    Description is given in atmos.hxx                     */
/* Editor tabular size: 4                                                   */
/* Version 2.0	New decomposition, independent executable option (not just  */
/* a test)                                                                  */
/****************************************************************************/
 
/* Insert the macro ATMOSTEST in compilation for production of independent
executable (e.g. for tests) or uncomment below: */
//#define ATMOSTEST
#include <math.h>
#include "atmos.hxx"
#include "scind.hxx"
#include "weif.hxx"
#include "nr.h"
#include <string>
using namespace std;

#ifdef ATMOSTEST
#include "readmass.h"
#include <map>
#ifdef GCC33
#include <sstream>
#else
#include <strstream>
#endif
#include <iostream>

/*---------------------------- TEST PART ---------------------------------*/

/**
	@brief	Convert the string (e.g. of indices in mass-file) into double array
	@param	arr	Place to put the converted numbers
	@param	n	number of converted numbers; if >0 on enter, then specifies the
		maximal quantity of numbers to convert (arr capacity for example)
	@param	string with numbers to convert
*/
void str2da(double * arr, int * n, const char * str) {
	char * endptr = 0;
	const char * nptr = str;
	int i = 0;	// counter of converted numbers
	double d = 0;
	while (1) {
		d = strtod(nptr,&endptr) ;
		if (endptr>nptr && (*n)>0 && i<*n) {
			arr[i++] = d;	// successful conversion
			nptr = endptr;
		} else
			break;
	}
	*n = i;
}

/**
	@brief	Write the relative deviation of observed scintillation indices from
		calcn2() model ones: O-C=(Sm-So)/So
	@param	f	opened file stream pointer
	@param	scmod	model indices array
	@param	scob	observed non-shifted indices
	@param	scob_s	observed shifted indices. \b Null for Normal mode!
	@param	nsc0	number of indices in scobs[] and scobss[]; scmod[] is twice as long
	@param	prefix	What to put as beginning of line
	
	Function writes the deviations like (model-observ)/model for normal and
	differential scintillation indices. All deviations are written as one line. For 4
	apertures, 10 deviations are
	The difference is
	- no errors are specified
	- for generalized mode, both normal and shifted indices are given in one row
	
	The \a prefix is written as leading string; indices are written in
	sc::IDXFMT format with sc::IDXDLM delimiter.
*/
void writedev(FILE * f, const double * scmod, const double * scob, 
	const double * scob_s, int nsc0, const char * prefix) {
	fprintf(f, "%s", prefix);

	const double undef = -9.99;	// written if observed indices are non-positive
	/* Write the model indices: */
	for (int mode=0; mode<=(scob_s!=0); mode++) {
		const double * obs = mode ? scob_s : scob ;
		for (int i=0;i<nsc0;i++) 
			if (obs[i]>0)
				fprintf(f,sc::ERRFMT,sc::IDXDLM,(obs[i] - scmod[i])/obs[i]);
			else
				fprintf(f,sc::ERRFMT,sc::IDXDLM, undef);	
	}
	fprintf(f, "\n");
}
	
int main( int ac, char **av ) {

	if (ac<4) {
		cout << "ATMOS: atmospheric turbulence and parameters calculator\n"
		"Usage: \n./atmos <weight-dir> <mass-file> <zshift1> <zshift2> <secz> [<decomp.file>]\n"
		"  where mass-file may be also '-' for standard input.\n"
		"z* parameters are starting defaults and updated from mass-file data,\n"
		"weight-dir can be a name of a single weight file if no O-lines in mass-file.\n";
		exit (-1);
	}
	const char * wdir = av[1];
	string wfile = av[1];
	const char * dfile = (ac>5) ? av[6] : 0;
	char * endptr;
	double zshift1 ;	// zshifts of normal and 1st half of generalized modes
	zshift1 = strtod(av[3],&endptr);
	double zshift2 = strtod(av[4],&endptr);	// zshift of second half of gen. mode.
	double secz   = strtod(av[5],&endptr);

	const double ZMAX = 25; // km, we want it like this
	const double Z0 = 0;	// trivial


	// number of MASS channels:
	const int NCHAN = 4;
	// number of indices:
	const int NIDX = sc::napcomb(NCHAN)+NCHAN;	// SI,DSI,DESI
	double idx[NIDX*2], e2idx[NIDX*2] ; 	// normal and shifted indices
	// maximal expected quantity of numbers in a mass-file line:
	const int NNUMB = 2*NIDX; // indices and errors
	double da[NNUMB];
	double longitude = -4.72/*hours*/, latitude=-30.15;	// default CTIO coordinates

	FILE * fout = stdout;
	
	FILE * fin = strcmp(av[2],"-") ? fopen(av[2],"r") : stdin ;
	if (!fin) {
		cerr << "Problems with file "<<av[2]<<". Exit\n";
		exit(-1);
	}
	bool isgen = false;
	std::map <string,double> aperdata;
	std::map <string,string> withaper;
	char mode = '?';
	bool headout = false;	// flag of headers written (#R-header must know naper)
	while (!feof(fin)) {
		// read indices:
		rdm::readline(fin);
		string ut = rdm::getut();
		switch (char t = rdm::gettype()) {
			case 'P':
				// get the main parameters: site coords, optics parameters
				if (rdm::getname()=="GENERAL\\SITE\\LONGITUDE") {
					int h,m,s;
					sscanf(rdm::getvalue().c_str(),"%d %d %d",&h,&m,&s);
					longitude = abs(h) + m/60.0 + s/3600.0;
					if (rdm::getvalue().c_str()[0]=='-')
						longitude *= (-1);
//					cout << "LONG="<<longitude<<'\n';
				}
				if (rdm::getname()=="GENERAL\\SITE\\LATITUDE") {
					int d,m,s;
					sscanf(rdm::getvalue().c_str(),"%d %d %d",&d,&m,&s);
					latitude = abs(d) + m/60.0 + s/3600.0;
					if (rdm::getvalue().c_str()[0]=='-')
						latitude *= (-1);
//					cout << "LAT="<<latitude<<'\n';
				}
				if (strstr(rdm::getname().c_str(),"DEVICES\\APERTURES\\STOP")) {
					string val = rdm::getvalue();
#ifdef GCC33
					istringstream ist(val.c_str()) ; // string input
#else
					istrstream ist(val.c_str()) ; // string input
#endif
					string aper;
					double foclen, corr;
                   	ist >> aper >> foclen >> corr;
					aperdata[aper] = foclen;
				}
				if (rdm::getname()=="OPERATIONS\\NORMAL MODE\\WITHAPERTURE") {
					withaper["N"] = rdm::getvalue();
//					cout << "N->"<<withaper["N"]<<'\n';
				}
				if (rdm::getname()=="OPERATIONS\\GENERALIZED\\WITHAPERTURE") {
					withaper["G1"] = rdm::getvalue();
//					cout << "G1->"<<withaper["G1"]<<'\n';
				}
				if (rdm::getname()=="OPERATIONS\\GENERALIZED\\WITHSECOND") {
					withaper["G2"] = rdm::getvalue();
//					cout << "G2->"<<withaper["G2"]<<'\n';
				}
				if (rdm::getname()=="OPTICS\\FEEDING OPTICS\\FOCALLENGTH") {
					aperdata["FEED"] = strtod(rdm::getvalue().c_str(),0);
//					cout << "Feed="<<aperdata["FEED"]<<'\n';
				}
				break;
			case 'O':
				{
					string val = rdm::getvalue();
#ifdef GCC33
					istringstream ist(val.c_str()) ; // string input
#else
					istrstream ist(val.c_str()) ; // string input
#endif
					string name;
					ist >> name;
					if (name.length()<4)
						// another word is also star name (which we don't need):
						ist >> name;
					int crd;
					const int ncrd = 3+3;	// number of coordinate numbers (hmsdms)
					for (int i=0; i<ncrd; i++) ist>>crd;
					// ok, we're by the spectrum:
					ist >> name;
					// convert the SED name into lower-case and produce the weight
					// file name:
					for (int i=0; i<(int)name.length(); i++) name[i] = tolower(name[i]);
					const char WFEXT[] = ".wf";	// weight file extension
					if (wdir[strlen(wdir)-1]=='/') 
						wfile = wdir + name + WFEXT;
					else
						wfile = string(wdir) + '/' + name + WFEXT;
					if (ist) ist >> secz ;	// last parameter should be zen.dist
					// a bug: catch former zdist instead of secz:
					if (secz>2)
						secz = 1./cos(secz*M_PI/180.) ;  
//					cout << "WEIGHT="<< wfile<<" secz="<<secz<<'\n';
				}
				
				break;
			case 'M':
				mode = rdm::getname()[0];
				if (mode=='N' || mode=='G') {
					// check if there is secz info:
					if (rdm::getvalue().length()) {
						char * endptr;
						double d = strtod(rdm::getvalue().c_str(),&endptr);
						if (d>=0 && endptr>rdm::getvalue().c_str())
							secz = d;
//						cout << " secz="<<secz<<'\n';
					}
					isgen = (mode=='G');
					// refresh zshifts:
//					cout <<rdm::getvalue()[0]<<": ";
				
					double focfeed = aperdata["FEED"];
					double foclens = aperdata[withaper["G2"]] ;
					// here we assume that for positive focal lengths, the negative
					// shift is returned (version of weif at least 1.6 !)
					zshift2 = foclens!=0 ? wf::getzshift(focfeed,foclens) : 0;
				
					foclens = isgen ? aperdata[withaper["G1"]] :  aperdata[withaper["N"]];
					zshift1 = foclens!=0 ? wf::getzshift(focfeed,foclens) : 0;
//					cout << "Z1=" << zshift1 << " Z2=" << zshift2 << '\n';
					atm::update(wfile.c_str(),dfile,zshift1,zshift2,Z0,ZMAX);
				
					if (nr::erget()) {
						fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
						return nr::erget() ;
					}
					if (!headout) {
						/* write headings: */
						atm::write(fout,"#A UT     ", atm::ATMPAR, atm::HEAD) ;
						atm::write(fout,"#T UT     ", atm::CN2PROF, atm::HEAD) ;
						atm::write(fout,"#R UT     ", atm::O_CSCIND, atm::HEAD) ;
						headout = true;
					}
				} else 
					mode = '?';
				break;
			case 'i': case 'j':
				if (mode!='?') {
					// convert the record into indices:
					int n = NIDX;
					/* We put the normal indices (SI,DSI,DESI) in first NIDX
					elements, shifted - after first NIDX skipped */
					str2da(idx+(t=='j')*NIDX,&n,rdm::getvalue().c_str());
					if (t=='j')
						atm::calcint(idx, &(idx[NIDX])) ;
					else
						if (!isgen) 
							atm::calcint(idx, 0) ;
						// otherwise - wait for generalized indices
				}
			break;
			case 'I': case 'J':
//			case 0:
			if (mode!='?') {
				int n = NIDX*2;
				/* We put the normal indices (SI,DSI,DESI) in first NIDX
				elements, shifted - after first NIDX skipped */
				str2da(da,&n,rdm::getvalue().c_str());
				// sort out indices:
				for (int i=0; i<NIDX; i++) {
					idx[i+(t=='J')*NIDX] = da[i*2];
					double eidx = da[i*2]*da[i*2+1] ;
					e2idx[i+(t=='J')*NIDX] = eidx * eidx;
				}

				if (!isgen || t=='J') {
					// DESI in A, maybe shifted
					double desi = idx[NIDX*(1+(t=='J'))-NCHAN+1];
					// its absolute error:
					double edesi = sqrt(e2idx[NIDX*(1+(t=='J'))-NCHAN+1]);
					// convert into relative error:
					if (desi>0) edesi/=desi; else edesi=0;
					// average integrals and produce atm.parameters:
					atm::avgint(desi, edesi, secz) ;

					if (nr::erget()) {
						// most possible - about integral <0. Skip this results and go
						fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
						nr::erreset();
					} else
						// write parameters on stdout:
						atm::write(fout,(string("A ")+ut).c_str(), atm::ATMPAR) ;

					// make profile restoration with any method:
					double seecn2 = 0;
					const char meths[3] = "LX";
					for (int m = 0; m<(int)strlen(meths); m++) {
						if (isgen)
							seecn2 =atm::calcn2(idx, &(idx[NIDX]),
								e2idx,&(e2idx[NIDX]),secz, meths[m]) ;
						else
							seecn2 =atm::calcn2(idx, 0, e2idx, 0, secz, meths[m]) ;
						if (nr::erget()) {
							// not convergence most probably...
							fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
							nr::erreset();
						} else {
							atm::write(fout,(string("T ")+ut).c_str(), atm::CN2PROF) ;
							atm::write(fout,(string("R ")+ut).c_str(), atm::O_CSCIND);
						}
					}
				}// otherwise - wait for generalized indices
			}
			break;
			default:;
		}
	}

	/* One check per all calls: */
	if (nr::erget()) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
		return nr::erget() ;
	}
	atm::done();
	
	fclose(fout) ;	
	fclose(fin) ;
	return 0 ;
}
#endif

/** private part of a namespace at */

namespace atm {

/*------------------------ Global parameters -----------------------------*/

/** Bad result placeholder and value */
double bad = 0 ;

/** Cn^2 profile restoration method:  */
char cn2method = '?' ;

/** Dimension of (z)cn2 vectors */
const int MAXNLAYER = (NFIXED>NFLOAT) ? NFIXED+1 : NFLOAT+1 ;

/** Actual length of the computed Cn^2-vector from calcn2() */
int ncn2 = 0 ;

/** Cn^2-vector computed by Cn2 restoration calcn2() and Chi2 as last elem.*/
double cn2[MAXNLAYER] ;

/** Seeing computed by restored Cn2() profile. Whole atm. - for Generalized
mode restoration; free atmosphere - for Normal mode. */
double cn2seeing = 0;
double getcn2see() { return cn2seeing;}

/** Altitudes of Cn^2 layers from calcn2() */
double zcn2[MAXNLAYER] ;

/** Basic set of weight functions: */
wf_t atwf ;

/** Combined matrix of weights for Cn2 restoration: used in calcn2() and in
meritfunc() */
double ** wcomb = 0;

/** Shifts by altitude [km] by which normal (first) and generalized (both) 
modes operate (normally non-positive) */
double zshift[NMODE] = {-99,-99} ;

/** Two Matrices of coefficients for the moments of Cn2  */
double** coef[2] = {NULL, NULL} ;	/**< [0]: for non-shifted measurements, [1] 
for generalized mode measurements */

/** Number of photometric channels */
int nchan = 0 ;

/** Number of weights respective to nchan */
int nw = 0 ;

/* Combined average indices and their errors for restoration of
Cn2-profile */
double * sicomb = 0, * e2sicomb = 0 ;

/* Combined (observed-model)/observed indices produced by calcn2() */
double * o_csi = 0;

/** Storage of accumulated vectors of instantaneous turbulence moments */
double ** cn2int ;

/** Physical length of cn2int storage (allocated space) */
int cn2intlen = 0;

/** Current length of cn2int storage (measurement number) */
int curmeas = 0;

/** User's altitude grid for Cn2 moments decomposition. ZMAX serves also as
an upper limit for Cn2 restoration method. Filled in update(). */
double z0, zmax, dz, dzmin ;

/* Vectors of average integrals and their errors */
double  aint[NPOWER] ;
double e2int[NPOWER] ;

/* Vector of atmospheric parameters and their errors made from aint,e2int */
double atmpar[LAST] ;
double eatmpar[LAST] ;

/* Currently loaded weight file name: to check whether it's changed since last call */
std::string curwffile = "" ;

}

/*============================== Functions ===============================*/

/*------------------------------ getval ----------------------------*/

double atm::getval(ordint what) {
	switch (what) {
		case FSEE:	return (atmpar[FSEE]>0 && atmpar[FSEE]<MAXANGLE) ? atmpar[FSEE] : bad;
		case WSEE:	return (atmpar[WSEE]>0 && atmpar[WSEE]<MAXANGLE) ? atmpar[WSEE] : bad;
		case FM0:	return (aint[IFSE]>MINM0 && aint[IFSE]<MAXM0) ?	aint[IFSE]:bad;
		case WM0:	return (aint[IWSE]>MINM0 && aint[IWSE]<MAXM0) ?	aint[IWSE]:bad;
		case FHEFF:	return (atmpar[FHEFF]>MINHEFF && atmpar[FHEFF]<MAXHEFF) ? atmpar[FHEFF] : bad;
		case WHEFF:	return (atmpar[WHEFF]>MINHEFF && atmpar[WHEFF]<MAXHEFF) ? atmpar[WHEFF] : bad;
		case ISOPL:	return (atmpar[ISOPL]>0 && atmpar[ISOPL]<MAXANGLE) ? atmpar[ISOPL] : bad;
		case M2:	return (aint[IISK]>MINM2 && aint[IISK]<MAXM2) ?	aint[IISK]:bad;
		case TC:	return (atmpar[TC]>0 && atmpar[TC]<MAXTC) ? atmpar[TC] : bad;
		default:
			nr::nrerror("ATMOS: getval(): wrong what-selection",nr::ERPAR);
			return bad ;
	}
}

/*------------------------------ geterr ----------------------------*/

double atm::geterr(ordint what) {
	double v, e ;
	switch (what) {
		case FSEE:	v = atmpar[FSEE]; e = eatmpar[FSEE]; break ;
		case WSEE:	v = atmpar[WSEE]; e = eatmpar[WSEE] ; break ;
		case FM0:	v = aint[IFSE]; 
					e = (e2int[IFSE]>0) ? sqrt(e2int[IFSE]) : bad; break ;
		case WM0:	v = aint[IWSE]; 
					e = (e2int[IWSE]>0) ? sqrt(e2int[IWSE]) : bad; break ;
		case ISOPL:	v = atmpar[ISOPL]; e = eatmpar[ISOPL] ; break ;
		case M2:	v = aint[IISK]; 
					e = (e2int[IISK]>0) ? sqrt(e2int[IISK]) : bad; break ;
		case FHEFF:	v = atmpar[FHEFF]; e = eatmpar[FHEFF] ; break ;
		case WHEFF:	v = atmpar[WHEFF]; e = eatmpar[WHEFF] ; break ;
		case TC:	v = atmpar[TC]; e = eatmpar[TC] ; break ;
		default:
			nr::nrerror("ATMOS: geterr(): wrong what-selection",nr::ERPAR);
			return bad ;
	}
	if (v>0 && v!=bad && e!=bad) 
		return e/v ;
	else
		return 0 ;
}

/*------------------------------ getncn2 -------------------------------*/

int atm::getncn2() {
	return ncn2 ;
}

/*------------------------------ getzcn2 -------------------------------*/

double atm::getzcn2(int i) {
	if (zcn2)
		return zcn2[i] ;
	else {
		nr::nrerror("ATMOS: getzcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ getcn2 --------------------------------*/

double atm::getcn2(int i) {
	if (cn2)
		return cn2[i] ;
	else {
		nr::nrerror("ATMOS: getcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ init() --------------------------------*/

inline void debug(const char* s) {
#ifdef ATMOSDEBUG
	printf("ATM:---------------%s----------------\n",s);
#endif
}

/*------------------------------ update --------------------------------*/

void atm::update(
	const char * wfile,
	const char * checkfile,
	double zshift1,
	double zshift2,
	double z0, double zmax, double dz, 
	double dzmin) {

const double KM2M = 1e3 ;


	if (nr::erget()) return ;

	if (!wfile) {
		nr::nrerror("ATMOS: update() on null weight file name", nr::ERNUL) ;
		return ;
	}
	
	/* Check the weight to coincide with the currently loaded by its file name */
	if (curwffile != string(wfile)) {
		/* weight is either not loaded or different */
		atwf.read(wfile) ;
	} else {
		/* Weight is the same. Check for the shifts and exit if they are so: */
		if ((zshift[0]==zshift1) & (zshift[1]==zshift2)) 
			goto CLEARLAB;
	}
	if (nr::erget()) return ;	/* returns if wf_t::read() got bad */
	
	/* Either weight functions or shifts have changed: redo decomposition */
	{	// this braket is to enclose declarations crossed by jump to CLEARLAB
	/* Cycle with 2 steps: 0-th for moments decomposition for the normal mode;
	   1-st for generalized (2 zshifts) mode */

	/* Check-file opening: */
	FILE * cf = 0 ;
	if (checkfile)
		if (!(cf=fopen(checkfile,"wt"))) {
			nr::nrerror((string("ATMOS: update(): problem to open the "
			"check-file ")+checkfile).c_str(), nr::ERFIO) ;
			return ;
		}

	/* Make the matrices of weights shifted by zshift1 and zshift2 */
	wf_t wfs[NMODE];
	/* Take care of atm::wcomb[][] if dimensions of atwf.ws[][] have changed: */
	if (atm::nw && atm::nw!=atwf.getnw()) {
		nr::free_dmatrix(wcomb,0,2*atm::nw-1,0,atm::nw);
		wcomb = 0;
	}
	atm::nw = atwf.getnw();

	int ibound = 0 ;	/* index of the boundary layer limit in WF z-grid */
	int nz = 0;
	
	// Save user's grid:
	atm::z0 = z0;
	atm::zmax = zmax;
	atm::dz = dz;
	atm::dzmin = dzmin;

	for (int mode=0; mode<NMODE; mode++)  {
		zshift[mode] = (mode ? zshift2 : zshift1) ;

		/* COMPUTE THE TRANSPOSED WEIGHT MATRIX FOR ALTITUDE POWERS
		DECOMPOSITION */
		// we make the decomposition on the "optimal" grid with the upper limit
		// equal to that given in the weight file:
		for (int imode=0; imode<=mode; imode++) {
			if (mode==0) {
				// Normal mode: omit lower atmosphere in decomposition:
				wfs[imode].setzgrid(atm::HBOUND,atm::zmax,atm::dz, atm::dzmin) ; 
			} else {
				if (!imode) wfs[imode].setaper(0);
				wfs[imode].setzgrid(atm::z0,atm::zmax,atm::dz, atm::dzmin) ; 
			}
			wfs[imode].interpolate(&atwf, zshift[imode]) ;	
			if (nr::erget()) return ;
		}

		nz = wfs[mode].getnz();
		if (mode==0) {
			if (nchan != atwf.getnaper()) {
				nchan = atwf.getnaper();
				/* Allocate the coefficients matrix: */
		        for (int i=0; i<NMODE; i++) 
					if (atm::coef[i]==0) {
//						cout << "Alloc "<<i<<"-th "<<NPOWER<<"x"<<atm::nw<<"\n";
						atm::coef[i] = nr::dmatrix(0,NPOWER-1,0,2*atm::nw-1) ;
					}
			}
		}
		/* SVD spaces */
		double ** wmatr=nr::dmatrix(1,nz,1,atm::nw*NMODE) ;
		double * wvect=nr::dvector(1,atm::nw*NMODE) ;
		double ** vmatr=nr::dmatrix(1,atm::nw*NMODE,1,atm::nw*NMODE) ;
		double * hvect=nr::dvector(0,nz-1) ;

		/* Find the zgrid-index of the boundary layer end: */

		for (int j=0; j<nz; j++) {
			if (wfs[mode].getalt(j)>=atm::HBOUND) {
				ibound = j;
				break ;
			}
		}

		if (ibound<0) {
			nr::nrerror("ATMOS: update(): failed to find the boundary layer"
			" index",nr::ERZGR); 
			goto CLEARLAB ;
		}
		if (cf) {
			fprintf(cf,"\nApproximated powers weight file %s, zshift=%5.2fkm", 
				wfile, zshift[0]);
			if (mode)
				fprintf(cf," and %5.2fkm",zshift[mode]);
			fprintf(cf,":");
		}
		/* Pointer to the string of previous use-weights flags */
		const char * prevuse = 0;
		int nwused=0;	// number of used weights for current approximation

		for (int ipow=int(atm::IFSE); ipow<=int(atm::IISK); ipow++) {
	
			/* Select the power: */
			double power ;
			switch (poweridx(ipow)) {
				case IFSE:
				case IWSE: power = POWSEE; break ;
				case IISP: power = POWISP; break ;
				case IEFF: power = POWEFF; break ;
				case IISK: power = POWISK; break ;
				default: nr::nrerror("ATMOS: update(): invalid power index",
					nr::ERPAR) ;
					return ;
			}
			if (cf) fprintf(cf, "\n# Ap%d: %6.3f ",(int)ipow,power) ;
			/* Generate the altitude powers: */
			for (int j=0; j<nz; j++) 
				hvect[j] = pow(wfs[mode].getalt(j)*KM2M, power) ;
			if (ipow==int(atm::IFSE) && ibound>0) {
				/* free atmosphere: force the hvect to resemble the smallest
					apertures pair weight behavior: */
				wfs[mode].copy(0,ibound-1,nchan,hvect) ;
				double d = wfs[mode].getval(ibound,nchan) ;
				for (int j=0; j<ibound; j++) hvect[j] /= d ;
				if (cf) fprintf(cf,"(W(AB)-dumped) ");
			}

			if (!prevuse || strcmp(prevuse,atm::USEWEIGHT[ipow])) {
				/* New SVD required: */
				/* Combined transposed matrix of weights for SVD */
				nwused = 0;
				for (int imode=0; imode<=mode; imode++) 
					for (int j=0; j<atm::nw; j++) 
						if (atm::USEWEIGHT[ipow][j+imode*atm::nw]!='0') {
							for (int i=0; i<nz; i++) 
								wmatr[i+1][nwused+1] = wfs[imode].getval(i,j) ;
							nwused++ ;
							if (cf) {
								if (imode)
									fprintf(cf,"%ss,",atwf.name(j));
								else
									fprintf(cf,"%s,",atwf.name(j));
							}
						}
				
				/* Run SVD of weight matrix: */
				nr::svdcmp(wmatr,nz,nwused,wvect,vmatr) ;
				if (nr::erget()) {mode=NMODE; goto ENDLAB ;}

				/* Maximal singular value: */
				double d = wvect[1] ;
				for (int i=2;i<=nwused;i++) 
					if (wvect[i]>d) d=wvect[i] ; /* maximal w */
				/* Minimal acceptable singular value: */
				d *= atm::MINRELW ;	

				int nbase = nwused ;
				/* Rejection of small singulars: */
				if (cf) fprintf(cf, " N(Base fun.): Init.=%d, ",nbase) ;
				for (int i=1;i<=nwused;i++) 
					if (wvect[i]<d) { 
						wvect[i] = 0 ; 
						nbase--; 
					}
				if (cf) fprintf(cf, " Smalls reject=> %d  ",nbase);
			}
			prevuse = atm::USEWEIGHT[ipow];

			/* Back-substitute to find the coefficients: */
			
			// normal mode coefficients for shifted weights and for M0 have to remain zero!
			for (int c=0; c<2*atm::nw; c++) atm::coef[mode][ipow][c] = 0;
			if (mode>0 || zshift[mode]!=0 || ipow!=atm::IWSE) {
				nr::svbksb(wmatr,wvect,vmatr,nz,nwused, hvect-1, 
					atm::coef[mode][ipow]-1);
				// now the coefficients are concentrated in first nwused
				// elements. Redistribute them starting from the tail:
				int k = nwused - 1 ;
				for (int c=atm::nw*(1+mode)-1; c>=0; c--)
					if (atm::USEWEIGHT[ipow][c]!='0') {
						atm::coef[mode][ipow][c] = atm::coef[mode][ipow][k] ;
						k--;
					} else
						atm::coef[mode][ipow][c] = 0 ;
			} // else - leave zero
		
		}
		if (cf) {
			/* Check the coefficients: */
			fprintf(cf,"\n# Coefficients:\n");
			for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
				/* Search for the maximal coefficient and compute the noise-factor
				NAF=sum(coef^2)/sum^2(coef): */
				double d = 0 ;
				for (int j=0; j<atm::nw*2; j++) {
					double ac = fabs(coef[mode][ipow][j]) ;
					if (d<ac) d=ac ;
				}
				double naf = 0, sumcoef = 0;
				for (int j=0; j<atm::nw*2; j++) {
					naf += coef[mode][ipow][j] * coef[mode][ipow][j] ;
					sumcoef += coef[mode][ipow][j] ;
				}
				if (sumcoef>0)	
					naf = naf/(sumcoef*sumcoef); 
				else
					if (naf>0) naf = 99.9 ; 

				fprintf(cf,"# Ap%d: Max=%9.3e NAF=%5.1f",ipow,d,naf);
				for (int j=0; j<atm::nw*2; j++) 
					fprintf(cf," %7.4f",d>0?atm::coef[mode][ipow][j]/d:0);
				fprintf(cf,"\n");
			}
			/* Check the approximated relations: */	
			fprintf(cf,"# h[km]  ") ;
			for (int i=0; i<=atm::IISK ; i++) 
				fprintf(cf,"#Ap%d/h^ ",i) ;
			fprintf(cf,"\n") ;
			for (int i=0;i<nz;i++) {
				fprintf(cf,"%6.3f",wfs[mode].getalt(i)) ;
				for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
			   		double power ;
					switch (poweridx(ipow)) {
						case IFSE:
						case IWSE: power = POWSEE; break ;
						case IISP: power = POWISP; break ;
						case IEFF: power = POWEFF; break ;
						case IISK: power = POWISK; break ;
						default: goto ENDLAB ;
					}
					/* power of altitude: */
					double a = pow(wfs[mode].getalt(i)*KM2M,power) ;
					double d = 0 ;
					for (int imode=0; imode<=mode; imode++)
						for (int j=0; j<atm::nw; j++)
							d+=wfs[imode].getval(i,j)*coef[mode][ipow][j+imode*atm::nw];
					fprintf(cf," %6.3f",a>0?d/a:0); 
				}
				fprintf(cf,"\n") ;
			}
		}
ENDLAB:
		/* Free the orthogonalized matrix, singular value vector and square 
		V-matrix: */
		nr::free_dmatrix(wmatr,1,nz,1,atm::nw*NMODE) ;
		nr::free_dvector(wvect,1,atm::nw*NMODE) ;
		nr::free_dmatrix(vmatr,1,atm::nw*NMODE,1,atm::nw*NMODE) ;
		nr::free_dvector(hvect,0,nz-1) ;
	}

	/* Close check-file */
	if (checkfile)
		fclose(cf) ;

	}	// closing braket to enclose declarations crossed by jump to CLEARLAB
CLEARLAB:
	/* reset counter of made measurements of integrals: */
	curmeas = 0 ;
	
	/* Reset measurement results: */
	atmpar[FSEE] = eatmpar[FSEE] = atmpar[WSEE] = eatmpar[WSEE] = atmpar[ISOPL] = eatmpar[ISOPL] = atmpar[WHEFF] = eatmpar[WHEFF] = 
	atmpar[FHEFF] = eatmpar[FHEFF] = atmpar[TC] = eatmpar[TC] = bad ;
	cn2method = '?' ;
	ncn2 = 0;	// hense, no (z)Cn2 zeroing needed
	if (aint && e2int) 
		for (int i=0; i<LAST; i++) aint[i] = e2int[i] = 0 ;

	// save the weight file name as a signal of successful work:
	curwffile = wfile ;
}


/*------------------------------ done --------------------------------*/

void atm::done(void) {
debug("atm::done()");
	atm::nchan = 0 ;
	
	if (atm::coef[0]){ 
		debug("delete coef[][]");
		nr::free_dmatrix(coef[0],0,NPOWER-1,0,atm::nw*2-1);
		nr::free_dmatrix(coef[1],0,NPOWER-1,0,atm::nw*2-1) ;
		coef[0] = coef[1] = 0 ;
	}

	/* Deallocate the storages for integrals: */
	if (cn2int) {
		debug("delete cn2int[][]");
		nr::free_dmatrix(cn2int,0,cn2intlen-1,0,NPOWER-1) ;
		cn2int = 0 ;
		cn2intlen = curmeas = -1 ;
	}
	
	if (wcomb) {
		// last parameter is an unused foo: nz actually
		nr::free_dmatrix(wcomb,0,2*atm::nw-1,0,atm::nw);
		wcomb = 0;
	}
	if (o_csi) {
		nr::free_dvector(o_csi,0,2*atm::nw-1);
		o_csi = 0;
	}

	atwf.setaper(0);		
	atm::nw = 0;

	debug("done Ok");
}

/*------------------------------ calcint --------------------------------*/

void atm::calcint(double * scind, double * scinds) {

	if (!coef[0] || !scind) {
		nr::nrerror("ATMOS: calcint() on empty index or coef. array", nr::ERNUL) ;
		return ;
	}
	/* (Re)allocate the storages for integrals: */
	if (curmeas>=cn2intlen) {
		if (atm::cn2intlen>0) {
			double**tmp=nr::dmatrix(0,cn2intlen+NMEAS-1,0,NPOWER-1) ;
			// in a cycle, reassign the rows which are existent removing 
			// the ones in tmp which are not needed. Leave all the new rows intact
			for (int i=0; i<curmeas; i++) {
				delete[] tmp[i];	// simply reassign the allocated array
				tmp[i] = cn2int[i]; // ...
			}
			delete[] cn2int;	// no nr::free_dmatrix!!! It will kill the powers
			cn2intlen += NMEAS;
			cn2int = tmp ;
		} else {
			cn2intlen = NMEAS;
			cn2int = nr::dmatrix(0,cn2intlen-1,0,NPOWER-1) ;
		}
	}

	/* We detect the generalized mode by the appearance of non-NULL scinds */
	bool isgen = (scinds!=0) ;

	/* Compute the Cn2-integrals from indices: */
	for (int i=0; i<NPOWER; i++) {
		double cn2mom = 0 ;
		for (int j=0; j<atm::nw; j++) {
		 	cn2mom += scind[j] * coef[isgen][i][j] ;
		}
		if (isgen) 
			for (int j=0; j<atm::nw; j++) 
			 	cn2mom += scinds[j] * coef[isgen][i][j+atm::nw] ;
		cn2int[curmeas][i] = cn2mom ;
	}
	curmeas++ ;
}

/*------------------------------ getseeing -------------------------------*/

const double JPOW_SEE = 0.6 ;	/* Power of Cn-integral for seeing */

/**
	@brief Convert the given integral of turbulence into seeing
	@param	cn2 Integral of Cn2(h) over atmosphere
	@param	secz	secant of zenith distance
	@return	Seeing [arcsec]
*/
inline double getseeing(double cn2, double secz) {
	const double LPOW_SEE = -0.2 ;	/* Power of lambda for seeing */
	return atm::KS * pow( cn2 / secz, JPOW_SEE) * 
			pow(atm::atwf.getleff(),LPOW_SEE) ;
}

/*------------------------------ avgint ----------------------------------*/

/** Compute the formal relative error of the value of the power function of the
value \em mean given its squared error \em e2 */
inline double relerr(double mean, double e2, double power) {
	if ((e2<=0) || (mean<=0)) 
		return 0 ;
	else
		return fabs(power * sqrt(e2) / mean) ;
}

void atm::avgint(double desi, double edesi, double secz) {
const double JPOW_ISP = -0.6 ;	/* Power of Cn-5/3th moment for isoplan. angle*/
const double LPOW_ISP = 1.2 ;	/* Power of lambda for isoplan. angle */
const double SPOW_TC  = -0.6 ; /* Power of DESI for atm. time constant */

	if (curmeas<1) {
		nr::nrerror("ATMOS: avgint() with no measurements done",nr::ERPAR) ;
	}

	/* average integrals: */
	sc::avgmatrix(cn2int,NPOWER,curmeas,0,0,aint,e2int,sc::MAXLAG) ;

	if (nr::erget()) return ;	/* for both two checks and avgmatrix() error */

	/* Convert integrals into atmospheric parameters */
	
	/* Seeing : free and whole atmosphere */	
	atmpar[FSEE] = atmpar[WSEE] = eatmpar[FSEE] = eatmpar[WSEE] = bad ;
	int isbadint = false ;
	if (aint[IFSE]>0) {
		atmpar[FSEE] = getseeing(aint[IFSE], secz) ;
		eatmpar[FSEE] = atmpar[FSEE] * 
			relerr(aint[IFSE],e2int[IFSE],JPOW_SEE);
		if (aint[IWSE]>0) {
			atmpar[WSEE] = getseeing(aint[IWSE], secz) ;
			eatmpar[WSEE] = atmpar[WSEE] * 
				relerr(aint[IWSE],e2int[IWSE],JPOW_SEE);
		} else {
			atmpar[WSEE] = eatmpar[WSEE] = 0;	// not generalized mode?
		}
	} else {
		isbadint = true ;
	}
	
	/* Isoplanatic angle: */
	atmpar[ISOPL] = eatmpar[ISOPL] = bad ;
	if (aint[IISP]>0) {
		double zpow = pow(secz,POWISP+1) ;
		atmpar[ISOPL] = KP * pow(aint[IISP] / zpow, JPOW_ISP)
			* pow(atwf.getleff(), LPOW_ISP) ;
		eatmpar[ISOPL] = atmpar[ISOPL] *
			relerr(aint[IISP],e2int[IISP],JPOW_ISP); 
	} else {
		isbadint = true ;
	}

	/* Effective altitude of turbulence in [meters]: (free atm.) */
	atmpar[FHEFF] = atmpar[WHEFF] = eatmpar[FHEFF] = eatmpar[WHEFF] = bad ;
	if (aint[IEFF]>0 && aint[IISP]>0) {
		double power = 1.0/(POWISP-POWEFF) ;
		double zpow = pow(secz, POWISP-POWEFF) ;
		atmpar[FHEFF] = pow(aint[IISP]/aint[IEFF]/zpow,power) ;
		double d = relerr(aint[IISP],e2int[IISP],power) ;
		eatmpar[FHEFF] = d*d;
		d = relerr(aint[IEFF],e2int[IEFF],power) ;
		eatmpar[FHEFF] += d*d;
		eatmpar[FHEFF] = sqrt(eatmpar[FHEFF]) * atmpar[FHEFF] ;
	} else {
		isbadint = true ;
	}
	
	/* Whole atm. effective altitude: */
	if (aint[IEFF]>0 && aint[IWSE]>0) {
		atmpar[WHEFF] = aint[IEFF]/aint[IWSE] / secz ;
		double d = relerr(aint[IEFF],e2int[IEFF],POWEFF) ;
		eatmpar[WHEFF] = d*d;
		d = relerr(aint[IWSE],e2int[IWSE],POWEFF) ;
		eatmpar[WHEFF] += d*d;
		eatmpar[WHEFF] = sqrt(eatmpar[WHEFF]) * atmpar[WHEFF] ;
	} 

	atmpar[TC] = eatmpar[TC] = bad ;
	if (desi>0) {
		atmpar[TC] = KT * pow(desi,SPOW_TC) ;
		eatmpar[TC] = atmpar[TC] * edesi * fabs(SPOW_TC) ;
	} else {
		isbadint = true ;
	}
	if (isbadint) {
		nr::nrerror("ATMOS: at least one Cn2-integral <0. No results saved",
			nr::ERNOD) ;
	}
}

/*---------------------------- meritfunc --------------------------------*/
/* Reopen atm once again: */
namespace atm {

int curnweight = 0 ;	// current number of weights (as 10 or 20) for meritfunc()

/**
	@brief	Merit function for the profile-restoration at fixed altitudes
	@param	y	vector[1..ncn2] of sqrt(Cn2) of layers (for powell()) 
					OR vector[\b 0..ncn2] of Cn2 
					if \a z[] is explicitly given (for floating layers)
	@param	z	vector of integer-type altitudes
	@return	merit Chi-square
	
*/
double meritfunc(double * y) ;
double atm::meritfunc(double * y, int * z) ;
}
// for powell():
double atm::meritfunc(double * y) {
	double chi2 = 0 ;	/* Khi-square */
	for (int j=0; j<curnweight; j++) {
		if (atm::FITSCIND[j]!='0') {
			double sc2 = 0 ;	/* scint. index */
			for (int i=0; i<ncn2; i++) 
				sc2 += y[i+1] * y[i+1] * wcomb[j][i] ;
			o_csi[j] = sicomb[j]-sc2 ;
			chi2 += o_csi[j]*o_csi[j] / e2sicomb[j] ;
			if (sicomb[j]>0) o_csi[j] /= sicomb[j] ;
		}
	}
	return chi2 ;
}
// for floating layers method:
double atm::meritfunc(double * y, int * z) {
	double chi2 = 0 ;	/* Khi-square */
	for (int j=0; j<curnweight; j++) {
		if (atm::FITSCIND[j]!='0') {
			double sc2 = 0 ;	/* scint. index */
			for (int i=0; i<ncn2; i++) 
				sc2 += y[i] * wcomb[j][z[i]] ;
			o_csi[j] = sicomb[j]-sc2 ;
			chi2 += o_csi[j]*o_csi[j] / e2sicomb[j] ;
			if (sicomb[j]>0) o_csi[j] /= sicomb[j] ;
		}
	}
	return chi2 ;
}

/*---------------------------- geto_csi ----------------------------------*/

const double * atm::geto_csi() { return atm::o_csi; }

/**
	@brief Matrix SVD invertor
	@param	ainv	Returned inverted matrix
	@param	a	Matrix to be inverted (content is \b modified on return!)
	@param	nrow	(row number in a) = (column number in ainv)
	@param	ncol	(column number in a) = (row number in ainv)
	@param	w	Decomposition work vector of singulars [1..ncol]
	@param	v	Decomposition work orthogonal matrix [1..ncol][1..ncol]
	@param	wcutoff	Minimal acceptable ratio of a w-element to maximal w-element
		for deletion of small singulars in w-vector
	
	The SVD decomposition of input matrix \em a is done and then the
	multiplication of resulted \em v, \em diag(1/w) and decomposed orthogonal
	matrix \em u being the transposed matrix on the place of \em a is done. This
	product is computed in \em ainv.
*/
void invwmatr(double ** ainv, double ** a, int nrow, int ncol, 
	double * w, double ** v, double wcutoff) {

	nr::svdcmp(a,nrow,ncol,w,v) ;
	/* Cancel small singulars */
	double d = 0 ;
	for (int i=1; i<=ncol; i++)	if (d<w[i]) d=w[i];
	d *= wcutoff ;	/* Minimal acceptable singular value */
	for (int i=1; i<=ncol; i++)
		if (w[i]>d) 
			w[i] = 1. / w[i] ;
		else 
			w[i] = 0 ;

	/* multiply three matrices: W^{-1}=(V*diag(1/w))*U^T */
	for (int i=1; i<=ncol; i++) 
		for (int j=1; j<=nrow; j++) {
			ainv[i][j] = 0 ;
			for (int k=1; k<=ncol; k++)
				ainv[i][j] += v[i][k]*w[k]*a[j][k] ;
		}
}


/*---------------------------- calcn2 ----------------------------------*/

double atm::calcn2(double * scind, double * scinds,
		double * e2scind, double * e2scinds, double secz, char meth) {

const double AVGCN2	= 6.95e-13 ;	/* Cn2-integral giving 1" seeing */
const double FTOL = 1e-3 ;		/* Tolerance in powell() */

	if (nr::erget()) return bad;

	/* COMPUTE THE COMBINED WEIGHT MATRIX FOR CN2-profile RESTORATION.
	It is computed on a grid which is stretched with a factor of secant(z_dist)
	with respect to the "standard" grid with equal step DZCN2=1km. Meanwhile, the
	altitudes in zcn2[] are assigned as seen in zenith, with step DZCN2. */
	
	/* Secant of the zenith distance stretchs the zcn2[] grid and amplifies
	the strengths cn2[], so needs to be correctred twice! */

	int nz = 0;
	bool isgen = (scinds!=0) ;	// generalized mode flag

	double cn2zmax = atm::zmax*secz;
	if (cn2zmax>wf::ZMAX) cn2zmax = wf::ZMAX ;
	
	// equidistant grid with a step corrected by secant of zenith distance:
	wf_t wfs;	// computed twice in this loop: for shifted and not weights
	// - we need this weight structure for altitude referencing further
	for (int mode=0; mode<=isgen; mode++) {
		double cn2z0 = isgen ? Z0CN2_GEN*secz : Z0CN2_NORM*secz ;

		double cn2minzstep = (meth==FIXEDLAY) ? 
			DZMINCN2_FIX*secz : DZMINCN2_FLO*secz ;

		double cn2zstep = (meth==FIXEDLAY) ? 
			DZCN2_FIX : DZCN2_FLO ;
		// step (fixed) must be stretched; modifier (prop. grid) - not
		if (cn2minzstep==0) cn2zstep *= secz;	

		// clear wfs (works before second turn in gen. mode):
		if (mode) wfs.setaper(0);
			
		wfs.setzgrid(cn2z0,cn2zmax,cn2zstep,cn2minzstep) ;
		wfs.interpolate(&atwf, zshift[mode]) ;
		nz = wfs.getnz();
		if (!wcomb && !mode) {
			// create weight matrix for Cn2 restoration:
			wcomb = nr::dmatrix(0,2*atm::nw-1,0,nz-1);				
		}
		if (!mode) {
			// create the model indices O-C vector:
			if (!o_csi) o_csi = nr::dvector(0,2*nw-1);
			for (int w=0; w<nw*2; w++) o_csi[w] = 0;
		}
		for (int w=0; w<nw; w++)
			wfs.copy(0,nz-1,w,wcomb[w+mode*nw]);
	}

#ifdef ATMOSDEBUG
// show the altitude grid for restoration:
for (int i=0; i<nz; i++) {
	cout << wfs.getalt(i)/secz << ' ';
}
cout << '\n';
#endif

	/* Number of weights i.e. measured indices (not all may be used! see
		atm::FITSCIND): */
	atm::curnweight = atm::nw*(1+isgen);

	/* we need the combined indices and err-s global vectors for merit-function */
	atm::sicomb =   nr::dvector(0,curnweight-1) ;
	atm::e2sicomb = nr::dvector(0,curnweight-1) ;

	for (int i=0; i<nw; i++) {
			sicomb[i] = scind[i] ;
			e2sicomb[i] = e2scind[i] ;
	}

	if (isgen)
		for (int i=0; i<nw; i++) {
			/* works in generalized mode only */
			sicomb[i+nw] = scinds[i] ;
			e2sicomb[i+nw] = e2scinds[i] ;
		}

	/* arrays for selected matrix inversion: far larger than needed */
	double **w4=nr::dmatrix(1,curnweight,1,MAXNLAYER) ;
	double **w4inv=nr::dmatrix(1,MAXNLAYER,1,curnweight) ;
	double *wvect=nr::dvector(1,MAXNLAYER) ;
	double **vmatr=nr::dmatrix(1,MAXNLAYER,1,MAXNLAYER) ;
	double minchi2 = MAXCHI2 ;	/* arbitrary large number */
	double m0prof = 0 ;
	if (meth==FIXEDLAY) {
		/* Fixed layers method */
		/* Generate the grids: altitudes and 1st approximation of the profile */
		atm::ncn2 = NFIXED + isgen ;
		if (atm::ncn2>nz) atm::ncn2 = nz;
		const double cn2_0 = sqrt(AVGCN2 / atm::ncn2) ;
		// altitude grid must be cut to the demanded number of layers! say,@16km
		for (int i=0; i<ncn2; i++) {
			atm::zcn2[i] = wfs.getalt(i) / secz ;
			atm::cn2[i] = cn2_0;
		}

		/* Diagonal "identity" matrix: */
		double ** xi = nr::dmatrix(1,atm::ncn2,1,atm::ncn2) ;
		for (int i=1; i<=ncn2; i++) 
			for (int j=1; j<=ncn2; j++) xi[i][j] = (double)(i==j) ;
			
		/* Solve the task: find the minimum chi-square */
		int iter = 0 ;
		nr::powell(cn2-1,xi,ncn2,FTOL,&iter,&minchi2,atm::meritfunc);
		if (nr::erget()==nr::recipes_ercode) {
			ncn2 = 0;	// not converged
			minchi2 = MAXCHI2;
		} else {
			/* Back-square the profile, correct for sec(z) and compute the sum: */
			for (int i=0; i<ncn2; i++) {
				cn2[i] = (cn2[i]*cn2[i])/secz ;
				m0prof += cn2[i] ;
			}
		}
		nr::free_dmatrix(xi,1,ncn2,1,ncn2) ;
	} else {
		/* "Floating" layers method */
		atm::ncn2 = NFLOAT + isgen ;
		double * tr_cn2=nr::dvector(0,ncn2-1) ;// trial Cn2 vector
		int ih[NFLOAT+1] ;	// memory is either exact (gen) or larger (norm)
		// storage of best-chi2 residuals 
		double * o_csi0 = nr::dvector(0,atm::curnweight-1) ;
		// altitude index limits for layers:
		int maxi[NFLOAT+1] = {0};	// bottom layer is fixed if isgen==true
		maxi[NFLOAT+isgen-1] = nz - 1;
		for (int i=0; i<NFLOAT-1; i++) 
				for (int j=0; j<wfs.getnz(); j++) 
					if (wfs.getalt(j)>MAXHLAY[i]*secz) {
						maxi[i+isgen]=j; 
						break;
					}

		if (isgen) ih[0] = 0 ; /* Fixed ground layer */
		// Wander with (three) floating layers which altitude indices are ih[i]:
		for (int i=isgen; i<NFLOAT+isgen; i++) ih[i] = (i>0)?ih[i-1]+1:0;
		do {
			/* Compose the selected weights matrix w4 */
			for (int j=0;j<curnweight;j++)
				for (int i=0;i<ncn2; i++)
					w4[j+1][i+1] = wcomb[j][ih[i]] ;

			/* Invert w4-matrix */
			invwmatr(w4inv, w4, curnweight, ncn2, 
				wvect, vmatr, MINRELW) ;

			/* Compute the strengths: */
			for (int i=0; i<ncn2; i++) {
				tr_cn2[i] = 0 ;
				for (int j=0; j<curnweight; j++)
					tr_cn2[i] += w4inv[i+1][j+1] * sicomb[j] ;
				if (tr_cn2[i]<0) tr_cn2[i] = 0 ;	  /* >=0 constraint */
			}

			/* Compute indices and chi-square: */
			double chi2 = meritfunc(tr_cn2, ih);

			/* Save the altitudes and strengths if this is the best
				case up to now */
			if (chi2<minchi2) {
				minchi2 = chi2 ;
				m0prof = 0 ;
				for (int i=0; i<ncn2; i++) {
					zcn2[i] = wfs.getalt(ih[i]) / secz ;
					// correct for the sec(z) and save:
					cn2[i] = tr_cn2[i] / secz ;
					m0prof += cn2[i] ;
				}
				// save the residuals; they'll modify further for worse Chi2!
				for (int i=0; i<atm::curnweight; i++) o_csi0[i] = o_csi[i];
			}
			// modify the layer indices:
			bool atlim = false;
			for (int i = NFLOAT+isgen-1; i>=isgen; i--) {
				if (ih[i]<maxi[i]) {
					ih[i]++;
					if (atlim) {
						for (int j=i+1; j<=NFLOAT+isgen-1; j++) ih[j] =
							ih[j-1]+1;
						atlim = false;
					}
					break ;
				} else {
					atlim = true;
					if (i==isgen) 
						goto ENDFLOAT;	// end of iterations
				}
			}
		} while (1) ; // caught here the signal to stop*/
		ENDFLOAT:
		nr::free_dvector(tr_cn2,0,ncn2-1) ;
		// restore the best residuals from saved and remove latters:
		for (int i=0; i<atm::curnweight; i++) o_csi[i] = o_csi0[i];
		nr::free_dvector(o_csi0,0,atm::curnweight-1);
	}

	/* Save the method identity: */
	cn2method = meth ;

#ifdef TESTIWW
	/* Compose the selected weights matrix w4 (having maybe 6 rows, not 4) */
	for (int j=0;j<curnweight;j++)
		for (int i=0;i<ncn2; i++) {
			int k = int(rint(zcn2[i])) ;
			w4[j+1][i+1] = wcomb[j][k] ;
		}
	/* Invert w4-matrix */
	invwmatr(w4inv, w4, curnweight, ncn2, wvect, vmatr, MINRELW) ;
	
	for (int h=0; h<=atm::zmax; h++) {
		printf("IWW %2d\t",h);
		double tot = 0;
		double prof = 0;
		for (int i=0; i<ncn2; i++) {
			double f = 0;
			for (int j=0; j<curnweight; j++) 
				f += w4inv[i+1][j+1] * wcomb[j][h] ;
			prof += f*cn2[i];
			tot += f;
			printf("%5.2f\t",f);
		}
		if (prof<0) prof=0;
//		printf("%8.2e\n",prof);
		printf("\n");
	}
#endif
	if (minchi2>MAXCHI2)	minchi2 = MAXCHI2 ;
	cn2[ncn2] = minchi2 ;	/* save chi-square in last element */
	nr::free_dmatrix(w4,1,curnweight,1,MAXNLAYER) ;
	nr::free_dmatrix(w4inv,1,MAXNLAYER,1,curnweight) ;
	nr::free_dvector(wvect,1,MAXNLAYER) ;
	nr::free_dmatrix(vmatr,1,MAXNLAYER,1,MAXNLAYER) ;
	nr::free_dvector(sicomb,0,curnweight-1) ;
	nr::free_dvector(e2sicomb,0,curnweight-1) ;

	nr::free_dmatrix(wcomb,0,2*nw-1,0,nw);// last parameter is an unused foo
	wcomb = 0;

	/* compute and return the seeing which corresponds to restored profile.
	Zenith distance is set to 1 since Cn2[] is already corrected for it */
	return (cn2seeing=getseeing(m0prof,1)) ;
}

/**
	@brief	Get the value of the instrument profile centered at \em z0
	@param	z	Altitude to give the result for
	@param	z0	Center altitude for the profile
	@param	resol 	relative resolution in altitude (FWHM/altitude)
	@param	resmin 	minimal altitude resolution (at lowest z) [km]
	@return	PSF value
	
	The instrument profile with the altitude resolution \em res is an arbitrary
	function symmetric in the domain of the altitude logarithm. At lowest
	altitudes, the gaussian is with half-width \em resmin in a linear domain of
	altitudes. The MASS resolution \em res is about 0.5. In current
	implementation, the PSF is a gaussian.
*/
inline double psf(double z, double z0, double resol, double resmin) {
	double ldx = (resmin<resol*z0) ? (log(z)-log(z0))/log(resol) :
				(z-z0)/resmin ;
	return exp(-ldx*ldx) ;
}

void atm::interpolate(double * z, double * intcn2, int nz, 
	double * zin, double * cn2in, int nzin) {
	if (!z || !intcn2 || nz<1) {
		nr::nrerror("ATMOS: intercn2() with null arguments",nr::ERNUL);
		return ;
	}

	const double resol = 0.4 ;	// altitude resolution of MASS
	const double resmin = 0.5; // minimal absolute resolution [km]

	/* In case the default parameters are left, override them with last-computed
	profile data: */
	double * zprof = (zin) ? zin : zcn2;
	double * cn2prof = (cn2in) ? cn2in : cn2;
	int nprof = (nzin) ? nzin : ncn2 ;

	int iupper = 0 ;	// index of zprof[] node which is above the current alt.
	int ilower = -1 ;	// index of zprof[] node which is below the current alt.
	for (int i=0; i<nz; i++) {
		if ((z[i]>zprof[iupper]) && (iupper<nprof)) {
			iupper++ ; ilower++ ;
		}
		// distance to lower layer in units of resolution:
		const double LARGE = 10 ;
		double dl = LARGE ;
		if (ilower>=0) {
			double curres = resol * zprof[ilower] ;
			if (curres<resmin) curres = resmin ;
			dl = fabs(z[i]-zprof[ilower])/curres ;
		}
		// distance to upper layer in units of resolution:
		double dh = LARGE ;
		if (iupper<nprof) {
			double curres = resol * zprof[iupper] ;
			if (curres<resmin) curres = resmin ;
			dh = fabs(z[i]-zprof[iupper])/curres ;
		}
//	printf("%d: %d %d\n",i,ilower,iupper);
		if (dl+dh<2) {
			// two close layers: linear interpolation
			double dz = (z[i]-zprof[ilower]) / (zprof[iupper]-zprof[ilower]) ;
			intcn2[i] =  cn2prof[ilower] * (1-dz) + cn2prof[iupper] * dz ;
		} else {
//	printf("%d: %3.1f %3.1f\n",i,dl,dh);
			double lcn2 = (ilower>=0) ? cn2prof[ilower] * psf(z[i], zprof[ilower], resol, resmin) : 0 ;
			double hcn2 = (iupper<nprof) ? cn2prof[iupper] * psf(z[i], zprof[iupper], resol, resmin) : 0 ;
			intcn2[i] = lcn2 + hcn2 ;
		}
	}
}


/*----------------------------- write -------------------------------*/

void atm::write(FILE * f, const char * stamp, atm::what what, 
	bool header) {

	if (!f) nr::nrerror("ATMOS: write() with null FILE", nr::ERPAR) ;
	if (nr::erget()) return ;
	
	/* write the stamp: */
	if (stamp)
		if (!fprintf(f, "%s",stamp)) {
			nr::nrerror("ATMOS: write(): cannot write file", nr::ERFIO) ;
			return ;
		}
		
	switch (what) {
		case ATMPAR:
		for (int i=0; i<int(atm::LAST); i++) {
			atm::ordint id = atm::ordint(i) ;
			const char * fl = getfmt(id) ;
			if (header) {
				const char * pn = getname(id) ;// parameter name
				// length of a number in current format: NO FLAGS ASSUMED
				char * endptr ;
				int fmtlen = strtol(rindex(fl,'%')+1, &endptr, 10) ;
				string heading = string(ATMDLM)+pn;
				int spaceadded = fmtlen-strlen(pn) ;
				for (int c=0; c<spaceadded;c++) heading+=' ';
				// length of an accuracy number:
				fmtlen = strtol(rindex(ERRFMT,'%')+1, &endptr, 10) ;
				heading += string(ATMDLM)+"e_"+pn;
				spaceadded = fmtlen-strlen(pn)-2 ;	// last "2" is len("e_")
				for (int c=0; c<spaceadded;c++) heading+=' ';
				fprintf(f,"%s",heading.c_str()) ;
			} else {
				/* write the average integrals and their errors: */
				double d = getval(id); 
				double e = geterr(id) ;
				fprintf(f,fl,atm::ATMDLM,d) ;
				fprintf(f,atm::ERRFMT,atm::ATMDLM,e) ;
			}
		}
		break;
		case CN2PROF:
			if (header) {
				fprintf(f, "%sMet%sNz%sChi2  ",ATMDLM, ATMDLM, ATMDLM) ;
				// new addition: seeing by cn2 profile:
				fprintf(f,"%sSeeCn", ATMDLM);
				for (int i=0; i<NFIXED; i++)
					fprintf(f,"%sz_%d %sCn2_%d   ",ATMDLM,i+1,ATMDLM,i+1) ;
			} else {
				/* save the dimention, altitudes and strengths of the restored 
					profile: */
				fprintf(f,"%s  %c%s%2d",ATMDLM,cn2method,ATMDLM,getncn2());
				fprintf(f,CHI2FMT, ATMDLM,getcn2(getncn2())) ;
				// new addition: seeing by cn2 profile:
				fprintf(f,ANGFMT, ATMDLM, cn2seeing);
				for (int i=0; i<getncn2(); i++) {
					fprintf(f,ALTFMT,ATMDLM,getzcn2(i)) ;
					fprintf(f,CN2FMT,ATMDLM,getcn2(i)) ;
				}
			}
		break;
		case O_CSCIND:
			if (header) {
				fprintf(f, "%sMet",ATMDLM);
				// length of an accuracy number:
				char * endptr = 0;
				int fmtlen = strtol(rindex(ERRFMT,'%')+1, &endptr, 10) ;
				string heading = string(ATMDLM)+"R_";
				string spaces;
				int spaceadded = fmtlen-2-2 ;	// 2+2 is strlen("R_")+strlen("00")
				for (int c=0; c<spaceadded;c++) spaces+=' ';
				for (int i=0;i<2*atm::nw;i++) 
					fprintf(f,"%s%-2d%s",heading.c_str(),i,spaces.c_str()) ;
			} else {
				/* write the method identity: */
				fprintf(f,"%s  %c",ATMDLM,cn2method);
				/* Write the normalized index residuals (obseved-model) : */
				for (int i=0;i<curnweight;i++) 
					fprintf(f,atm::ERRFMT,atm::ATMDLM,o_csi[i]);
			}

	}
	/* Set the carriage-return character: */
	if (!fprintf(f,"\n"))
		nr::nrerror("ATMOS: write(): cannot write file 2", nr::ERFIO) ;
}
