/*--------------------------------------------------------------------------*/
/** @file nr.h
                  MASS project: TURBINA module file                       
                       header file for nr.c                             
                                                                          
 The module NR contains the set of procedures from "Numerical Recipes in C: The
 art of scientific programming". All the files of the used recipes are
 concatenated in nr.c by means of #include directives. At the top of this concatenation, in this header file, 
 we set a heading 
 \verbatim #define float double \endverbatim 
 to make all the mathematics done in double-precision, as accepted in MASS 
 Software (see nrutil.h documentation). 

 The memory- and error-handling utilities are given in a separate module
 NRUTIL. 

 \b Usage: 
 
 The memory allocation within the MASS "scientific" modules (scind, weif, scan
 and atmos) is done with dvector(), dmatrix() for floating point arrays;
 deallocation is made with respective free_dvector() and free_dmatrix(). Thus,
 we use explicitly the double-type arrays. 

 The only modified scientific recipe is the \em qromo() integrator which is 
 given in qromo2.c source file to implement the relative convergence criterion.
 The use of all the other recipes is normal, as described in the Book. They are
 only modified to conform C++ conventions and return after nrerror() call.

 @version 1.5: Adaptation of recipes to be compatible with C++              */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/


#ifndef _NR_H
#define _NR_H

/** All the mathematics with "float"-type is done actually in double-type in
nr.c, nrutil.cpp and MASS applications */
#define float double

/* Force C++ compilation of nr and nrutil modules: */
#ifndef NR_CPP
#define NR_CPP
#endif

//#include <stdio.h> /* - needed in for commented out spctrm() */

#ifdef __cplusplus

#ifndef NR_CPP
extern "C" {
#else
/**	Numerical Recipes namespace. */
namespace nr {
#endif
#endif

/* Here follows the list of C++-modified recipes greped from the original nr.h
(ANSI part): */

	void  bcucof(float *y, float *y1, float *y2, float *y12, float d1,
		float d2, float **c);
	void  bcuint(float *y, float *y1, float *y2, float *y12, float x1l, 
		float x1u, float x2l, float x2u, float x1, float x2, 
		float *ansy, float *ansy1, float *ansy2);
	float bessj1(float x);
	float brent(float ax, float bx, float cx, float (*f)(float), float tol,
		float *xmin);
	float f1dim(float x);
	void  four1(float *data, int nn, int isign);
	float gammln(float xx);
	float gasdev(int *idum);
	void  hunt(float *xx, int n, float x, int *jlo);
	void  linmin(float *p, float *xi, int n, float *fret, float (*func)(float*));
	void  locate(float *xx, int n, float x, int *j);
	float midpnt(float (*func)(float), float a, float b, int n);
	void  mnbrak(float *ax, float *bx, float *cx, float *fa, float *fb, 
		float *fc, float (*func)(float));
	float poidev(float xm, int *idum);
	void  polcoe(float *x, float *y, int n, float *cof);
	void  polcof(float *xa, float *ya, int n, float *cof);
	void  polin2(float *x1a, float *x2a, float **ya, int m, int n, float x1, 
		float x2, float *y, float *dy);
	void  polint(float *xa, float *ya, int n, float x, float *y, float *dy);
	void  powell(float *p, float **xi, int n, float ftol, int *iter,
		float *fret, float (*func)(float *));
/** 
	@brief	Modified integrator for weight calculations 
	@param	...	see the NR Book
	@param	EPS	>0: absolute convergence, as in original qromo() 
	where EPS is a definition;
	<0: signals to use the relative convergence
	@return void
*/	
	float qromo(float (*func)(float), float a, float b,
		float (*choose)(float (*)(float),float,float,int)
		,double EPS); /* Last "double" stays for EPS in modified proc.in qromo2.c */
	float ran1(int *idum);
	float ran2(long *idum);
	void  ratint(float *xa, float *ya, int n, float x, float *y, float *dy);
	void  sort(int n, float *ra);
	void  splie2(float *x1a, float *x2a, float **ya, int m, int n,
		float **y2a);
	void  splin2(float *x1a, float *x2a, float **ya, float **y2a, int m,
		int n, float x1, float x2, float *y);
	void  spline(float *x, float *y, int n, float yp1, float ypn, float *y2);
	void  splint(float *xa, float *ya, float *y2a, int n, float x, float *y);
	void  svbksb(float **u, float *w, float **v, int m, int n, float *b,
		float *x);
	void  svdcmp(float **a, int m, int n, float *w, float **v);
	void  svdfit(float *x, float *y, float *sig, int ndata, float *a, 
		int ma, float **u, float **v, float *w, float *chisq,
		void (*funcs)(float,float *,int));
	void  svdvar(float **v, int ma, float *w, float **cvm);

/* end of the content of the standard nr.h */

#ifdef __cplusplus
}	/* closing either 'extern "C"' or 'namespace nr' */
#endif

#undef float	/* to avoid the conflicts of operations for float and double */

#endif
