/*
	C++ implementation of the Numerical Recipes utilities. nrerror() can
	generate the exception with a C-string argument (see nrutil.h for
	declaration of this exception class) or set the error code and message in
	global variables. All malloc() and free() calls are  replaced with new[]
	and delete[] operators (which generate exception bad_alloc() in standard
	C++). The option NR_DGN allows to dump the allocation/deallocation
	information on the standard output stream.

	@author	N.Shatsky (20.12.2001)
*/

#include "nrutil.h"

#define float double

/* Reopen nr: */
namespace nr {
	int ercode = 0 ;
	std::string ermsg ;

	/* Error messages corresponding to sc::ER... codes: */
	const int ERCODEMSGLEN = 60 ;
	const char ercodemsg[-nr::ERHI+1][ERCODEMSGLEN] = {
	"NO ERROR",   /* 0: Ok */
	"Numerical Recipes error",				/* -1 */
	"MEMORY allocation error",      		/* -2 */
	"NULL pointer supplied", 				/* -3 */
	"Allocation attempt on non-NULL array",	/* -4 */
	"Non-sense supplied parameter",   		/* -5 */

	"Bad file info or I/O error",    		/* -6 */
	"Not expected count series", 			/* -7 */
	"No/zero data supplied",				/* -8 */
	"z-grids are different", 				/* -9 */
	"Spectrum is incompatible",				/*-10 */
	"Apertures number or geometry deviate from needed", /* -12 */
	"Spectrum and Response curve do not match in w/l"/* -13 */
	} ;
}

int nr::erget() { return ercode ; }

void nr::erreset() { ercode = 0 ; }

const std::string& nr::ermessage() { return ermsg ; }

const char * nr::ercodemessage(int ercode) {
 
        if ((ercode<=ERLO) && (ercode>=ERHI)) {
                return nr::ercodemsg[-ercode];
        } else {
                return "";
        }
}


void nr::nrerror(const char * error_text, int error ) {
	if (!ercode) {
		ercode = error ;
		ermsg = error_text ;
	}
}

/**
	Template function for creation of vector arrays of any type \b T: 
*/
template<class T> T* anyvector (int nl, int nh) {
	T * v = new T[nh-nl+1] ;
	v -= nl ;
#ifdef NR_DGN
	cout << "#---------- allocate " << (v-nl) << " vector[" 
		<< nl << ',' << nh << "]\n" ;
#endif	
	return v ;
}

/**
	Template function for deletion of vector arrays of any type \b T: 
*/
template<class T> void free_anyvector (T * v, int nl, int nh) {
	delete[](v+nl) ;
#ifdef NR_DGN
	cout << "#-------- DEallocate " << (v-nl) << " vector[" 
		<< nl << ',' << nh << "]\n" ;
#endif	
	return ;
}

float*   nr::vector(int nl, int nh) { return anyvector<float>(nl,nh) ; }
double* nr::dvector(int nl, int nh) { return anyvector<double>(nl,nh) ; }
int*    nr::ivector(int nl, int nh) { return anyvector<int>(nl,nh) ;   }

void  nr::free_vector(float * v, int nl, int nh) 
{ free_anyvector<float>(v,nl,nh); return ; }
void nr::free_dvector(double* v, int nl, int nh) 
{ free_anyvector<double>(v,nl,nh); return ; }
void nr::free_ivector(int   * v, int nl, int nh) 
{ free_anyvector<int>(v,nl,nh) ; return ; }

/**
	Template function for creation of matrix of any type \b T
*/
template<class T> T** anymatrix(int nrl,int nrh, int ncl, int nch) {
	T **m = new (T*)[nrh-nrl+1] ;
	m -= nrl;

	for(int i=nrl; i<=nrh; i++) {
		m[i] = new T[nch-ncl+1] ;
		m[i] -= ncl;
	}
#ifdef NR_DGN
	cout << "#---------- allocate " << m << " matrix[" << 
		nrl << ',' << nrh << "][" << ncl << ',' << nch << "]\n" ;
#endif
	return m ;
}

float  **nr::matrix(int nrl, int nrh, int ncl, int nch) 
{ return anymatrix<float>(nrl,nrh,ncl,nch) ; }
double **nr::dmatrix(int nrl, int nrh, int ncl, int nch) 
{ return anymatrix<double>(nrl,nrh,ncl,nch) ; }
int **nr::imatrix(int nrl, int nrh, int ncl, int nch) 
{ return anymatrix<int>(nrl,nrh,ncl,nch) ; }


/**
	Template function for deletion of matrix of any type \b T
*/
template<class T> void free_anymatrix(T ** m, int nrl,int nrh, int ncl, int nch) 
{
	for(int i=nrh; i>=nrl; i--) delete[] (m[i]+ncl) ;
	delete[] (m+nrl) ;
#ifdef NR_DGN
	cout << "#-------- DEallocate " << m << " matrix[" << 
		nrl << ',' << nrh << "][" << ncl << ',' << nch << "]\n" ;
#endif
	return ;
}

void  nr::free_matrix(float  **m, int nrl, int nrh, int ncl, int nch)
{ free_anymatrix<float> (m,nrl,nrh,ncl,nch) ; return ; }
void nr::free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch)
{ free_anymatrix<double>(m,nrl,nrh,ncl,nch) ; return ; }
void nr::free_imatrix(int    **m, int nrl, int nrh, int ncl, int nch)
{ free_anymatrix<int>   (m,nrl,nrh,ncl,nch) ; return ; }


/**
	Template function for creation of a submatrix of any type \b T
*/
template<class T> T** nrsubmatrix(T** a, 
	int oldrl,int oldrh, int oldcl, int oldch, int newrl, int newcl) {
	T **m = new (T*)[oldrh-oldrl+1] ;
	m -= newrl;

	for(int i=oldrl, j=newrl; i<=oldrh; i++,j++) m[j] = a[i]+oldcl-newcl; 
#ifdef NR_DGN
	cout << "#---------- allocate " << m << " submatrix[" << 
		newrl << ",...][" << newcl << ",...] from " << a << "\n" ;
#endif
	return m ;
}

float **nr::submatrix(float **a, int oldrl, int oldrh, int oldcl, int oldch,
	int newrl, int newcl) 
{ return nrsubmatrix<float>(a,oldrl,oldrh,oldcl,oldch,newrl,newcl) ; }


/**
	Template function for deletion of a submatrix of any type \b T
*/
template<class T> void nrfree_submatrix(T** b,int nrl,int nrh,int ncl,int nch) {
	delete[] (b+nrl) ;
#ifdef NR_DGN
	cout << "#-------- DEallocate " << b << " submatrix[" << 
		nrl << ',' << nrh << "][...]\n" ;
#endif
	return ;
}

void nr::free_submatrix(float **b, int nrl, int nrh, int ncl, int nch) 
{ nrfree_submatrix<float>(b,nrl,nrh,ncl,nch) ; return ; }

/**
	Template function for conversion of a matrix represented as a 1-dim array 
	into a NR-compatible 2-dim array of any type \b T
*/
template<class T> T** nrconvert_matrix(T* a, int nrl,int nrh, int ncl, int nch) 
{
	int nrow=nrh-nrl+1;
	int ncol=nch-ncl+1;
	T **m = new (T*)[nrow] ;
	m -= nrl;

	for(int i=0, j=nrl; i<=nrow-1; i++,j++) m[i] = a+ncol*i-ncl ;
#ifdef NR_DGN
	cout << "#---------- allocate " << m << " convert-matrix[" << 
		nrl << ',' << nrh << "][" << ncl << ',' << nch << " from " << a << "\n";
#endif
	return m ;
}

float **nr::convert_matrix(float *a, int nrl, int nrh, int ncl, int nch)
{ return nrconvert_matrix<float>(a,nrl,nrh,ncl,nch); }

void nr::free_convert_matrix(float **b, int nrl, int nrh, int ncl, int nch)
{ nrfree_submatrix<float>(b,nrl,nrh,ncl,nch) ; return ; }

