/***************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            scan.cpp                                      */
/*                    Description is given in scan.hxx                      */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition SCANTEST in compilation for the module self-tests:
gcc -c scan.c .... -DSCANTEST=1 */
//#define SCANTEST

#include <math.h>
#include "scan.hxx"	// for count_t only
#include "nrutil.h"
#include "nr.h"	/* nrutil.h and nr.h must be included as last includes! */

#ifdef SCANTEST
#include "scind.hxx"	// for sc::apername() only
/*---------------------------- TEST PART ---------------------------------*/

#define NTEST 100

/* Define this to view instant scans with xmgrace: */
//#define VIEWSCAN
/* Define this to view merged scans with xmgrace: */
//#define VIEWMSCAN
#if defined(VIEWSCAN) || defined(VIEWMSCAN)
#undef NTEST
#define NTEST 1
#endif

/*----- Comes normally from configuration file: cf_... vars and defs --------*/

/* Apertures: */
/* number of apertures in segmentator : */
const int CF_NCHAN = 4 ;
const int CF_NSCAN = 25 ;
/* Physical size of segmentator annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.102, 0.188, 0.354};
double cf_aper_out[] = { 0.100, 0.185, 0.350, 0.650};
/* Exposure section: */
const double CF_MICROEXP = 0.004 ; 	/* < microexposure: times per scan count */
const int CF_NCNT = 100 ;	/* number of microexposures (or microsteps) per scan */
/* Magnification of the system: */
double cf_magnif = 20.0 ;
const double SCALE = 1.5 ;		/* Focal plane scale, in arcsec/microstep */
const double cf_bkgr[] = {30., 32., 40., 70.}; /* background counts measured in
channels per second */

const int CF_DEFRANGE = 12 ;	/* pixels from the center (see focus() docs)*/
/*------------- Other supplementary info: ------------------------*/
double sys_qe = 0.05 ;	/* System quantum efficiency */
double flux0 = 1E6 ; 	/* Photons per second per cm^2 from the mag=0 A0-star*/

/* Coordinates of simulated crossing the edges of the triangle aperture */
const double x10 = 0.25*CF_NCNT, x20 = 0.65*CF_NCNT ;
const double TAUWANDER = 0.5 ;	/* Characteristic time [sec] of image wandering */
/* Number of scanning microsteps per wandering characteristic time: */
const int steperwan = (int)(TAUWANDER / CF_MICROEXP + 0.5) ;

const double ERF05 = 0.476936 ;	/* Error function argument when it is equal to 0.5*/
/** 
	@brief	Image wandering simulation in one dimention
	@param	correlength	duration of correlation of deviations in calls
	@param	sigma_wander	RMS deviation of wandering
	@param	idum	random seed
	@return	random wandering deviation
	
	Function simulates the process of stellar image wandering in one dimention
	using the RMS deviation and the time span in units of function calls at
	which the deviations are correlated. This span equals to the characteristic
	time of wandering divided by the time interval with which the function is
	called.
	
	Initialize the function by the call with idum=0, use it, and then
	deinitialize by the call again with idum=0.
*/
double wanderdev(const int correlength, 
				const double sigma_wander,
				int *idum) {
static int n  = 0, len = 0 ;
static double fact = 0;
double dev ;
int i ;
static double * buf = NULL ;

	if ((*idum==0) || (len==0)) {
		if (buf==NULL) {
			/* Initialize: */
			len = correlength ;
			/** dispersion of \em len coadded normal-distributed 
				randoms with unity dispersion */
			fact = sqrt((double)len) ; 
			/* Allocate the buffer with for len randoms */
			buf = nr::dvector(0,len-1);
			/* fill in this buffer: */
			for (n=0; n<len; n++) {
				buf[n] = nr::gasdev(&(*idum)) * sigma_wander / fact ;
//				printf ("%f\n",buf[n]) ;
			}
		} else {
			/* deallocate: */
			nr::free_dvector(buf,0,len-1) ;
			buf=NULL; n=len=0;
			return 0 ;
		}
	} else {
		n++;
		n = n % len ;
		buf[n] = nr::gasdev(&(*idum)) * sigma_wander / fact ;
	}
	/* Compute the deviation: */
	dev = 0 ;
	for (i=0; i<n; i++)
		dev += buf[i] ;
	return dev ;
}

int main( int ac, char **av ) {

/* Receiving buffers for channels counts: */
count_t channelA[CF_NCNT] ;
count_t channelB[CF_NCNT] ;
count_t channelC[CF_NCNT] ;
count_t channelD[CF_NCNT] ;
count_t * channel[] = { (count_t*)(&channelA), (count_t*)(&channelB), 
					(count_t*)(&channelC), (count_t*)(&channelD) } ;

const double SCALE = 1.5 ;	/* 1.5 arcsec per microstep */
const double A = 2 ;	/* 2*tan(45degr) angles between the edges and radius-vector */
const double B = 0.2 ;	/* semidifference of tangent of these angles */

double area, flux, 
	magnitude = 0, 
	sigma_wander = 1, /* arcsec */
	seeing = 1, /* arcsec */ 
	sigma_scint = 0.1, 
	defocus = 0, 
	x1, 
	factor, mean, dev, dx, x,
	cx1, cx2 ;

int j, idum=(-13), test, meas, which;

if (ac<2) {
	printf(" MASS project: SCAN module test\n");
	printf(" Use: %s <star magnitude> <wander> <seeing> <scint> <defocus>\n",av[0]);
	printf(" where image wandering and seeing are in arcseconds \n");
	printf("   and scintillation is a relative flux variation (<1) \n");
	return -1 ;
}

magnitude = atof(av[1]) ;
if (ac>2) { x1 = atof(av[2]) ; if (x1>0) sigma_wander=x1; }
if (ac>3) { x1 = atof(av[3]) ; if (x1>0) seeing=x1; }
if (ac>4) { x1 = atof(av[4]) ; if (x1>0) sigma_scint=x1; }
if (ac>5) { x1 = atof(av[5]) ; if (x1>0) defocus=x1; }
if ((magnitude<-2)) {
	printf(" Error: magnitude is >=-2 \n");
	return -1 ;
}

/* Cycle of all the system work scenario repetitions : */

for (test=0; test<NTEST; test++) {
	scan_t scan[CF_NCHAN];
	for (int i=0; i<CF_NCHAN; i++) scan[i].set(CF_NCNT); 

	scan_t::loadconst(SCALE, A, B, x10, x20) ;

/*
	Test the set repetitive usage: 
	scan[0].set(CF_NCNT) ;
	scan[0].set(CF_NCNT+1) ;
	scan[0].set(CF_NCNT-2) ;
	scan[0].set(-1) ;
	if (nr::erget()) {
		cout << nr::ermessage() << "\n";
		nr::erreset() ;
	}
	scan[0].set(0) ;
	exit(0);
*/
	
	for (meas=0; meas<CF_NSCAN*2; meas++) {
		/* Direction sign : */
		which = 1-2*(meas%2) ;
		for (int i=0; i<CF_NCHAN; i++) {
			mean = 0 ;
			cx1 = cf_magnif*cf_aper_out[i]/2 ;
			cx2 = cf_magnif*cf_aper_inn[i]/2 ;
			area = (cx1*cx1-cx2*cx2) * M_PI;
			flux = flux0 * area * pow(10.0,-0.4*magnitude) * sys_qe ;
			for (j=0; j<CF_NCNT; j++) {
				dev = wanderdev(steperwan,sigma_wander,&idum);
				/* which edge is closer? */
				if (which>0)
					x = (double)j;
				else
					x = (double)(CF_NCNT-j-1) ;
				cx1 = x - (x10+dev) ;
				cx2 = (x20+dev) - x ;
				/* Coordinate difference with the edge: */
				dx = (fabs(cx1)<fabs(cx2)) ? cx1 : cx2 ;
				/* Stellar profile fraction which is seen: */
				factor = 0.5 + erf(dx/(seeing/SCALE) * ERF05) / 2.0 ;
				/* Scintillation: */
				dev = pow(10.0,0.4*nr::gasdev(&idum)*sigma_scint) ;
//				printf("%d %f %f %f %f\n",(int)x,dev,dx,factor,dev);
				factor *= dev;
//				printf("%d %f\n",(int)x,factor);
				channel[i][j] = (count_t)(0.5+
					nr::poidev((flux*factor + cf_bkgr[i])*CF_MICROEXP, &idum)) ;
				mean += (double)channel[i][j] ;
			}

			mean /= CF_NCNT ;
			scan[i].accum(channel[i], scan_t::direct(which)) ;
		}
		#ifdef VIEWSCAN
		for (int i=0; i<CF_NCHAN; i++) 
			for (j=0; j<CF_NCNT; j++) {
				if (which>0)
					x = (double)j;
				else
					x = (double)(CF_NCNT-j-1) ;
				printf("%d %d\n",(int)x,channel[i][j]);
			}
		#endif
	}
	if (nr::erget()) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
		return nr::erget() ;
	}

	/* Average scans: */
#ifndef VIEWSCAN
	printf("\n# Average scans");
	if (defocus!=0) printf(" and compute the focus:\n");
	else printf(" and compute the edge positions:\n");
#endif	
	for (int i=0; i<CF_NCHAN; i++) scan[i].merge() ;
#ifdef VIEWMSCAN
	printf("# Ok. Here they are:\n# x");

	for (int i=0; i<CF_NCHAN; i++) printf("\t%s",sc::apername(CF_NCHAN,i));	
	printf("\n");
	for (int j=0; j<CF_NCNT; j++) {
		printf("%d",j);
		for (int i=0; i<CF_NCHAN; i++) 
			printf("\t%d",(int)((scan[i].get(scan_t::MERG))[j]+0.5)) ;
		printf("\n");
	}
#endif
	for (int i=0; i<CF_NCHAN; i++) {
		scan[i].center (&cx1, &cx2, scan_t::MERG) ;
		if (defocus==0) {
			printf("# Chan %s: x1=%f, x2=%f i.e. ",sc::apername(CF_NCHAN,i),
				cx1, cx2);
				
			printf("sh_x=%5.2f\", sh_y=%5.2f\"\n", scan[i].xshift(), scan[i].yshift());
		} else {
			double sharp = 0 ;
			double cdefocus = scan[i].defocus (int((cx1+cx2)/2), scan_t::MERG,
				&sharp) ;

			printf("# Chan %s: sharp=%f, defocus=%f\n",
				sc::apername(CF_NCHAN,i), sharp, cdefocus);
		}
	}
	printf("# Release module arrays: automatic\n") ;

	if (nr::erget()) {
		fprintf(stderr,"%s\n",nr::ermessage().c_str()) ;
		return nr::erget() ;
	}
	
	printf("#===================== %d ===================\n",test);

} /* of test-cycle */

idum = 0;
dev = wanderdev(0,0,&idum);
printf("# Ok.\n");

return 0 ;
} /* end of main() */

#endif

/* An address and value to return in case of failure */
double bad = 99 ;

/* Definition of the static calibration constants: */
double scan_t::a = 2 ; /* 45-degree approximation */
double scan_t::b = 0 ;	/* symmetric diaphragm */
double scan_t::scale = 1 ;	/* no scaling so far */
double scan_t::x1c = 0 ;
double scan_t::x2c = 0 ;
int scan_t::defrange = 17 ;	/* model radius of maximal sensitivity to defocus in
							D-aperture*/
double scan_t::sharp0 = 1.029 ;		// model focal sharpness
double scan_t::focscale = 100 ;	// model scale

/*----------------------------- Constructor ------------------------------*/
scan_t::scan_t(int uncnt, double unonlin) { 
	ncnt=maxncnt=0 ; 
	x1c = x1 = uncnt*1/4;	/* optimal case: half a scan in FOV */
	x2c = x2 = uncnt*3/4;
	nonlin = (unonlin<0 ? 0 : unonlin) ;/* set in any case */
	set(uncnt,unonlin) ; 
#ifdef SCANDEBUG	
printf("scan_t Constructor with %d\n",uncnt) ;
#endif
}

/*---------------------------- set ----------------------------------*/

void scan_t::set(int uncnt, double unonlin) {

#ifdef SCANDEBUG
	cout << "SCAN: set(): Ncnt="<< uncnt << "\n";
#endif	

	if (uncnt<0) {
		nr::nrerror("SCAN: set() with negative length",nr::ERPAR);
		return ;
	}

	if ((uncnt<=maxncnt) && uncnt) {
		ncnt = uncnt ; 
	} else {
		/* Here we either have not enough space or have to deallocate it: */
		if (maxncnt) {
			nr::free_dvector(fscan,0,maxncnt-1) ;
			nr::free_dvector(bscan,0,maxncnt-1) ;
			nr::free_dvector(mscan,0,maxncnt-1) ;
			ncnt = maxncnt = 0 ;
		}
		if (uncnt) {
			maxncnt = ncnt = uncnt ;
			bscan = nr::dvector(0, ncnt-1) ;
			fscan = nr::dvector(0, ncnt-1) ;
			mscan = nr::dvector(0, ncnt-1) ;
		}
	}
		
	/* Initialize scans with zeroes: */
	nfscan = nbscan = 0 ;
	if (unonlin>=0) nonlin = unonlin ;	/* change only if valid, leave 
		otherwise */

	for (int i=0; i<ncnt; i++) bscan[i] = fscan[i] = mscan[i] = 0 ;

#ifdef SCANDEBUG	
	cout << "SCAN: set() became: ncnt="	<<ncnt<<" max="<<maxncnt<<"\n";
#endif
}


/*---------------------------- accum ---------------------------------*/

void scan_t::accum(const count_t * scan, direct which) {
	if (!scan) {
		nr::nrerror("SCAN: accum() with null scan",nr::ERNUL) ;
		return ;
	}
	switch (which) {
		case FORW : 
			nfscan++ ; 
			for (int i=0; i<ncnt; i++) 
				fscan[i] += scan[i] * (1.0 + nonlin * scan[i]) ;
			break ;
		case BACK : 
			nbscan++ ; 
			for (int i=0; i<ncnt; i++) 
				bscan[ncnt-1-i] += scan[i] * (1.0 + nonlin * scan[i]) ;
			break ;
		default:
			nr::nrerror("SCAN: accum() invalid which-specification",
				nr::ERPAR);
			return ;
	}

}

/*---------------------------- merge ---------------------------------*/

void scan_t::merge() {
	if (!ncnt) {
		nr::nrerror("SCAN: merge(): scan has zero length",nr::ERNUL) ;
		return ;
	}
	int tot = nfscan + nbscan ;
	if (tot>0) 
		for (int i=0; i<ncnt; i++) mscan[i] = fscan[i] + bscan[i] ;
	else
		nr::nrerror("SCAN: merge(): nothing to merge",nr::ERNOD) ;
}

/*---------------------------- get -----------------------------------*/

double * scan_t::get(direct which) const {
	if (!ncnt) {
		nr::nrerror("SCAN: get(): not initialized");
	} else {
		switch (which) {
			case MERG: return mscan ;
			case FORW: return fscan ; 
			case BACK: return bscan ; 
			default:
				nr::nrerror("SCAN: get() unknown which-specification",
					nr::ERPAR);
		}
	}
	return &bad ;	/* if not initialized or bad "which" */
} 

/*---------------------------- center --------------------------------*/

void scan_t::center(double * ux1, double * ux2, direct which, 
	bool fixcenter) {

	const int K_BKGR = 5 ;	/* least value corresponds to K_BKGR=1 */
	const int K_PLAT0 = 10 ;	/* crude plateau estimate: K_PLAT0-th largest */
	const double UP_LIN = 0.75 ;	/* upper boundary of the linear range of slope */
//	const double LO_LIN = 0.25 ;	/* lower boundary of the linear range of slope */
	const double SNR = 10 ;	/* Minimal allowed ratio of signal to noise */

	if (ncnt<K_BKGR+K_PLAT0+2) {
		if (!ncnt)
			nr::nrerror("SCAN: center(): scan has zero length (uninitialized?)") ;
		else
			nr::nrerror("SCAN: center(): scan is too short") ;
	}

	*ux1 = x1 = *ux2 = x2 = 0 ;

	double * scan = get(which) ; 
	if (nr::erget()) return ;

	/* Sort the copied scan: */
	double * sorscan = nr::dvector(0,ncnt-1) ;
	for (int i=0; i<ncnt; i++) sorscan[i] = scan[i] ;
	nr::sort(ncnt,sorscan-1) ; 

	if (nr::erget()) {
		nr::free_dvector(sorscan,0,ncnt-1);
		return ;
	}

	/* Find the background level as K_BKGR-th smallest count: */
	double bkgr = sorscan[K_BKGR-1] ;
	/* and plateau level: */
	double plat = sorscan[ncnt-1-K_PLAT0] ;
	
	/* True plateau level is a median of counts which are above UP_LIN fraction
	of the intensity range: */
	double level = bkgr + UP_LIN*(plat-bkgr) ;
	plat = 0 ;
	for (int i=K_BKGR-1; i<ncnt-K_PLAT0; i++) 
		if (sorscan[i]>level) {
			int j = i+(ncnt-K_PLAT0-i)/2 ;
			plat = sorscan[j] ;
			break ;
		}

	nr::free_dvector(sorscan,0,ncnt-1);

	if (plat>bkgr+SNR*sqrt(bkgr)) {
		
		/* 50% level: */
		level = (plat + bkgr) / 2 ;
		/* search for the level crossing while the star appears: */
		for (int i=0; i<ncnt-1; i++) {
			if ((scan[i]<=level) && (scan[i+1]>level)) {
				x1 = (double)i + (level-scan[i])/(scan[i+1]-scan[i]) ;
				break ;
			}
		}
		/* search for the level crossing where the star disappears: */
		for (int i=ncnt-2; i>=0; i--) {
			if ((scan[i]>=level) && (scan[i+1]<level)) {
				x2 = (double)i + (scan[i]-level)/(scan[i]-scan[i+1]) ;
				break ;
			}
		}
		if ((x1*x2==0) && (x1!=x2)) {
			/* one of coordinates not determined */
			nr::nrerror("SCAN: no complete coverage of the star?",
				nr::ERGEO);
		}
		if (ux1 && ux2) {
			*ux1 = x1; 
			*ux2 = x2 ;
		}
	} else {
		nr::nrerror("SCAN: centering with no star?", nr::ERNOD) ;
	}
	
	if (fixcenter && !nr::erget()) {
		x1c = x1 ;
		x2c = x2 ;
	}

//	printf ("sc[%i]=%f plat=%f\n",i,sorscan[i],plat);
}

/*------------------------ loadconst -------------------------------*/

void scan_t::loadconst(double uscale, double ua, double ub, 
	double ux1c, double ux2c,
	int udefrange, double usharp0, double ufocscale) {
	if (uscale>0) scale = uscale ;
	if (ua>0) a = ua ;
	if (ub!=-99) b = ub ;
	if (ux1c>0) x1c = ux1c ;
	if (ux2c>0) x2c = ux2c ;
	if (udefrange>0) defrange = udefrange ;
	if (usharp0>0) sharp0 = usharp0 ;
	if (ufocscale!=0) focscale = ufocscale ;
}

/*---------------------------- defocus ---------------------------------*/

double scan_t::defocus (int xcenter, direct which, 
	double * sharp) const {

	const double * scan = get(which) ; 
	if (nr::erget()) return bad;

	if (defrange<1) {
		nr::nrerror("SCAN: defocus() demands the positive sharpness range",
			nr::ERPAR) ;
		return bad;
	}
	if (xcenter<=0 || xcenter>=ncnt)  {
		nr::nrerror("SCAN: defocus() with invalid center position",
			nr::ERPAR) ;
		return bad;
	}
	
	int imin = ((xcenter>defrange) ? (xcenter - defrange) : 0) ;
	int imax = (((xcenter+defrange) >= ncnt) ? (ncnt-1) : (xcenter + defrange)) ;
	int n = (imax-imin+1) ;
	// average count:
	double avg = 0;
	for (int i=imin ; i<=imax; i++) avg += scan[i] ;
	if (avg<=0) return bad ;
	avg /= n ;
	// compute the sharpness: ratio of average centered squared count to squared
	//	average count:
	double sharpness = 0 ;
	for (int i=imin ; i<=imax; i++) {
		double d1 = scan[i]/avg ;
		sharpness += d1*d1 ;
	}
	sharpness /= n ;
	if (sharp) *sharp = sharpness ;
	double defocus = focscale * (sharpness - sharp0)/sharp0 ;
	return defocus ;
}	
