/*--------------------------------------------------------------------------*/
/** @file scan.hxx
                  MASS project: TURBINA module file                       
                       header file for scan.cpp                             
                                                                          
 The module SCAN contains a set of utilities which handle the scans obtained in
 any servo-regime of MASS device which involves the \em scanning: the counting
 synchronous with moving of the wheel of apertures. Here  "apertures" are the
 focal plane diaphragms or lens settings, not the entrance  apertures of the
 system which are played around in SCIND and WEIF modules.  Servo-regimes are
 the focusing and centering of the MASS device apertures.

 @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)               

 @version 1.5: C++ version of scan.c                                        */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _SCAN_H
#define _SCAN_H

#include <stdio.h>
#include "iocount.h"	// for count_t only
     
/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
 \b Usage:

 For each device channel where the scanning has to be done the object of the
 class scan_t has to be declared. The constructor by default which is involved
 in declaration of the array of scan_t allocates no space. So, before
 accumulation of scans, the function set() is called for scan of each channel.

 Each time the new scan is received, the scan accumulation accum() is called
 with the proper value of the flag scan_t::direct of scanning  direction. After
 finishing the scanning, the utility merge() should be called to obtain the
 resulting merged scan where forward and backward scans are coadded. The
 utility get() returns the pointer  to one of the scans (forward, backward or
 merged). The "scientific" functions center() and defocus() extract from the
 selected accumulated scan (again 3 kinds) the information on the position of
 the star in the centering triangle aperture and on the focus displacement,
 respectively. The results of center() are also stored internally and used to
 compute the calibrated shifts of the star xshift() and yshift(). For latter
 two functions and for defocus(), the calibration constants should be loaded 
 with loadconst().

 Before the new accumulation of scan, the function set() has to be called. If
 nothing has changed (dimention of scan or non-linearity), the cleaning of
 the storages is only done.
 
 The scans are made of individual counts from the MASS device channels and
 represent the \em count_t-type arrays. The latter type definition is 
 borrowed from the module SCIND.
*/
class scan_t {

public: 
/** \em which -flag to access forward scan */
enum direct {BACK=-1,MERG=0,FORW=+1} ; 

/**
	@brief	Initialize the scan arrays before accumulation
	@param	ncnt	Number of counts in a scan (forward or backward)
	@param	nonlin	non-linearity parameter for the detector (i.e. the deadtime
		of detector divided by the microexposure time, e.g.: 
		\em nonlin = 24ns / 4ms = 6.0e-6)
	
	The object contains the arrays \em fscan, \em bscan where the forward and
	backward scans are accumulated, and \em	mscan to coadd them afterwards in a
	merged scan. These vectors have the logical length \em ncnt. If their 
	physical length is smaller, they are reallocated with a new length \em ncnt.
	
	If \em ncnt is zero, the storages are deallocated and no allocation is done
	(the call from destructor is done like this).

	This function should be called, obviously, before the first scanning pass is
	finished. Default non-linearity is made invalid to leave it unchanged.
	
	@see accum()
*/
void set(int ncnt = 0, double nonlin = -1) ;

/**
	@brief Constructor: set the private fields zero and call set().
	
	The default constructor is realized with default parameters passed to set().
	Thus, for array of scan_t, it is obligatory to use set() after declaration;
	for a single scan_t instance, the constructor with proper length and
	non-linearity can be called immediately before accum(). If default (invalid)
	non-linearity is supplied, it is set zero in constructor.
*/
scan_t(int ncnt = 0, double nonlin = -1) ;

/**
	@brief	Destructor: deallocate the memory reserved for storages
	
	This function releases the memory which is allocated for forward, back and
	merged scans by set() with zero parameter. 
*/
~scan_t() { 
#ifdef SCANDEBUG
printf("scan_t Destructor \n") ;
#endif
set(0) ; }

/**
	@brief	Add the current scan to the respective scan storage
	@param	scan	pointer to the \em count_t -type scan array
	@param	which 	Direction flag: FORW: forward, BACK: backward 
	
	Function adds the given array to the forward and backward scans, latter
	with inversed sequence of counts. Thus, forward and backward scans must be
	identical (see get()) if the measurement consitions are stable and no noise
	is taken into account. The correction for non-linearity is done with the
	non-linearity constant loaded in set().

*/
void accum(const count_t * scan, direct which) ;

/**
	@brief	Merge back- and forward scans
	
	Function coadds the forward and backward scans into the merged scan.

	@see accum()
*/
void merge() ;

/**
	@brief Return the pointer to the needed scan
	@param	which BACK for backward scan , FORW for forward scan,
		MERG for merged scan
	@return pointer to the scan array (not to be deleted!) 

*/
double * get(direct which = MERG) const ;

/**
	@brief	Locate the coordinates where the star crosses the edges of a
	centering triangle aperture
	@param	x1	Resulting coordinate of the crossing point of the \em left 
	triangle edge [microstep, i.e. pixel of scan]
	@param	x2	Resulting coordinate of the crossing point of the \em right 
	triangle edge [microstep]
	@param	which BACK for backward scan , FORW for forward scan,
		MERG for merged scan
	@param	fixcenter	Device adjustment mode: if true, then the resulting
		\em x1 and \em x2 are treated as for the non-shifted star and saved
		in scan_t::x1c and scan_t::x2c. See loadcenter().
	
	Function searches for the positions on the scan where the scan counts cross 
	the 50\%-level of the light variation range: the points where the star 
	appears from and disappears behind the edges of the triangle aperture.
	"Appearance" stays for the result \em x1 and has a smaller index in (forward
	and merged) scan than \em x2 (see xshift() for the star shift calibration
	issue based on \em x1 and \em x2).
	
	The method is based on a linear interpolation of points on the middles of
	slopes of the curve. The points are taken which surround the middle of
	slopes. "Middle" is considered as a mean of background level (some \em k-th
	smallest count in a scan) and the plateau level (median of counts which have
	intensity more than 75\% of the highest one.

	Under average conditions (seeing and wandering of images = 1'',
	scintillations of 10\%, see testing main() of the module) the
	precision of centering is about 0.1 microsteps.

	\note
	If one of the crossing points is not covered by the aperture edge while
	scanning, the respective coordinate is returned as zero.
	
	\attention
	If the error is set in the system on enter (see nrerror()), the zero 
	results are	returned in \em x1 and \em x2.
	
*/
void center(double * x1, double * x2, direct which = MERG, 
	bool fixcenter = false) ;

/**
	@brief	Assign the static calibration parameters for calculation of the star
		shifts and image defocussing
	@param	scale	Focal scale near the center [arcsec/microstep]
	@param	a	Sum of tangent of angles between the edges of the triangle 
		diaphragm and the direction to the axis of the aperture wheel
	@param	b	semidifference of tangent of angles between the edges of the 
		triangle diaphragm and the direction to the axis of the aperture wheel
	@param	x1c	scan coordinate of appearance of the non-shifted star
	@param	x2c	scan coordinate of disappearance of the non-shifted star
	@param	defrange	Radius of sharpness calculation around the scan
		center [microsteps]
	@param	sharp0	Sharpness of the scan which corresponds to the perfectly
		focussed system
	@param	focscale	focus position change per unity of the relative
		sharpness change (i.e. \em (sh-sh0)/sh0) taken near the focus position

	The function assignes the parameters to the class variables to be used in
	subsequent calculations of the star displacement with xshift() and yshift().
	Only the non-defaulted parameters are assigned. Defaults are selected to
	have no sense.
	
	\note
	These calibration parameters serve for all instances of the type scan_t
	because they are static. E.g. for many channels of the device, the single
	call to scan_t::loadconst() has to be done. For the focussing issue, the
	parameters (last three) have a sense for only one (say, D) aperture.
	
	On startup, the following calibration parameters are set:
	\arg \c scale = 1
	\arg \c A = 2
	\arg \c B = 0
	\arg \c x1c = x2c = 0
	\arg \c defrange = 17
	\arg \c sharp0 = 1.029
	\arg \c focscale = 100.
*/
static void loadconst(double scale = 0, double a = 0, double b = -99, 
	double x1c = 0, double x2c = 0, int defrange = 0, double sharp0 = 0,
	double focscale = 0) ;

/**
	@brief	Accept the (modified) edge-crossing coordinates for on-axis star

	This is a shortened version of loadconst().
*/
static void loadcenter(double x1c, double x2c) {
	loadconst(0,0,-99,x1c,x2c) ;
}

/**
	@brief	Assign the new parameters for the defocussing determination

	This is a shortened version of loadconst().
*/
static void loadfocus(int defrange, double sharp0 = 0, 
	double focscale = 0) {
	loadconst(0,0,-99,0,0,defrange,sharp0,focscale) ;
}
	
/**
	@brief	Return the shift of the star relative to the center of the field 
	of view in scanning direction
	
	The function uses the \em x1 and \em x2 values computed and stored in
	private members by center() to derive the shift of the star from the center
	of the field of view.
	\image html scan.jpg "Scheme of MASS scanning for centering the star"
	\image latex scan.eps "Scheme of MASS scanning for centering the star" height=7cm
	
	As one can see from the picture, the difference of \em x1 and \em x2 is
	easily converted into the Y-shift, and the Sum of them - into the X-shift
	via the calibration parameters \em A and \em B (see loadconst()):
	\f[
		\Delta X = scale \cdot \frac{\Delta x1+\Delta x2}{2} - \Delta Y \cdot B
	\f]
	Here \em Dx1 and \em Dx2 are the differences of the current edge-crossing
	coordinates \em x1, \em x2 with those obtained for on-axis star.

*/
double xshift() const { return scale * ((x1+x2)/2 - (x1c+x2c)/2) - yshift()*b ; }

/**
	@brief	Return the shift of the star relative to the center of the field 
	of view in radial direction

	The radial shift is computed as:
	\f[
		\Delta Y = scale \cdot \frac{\Delta x1-\Delta x2}{A}
	\f]
	
	@see xshift()
*/
double yshift() const { return ((x2-x1) - (x2c-x1c)) / a ; }

/**
	@brief	Compute the sharpness of the scan and convert it into defocus
	@param	xcenter	scan center to compute the sharpness around
	@param	which BACK for backward scan , FORW for forward scan,
		MERG for merged scan
	@param	sharp	resulting sharpness placeholder; not returned by default
	@return defocus measure in units of \em focscale in
		loadconst() (normally, [mm])
	
	Function computes the \em sharpness of a scan which depends significantly
	on the defocussing of the system. 

	In current implementation, the sharpness \em S is the ratio of the mean
	square of the signal to the squared mean of the signal:
	\f[
		S = \frac{\frac{1}{2R+1}\sum\limits_{i=Xc-R}^{Xc+R}
		scan[i]^2}{\overline{scan}^2}
		\mbox{~~~,~where~~~~}
		\overline {scan} = \frac{1}{2R+1}\sum\limits_{i=Xc-R}^{Xc+R} scan[i]
	\f]
	I.e., the signal is taken from scan at the positions closer than \a R=defrange 
	to the position \a Xc=xcenter.
	
	Then, the relative sharpness deviation is computed and converted into the
	defocus measure:
	\f[
		Defocus = focscale \cdot \frac{S-S_0}{S_0}
	\f]
	where \em So is the focal sharpness \a sharp0 (see loadfocus()).
*/
double defocus (int xcenter, direct which = MERG, double * sharp = 0) const ;

private:

	int ncnt ;	/**< current scan length in counts */
	int maxncnt ;	/**< maximal length of scan (memory storage size) */

static double scale ;	/**< calibration constant 1: focal plane scale near the 
		center  [arcsec/microstep] (microstep is a "pixel" of scan) */
static double a ;	/**< calibration constant 2: Sum of tangent of angles between 
		the	edges of the triangle diaphragm and the direction to the axis of 
		the aperture wheel */
static double b ;	/**< calibration constant 3: semi-difference of tangents of 
		angles between the edges of the triangle diaphragm and the direction to 
		the axis of the aperture wheel */
static double x1c ;	/**< centering parameter: star appearance coordinate on scan
		when it is not shifted */
static double x2c ;	/**< centering parameter: star disappearance coordinate on 
		scan when it is not shifted */
static int defrange ;	/**< radius of the sharpness determination around the
	scan center */
static double sharp0 ;	/**< sharpness of a scan when the system is focussed */
static double focscale ; /**< rate of the focus position change per unity of
relative sharpness change, usually in [mm/1] */
	double x1 ;	/**< current (last obtained) position of appearance of star */
	double x2 ;	/**< current position of disappearance of star */
	
/*---------------------------------------------------------------------*/
/*--- data storages for accumulation of scans -----------*/
/** Accumulated forward scan[ncnt] */
	double * fscan ;
/** Accumulated backward scan[ncnt] */
	double * bscan ;
/** Merged scan */
	double * mscan ;

	int nfscan ;	/**< Current amount of accumulated forward scans */
	int nbscan ;	/**< Current amount of accumulated backward scans */
	double nonlin ;	/**< non-linearity parameter, in [deadtime/exposure] */

} ; // of class scan_t

#endif
