/*--------------------------------------------------------------------------*/
/** @file scind.hxx
                  MASS project: TURBINA module file                       
                       header file for scind.cpp                             
                                                                          
 The module SCIND implements the calculation of stellar  scintillation indices
 observed in a number of apertures (channels). The correction of observed
 normalized dispersion of the signal involves the \em  background sky level,
 non-linearity of the detector and the \em  non-Poisson factor (close to 1)
 which converts the mean of the detector signal into dispersion (the unity for
 an ideal detector).

 As a result, the \em normal scintillation index is produced which is free 
 from photonic statistics influences and detector imperfections. The 
 differential indices related to the signals \em covariance of different
 channels are also  computed. The notions of "aperture" and "channel" are 
 identical here. The computed indices can be accessed, read or written to the
 disk with a number of additional utilities.

 This code originates from the program \b select.c written by V.Kornilov for
 DASS project. All the functions and constants which deal with scintillation 
 indices are available in the namespace "sc".

 \b Usage:

 For the sake of performance and simplicity, the work-arrays which are used for
 calculations of indices are allocated only once, by the initiation utility
 sc::init(). Deallocation is devoted to done(). After init(), all the necessary
 computations are done by sc::calcmom() and sc::calcidx() a certain number of
 times, during which the (instant) indices or raw count statistic moments may
 be saved to disk by sc::write(). Note that sc::write() for writing the moments
 should be called after sc::calcmom() but before sc::calcidx(). For indices,
 sc::write() may be called \em after sc::calcidx() ONLY if they were computed
 successfully, i.e. if nr::erget() returns 0, not nr::ERNOD. If the value of 
 some parameter of sc::init() has changed (or if even nothing has changed), 
 this function may be called again, one does not need to call done() before. 
 The additional space will be reallocated if needed.

 The index storages filled by sc::calcidx() may be accessed with sc::getidx().
 The sequential number of a certain index in accessed array may be obtained
 with sc::ind_seqnum() from the name of an aperture or of the combination of
 two apertures. Vice versa, the character name of an index which is accessed as
 \em i-th in index storage is returned by sc::apername().

 After some accumulation of indices (in local index storages), the indices can
 be averaged by sc::average(). These average indices can already be used for
 calculations of atmosphere models and parameters (see atm::calcint() and
 atm::calcn2() in the module ATMOS). Also, they may be saved to disk with
 sc::writeavg(). 
 
 The counter of accumulated indices must be reset by sc::again() before
 beginning of the next accumulation time and then the cycle of instantaneous
 index computations with sc::calcidx() can be restarted. Alternative to
 sc::again() is sc::init().

 After finishing the job, the memory should be cleaned with sc::done().

 \note

 In principle, index averaging may be done in any moment providing thus some
 on-fly smoothing of data in a time-sliding window. It is possible since the
 index storages work as some circular buffers, i.e. the rows which are not
 covered in a current accumulation time still contain information from the
 previous accumulation time. Empty rows of index storages won't be used, since
 the internal storage which keeps the measurement mode is reset to
 "mode=unknown" each time sc::init() is called.

 \attention
 
 Internally in the module, the counts are retained related to the original
 micro-exposure time, i.e. reflect the real number of pulses accumulated during
 the time piece. So are the counts saved in "raw moments" file
 (sc::write()). Meanwhile, the results of sc::getmean(), sc::getsig(),
 sc::getavgflux() are  already converted into the units of
 micro-exposure (normally [ms] as assumed in documentation of functions in sc)
 which was supplied to sc::calcidx() and saved internally. Also, the fluxes
 written by sc::writeavg() are also scaled to the micro-exposure.

 @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)               

 @version 1.5: Translation in C++                                    

	1.51: a few sysnopsis changes, statflux err(flux) bug corrected

	2.0: count moments and index calculation separated; filtered averaging
	for star escapes treatment; some bad-call checkings corrected.

	2.1: Count second moments normalized, new MOMFMT hense. Output is
	tested: the same as for v1.51. Headers repetitive output blocked.
*/
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _SCIND_H
#define _SCIND_H

#include <stdio.h>
#include <math.h>

/** Type of PM tube counts  */
#include "iocount.h"	// count_t is defined there

/**
	@brief
	Declarations of all functions and constants to handle scintillation indices
*/
namespace sc {

/** Enumeration of entities to access or write on the disk, to use in
get...() and write...() functions. */
enum scwhat {
	DSI, 	/**< Aperture indices writing (with DESI) or access */
	DESI, 	/**< Differential exposure indices access */
	AFLUX,	/**< Average flux in channels access or writing */
	AFLUXP,	/**< Average flux with non-Poisson parameter in channels writing */
	MOM  	/**< Count moments writing */
} ;

/** Formula to get a Number of entrance apertures and their non-redundant 
combinations having \em a apertures */
inline int napcomb(int a) { return ((a)*(a)+(a))/2 ; }

/** Use true logarithmic calculations of scint. indices instead of relative
differential calculations */
//#define LOGIDX

/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
	@brief	Return the name for a given index of entrance aperture or 
		combination of apertures
	@param	nchan	number of apertures
	@param  index	sequential number of aperture
	@return	pointer to name if Ok, or "" if invalid index
	    
	Name of an aperture corresponds to the given index according to
	convention: For index=0..nchan-1: name='A','B',... (for normal indices
	of scintillation in a single aperture); for index=nchan: name='AB'
	(first combination of apertures), nchan+1: 'AC',...,'BC','BD' and so
	on.

	This name is copied in the static character string which contains
	3 characters (two for the combination and one for trailing zero) and
	returns the pointer to this string. Thus, if more than one call to this
	function is done in one statement, the result is unpredictable.

*/
const char * apername(int nchan, int index) ;
//throw(std::out_of_range) 

/**
	@brief	Return the sequential number of index in storage by its name (ID)
	@param	nchan	number of channels
	@param	name	ID of an aperture or the combination of apertures (e.g. "B" or 
		"AD")
	@return	sequential number in range [0..napcomb()-1] or -1 on error (error is
	set with nr::nrerror() then)

	This function is needed to access certain indices with getidx(), getavgidx() and
	geterridx() to compute their last parameter by the index "name".
	
	@see apername()
*/
int ind_seqnum(int nchan, const char * name) ;
	
/**	
	@brief	Initialize the structures to put the scintillation indices
	@param	maxndata	Maximal expected number of counts in base time 
	series (one simultaneous measurement of \em nchan channels)
	@param	maxnmeas	Maximal expected number of measurements 
	(accum.time/base.time) to keep in the storage
	@param	nchan	Number of channels (entrance apertures)
	
	The module contains the global arrays which keep the normal, differential
	aperture and differential exposure indices computed for all apertures.
	These are the matrices in which the rows contain the simultaneously
	obtained indices for one particular base time.

	For the sake of a simple code structure, some intermediate data processing
	is done in advance to calculation of indices itself. For this and for
	acceleration of logarithmic indices derivation, some work  arrays are also
	allocated.

	The dimensions of allocated arrays are determined by the numbers \em
	maxndata, \em maxnmeas and \em nchan which are copied in global variables
	which are private to the module. Another private global variable counts the
	performed index calculations and initialized to zero here. The number of
	counts in any subsequent \em base-time series should not exceed the value
	of the parameter \em maxndata; the same for the measurements number (base
	times per accumulation time) which should not exceed the parameter \em
	maxnmeas.

	This function should be called, obviously, before the first calculation of
	indices. 
	
	\attention In addition to init() call, ALL the members of (private to
	module) channel parameter structures should be initialized by \em nchan
	calls to chan_init() made after the call to init().

	If the parameters specify the storage dimensions which are less than that
	which were previously allocated, the storages are left unchanged.

	\note
	This function must be kept in strict sync with calcidx().

	@see chan_init() calcidx() done()
*/
void init(int maxndata, int maxnmeas, int nchan) ;
//throw(std::out_of_range,std::bad_alloc) ;

/**
	@brief 	Initialize channel data in static channel array element
	@param	chan	sequential number of the channel (0,1..maxnchan-1) (see
		init())
	@param	bkgr	background level in [counts per units of \em microexp
		parameter in calcidx()]
	@param	deadtime	non-linearity parameter (the dead-time of PMT) in the same
		units as the \em microexp parameter in calcidx(), e.g. for [sec] it is 
		about (20-30)*1e-9
	@param	nonpois	non-poisson factor (around unity)
	@param	data	pointer to the buffer - receiver of channel PMT counts
	
	Function initializes the \em chan-th channel with the given parameters.
	\attention
	This channel data are expected to be allocated by the call to init().

	The assignment is made only if the respective parameter has a sense:
	non-negative for double-type parameters and non-null for the pointer. Thus,
	a single all-initializing call may be split into a sequence of calls which
	initialize (or update) some particular members  (e.g., changing the
	non-linearity and background level in the channel) leaving the rest
	non-sense or default.

	\note
	No memory allocation is done here for data[] array, it is only pointed to
	the buffer already allocated for receiving the channel counts.
	
	@see init()
*/
void chan_init(int chan, double bkgr, double deadtime = -1, double nonpois = -1,
	count_t * data = 0) ;
//throw (std::out_of_range, std::domain_error) ;

/**
	@brief	Compute the statistic moments of counts: means and normalized 
		dispersions and auto/cross-covariances
	@param	ndata	Actual number of counts in the channel buffers (equal for
		all channels)
	@param	ncorr	array of numbers of corrected counts in channels 
		(replaced with mean if missing) 

	This is the first step in scintillation indices calculations. It computes
	all the needed count statistics and stores them in the module storages.
	From that storages, the second procedure sc::calcidx() takes them to convert
	into the indices themselves.
	
	Calculations of statistics are done with NO correction for nonlinearity 
	which is applied in	sc::calcidx() before calculation of indices. Meanwhile,
	the dispersions and auto-covariances are corrected for the data-loss in a
	line (provided for each channel as a \a ncorr[] array elements). Latter
	correction assumes that the missing counts in a channel buffer were
	replaced with the mean count in advance to the call of this utility.

	Statistics include:
	- \em mean count for all channels
	- normalized \em dispersion and \em auto-covariances with lags 1 and 2 for all channels
	- normalized counts \em cross-covariances between channels with lags 0 and 1.
		
	The parameter \em ndata must not exceed the value \em maxndata with which
	the init() was called.
	
	The implementation of this function depends on the \em LOGIDX compilation
	option: if it's set, then the "true" calculations of statistics of 
	\em logarithms of counts is implemented as demanded by the exact
	definition of a scintillation index; otherwise - the relative differential 
	approximation is used. Here below we refer to these alternative modes of 
	implementation as "\em if \em LOGIDX".
	
	Calculations of statistic moments go through the following steps:
	- compute the ``raw'' means of counts \f$\overline x\f$ ;
	- if \em LOGIDX, compute logarithms of all individual counts and their 
	means;
	- calculate the dispersions and auto-covariances (lag=1,2) of counts 
	(or - if \em LOGIDX - of their logarithms) \em x :
	\f[
		\sigma^2_k(x) = \frac{1}{N-k-N_{corr}}\sum\limits_{i=1}^{N-k}(x_i-\overline
		x)(x_{i+k}-\overline x),~~k=0, 1, 2
	\f]
	where \f$ N_{corr}\f$ is a number of pixels in the channel which were 
	replaced with mean \f$\overline x\f$ due to the loss in the line (see 
	below);
	- calculate the cross-covariances of counts \em x and \em y of all
	combinations of channels with lags 0 and 1:
	\f[
		\rho_0(xy) = \frac{1}{N}\sum\limits_{i=1}^{N}(x_i-\overline	x)
		(y_{i}-\overline y)~,~~~~
		\rho_1(xy) = \frac{1}{2(N-1)}\sum\limits_{i=1}^{N-1}(x_i-\overline x)
		(y_{i+1}-\overline y)+(x_{i+1}-\overline x)(y_i-\overline y),~~~
	\f]
	Note that the second formula is symmetric.
	- if \em not \em LOGIDX, \em normalize the dispersions and auto-covariances:
	divide them by the squared mean count and by the product of mean counts 
	in two channels, respectively ;

	The pre-allocated arrays \em mean, \em sig and \em rho1, \em rho2 are used
	to  store the results of statistical calculations. They contain,
	respectively: mean counts, normalized dispersions and cross-covariances of
	counts with a zero lag, normalized auto- and cross-covariances with a lag 1
	and normalized auto-covariances with a lag 2. Note, that, the \em i-th
	element of the dispersion/covariance arrays addresses the same channel or
	combination of channels as \em i-th element of the scintillation index
	array, but cross-covariances with lag 2 are not computed. In other
	respects, the global arrays \em rho1 and \em rho2 are used and accessible
	similarly to \em sig.

*/
void calcmom(int ndata, int * ncorr) ;
//throw (std::domain_error, std::range_error, std::invalid_argument, std::out_of_range, sc::dataproblem) ;

/**
	@brief	Compute the normal and differential (aperture and exposure) 
		scintillation indices
	@param	microexp	Exposure time of each count (\em micro-exposure) to scale
		the non-linearity parameter and background in calculations (see
		chan_init())
	@param	isgen	Generalized mode flag

	This is the main utility in the module SCIND. It uses the statistic moments
	of counts in channels (means and normalized dispersions and
	auto/cross-covariances) computed by sc::calcmom() and stored in the module
	storages.

	This utility computes a set of all \em instantaneous (i.e. determined using
	the count statistics over the \em base \em time) scintillation
	indices (SI) which can be derived for a given number of channels. Thus, 
	both \em normal and \em differential indices (DSI) are calculated.
		
	Calculations of scintillation indices go through the following steps:

	- correct dispersions, covariances and mean counts for non-linearity of
	PMT \em L=deadtime/microexp :
	\f[
		\overline x_{corr} = \overline x (1 + L \overline x)~~,~~~~~
		\sigma^2_{corr}(x) = \frac{\sigma^2(x)}{(1 - 2 L \overline x)^2}~~,~~~~~
		\rho_{corr}(xy) = \frac{\rho(xy)}{(1 - 2 L \overline x)
		(1 - 2 L \overline y)} ;
	\f]
	Note, that the last two formulae are applicable to both usual and normalized
	second moments (latters are in play, actually);
	- compute scintillation indices free from photon noise (using non-Poisson
	factors \em P for conversion of means into photon noise dispersion) and
	influence of background \em B=bkgr*microexp, corrected to zero-exposure:
	normal and cross-channel (differential):
	\f[
		s^2(x) = \left(1.5 (\sigma_0^2(x) - \frac{P}{\overline x}) - 
		0.5 \sigma_1^2\right) \frac{\overline x^2}{(\overline x-B_x)^2},~~~~
				
		s^2(xy) = s^2(x) + s^2(y) - 2 \left(1.5 \rho_0(xy) - 0.5 \rho_1(xy)\right)
		\frac{\overline x \overline y}{(\overline x-B_x)(\overline y-B_y)}
	\f]
	Here already the non-linearity corrected means, and normalized dispersions 
	and covariances	are used ;
	- compute DESI for all channels, using auto-covariances with lags 0 
	(i.e. dispersion), 1 and 2.
	\f[
		s^2_{de}(x) = \frac{2}{9}\left( 3 s^2(x) +
		\left(\sigma^2_2-4\sigma^2_1\right)
		\frac{\overline x^2}{(\overline x - B_x)^2}\right)
	\f]
	- check that the normal and differential aperture indices belong to the
	interval sc::MINSCIND - sc::MAXSCIND. If all indices are in this range,
	increment the counter of data in storages (that which is returned by
	sc::getcurmeas()). If out of range, the error nr::ERNOD is set (returned by
	nr::erget()). It must be reset by nr::erreset() to continue measurements.

	See ``Multi-Aperture Scintillation Sensor. Final design report'', Appx. C
	(p.87).
			
	The order of indices in which they are assigned to the elements of an
	output array (written to the file or exported by getidx()) is exactly the
	same as the order of \em weights in the weight matrix which is computed in
	the module WEIF: first normal indices, then the covariance of the first
	channel data with the second, with the third etc..., then the covariances
	of the second channel with	third etc. See sc::apername() and
	sc::ind_seqnum().

	\b Performance (including time-consumption by sc::calcmom()):
	
	The test consists of calculation of scintillation indices for 4 channels
	including the DESI calculations for all channels. It takes:

	- In relative approximation of indices (normalized dispersion or 
		covariance):  \em 1.7ms at P-III 667 MHz,
	- In true logarithmic form (\em LOGIDX definition set): 
		\em 3ms at P-III 667 MHz,

	The list of features:
	
	- the resulting indices in \em scind are \b already \b corrected \b to 
	\b zero-exposure;
	- The code assumes the background level and non-linearity parameter to be
	scaled to the micro-exposure time used for the count series. E.g. if the
	count (micro)exposure is 1ms, then the background level of 30 cnt/s should 
	be set as \em bkrg=30*0.001=0.03; deadtime of PM of 20ns - as \em
	nonlin=20ns/0.001s=2e-5.
	- The pre-allocated arrays \em mean, \em cmean, \em sig are used to 
	read/store the results of statistical calculations. They contain,
	respectively: mean counts (from sc::calcmom()) and means corrected for 
	non-linearity, normalized dispersions and covariances of counts corrected for
	non-linearity. Note, that, the \em i-th element of the
	dispersion/covariance array addresses the same channel or combination of
	channels as \em i-th element of the scintillation index array. The global
	arrays \em r1 and \em r2 are also  used and are accessible similarly to \em
	sig.

	The results of index calculations are put in the current measurement's row
	(see getcurmeas()) of the local index storing matrices.

	The calculations are made irrespective of \em isgen parameter of this
	function. This parameter is only saved in additional local storage to be
	able to disentangle the normal and generalized data while averaging the
	indices.

	Before exiting and upon the success of calculations, the counter \em
	curmeas of the performed measurements is incremented.  If the value of \em
	curmeas equals or exceeds the declared in init() maximal number \em
	maxnmeas (the length of allocated storages for indices), then the error 
	sc::EROFL is returned and no calculation is done.

    \attention The utility corrects the count statistics for detector
	nonlinearity and puts the \em second moments back on place of non-corrected
	values! Thus, saving the moments with sc::write() must be done after
	sc::calcmom() and BEFORE sc::calcidx().
*/
void calcidx(double microexp, bool isgen) ;
//throw (std::domain_error, std::range_error, std::invalid_argument, std::out_of_range, sc::dataproblem) ;

/**
	@brief	Get the number of measurements done in a current accumulation time
	@return	measurement number 
	
	Example: if again() was just used, the function will return "0" since no
	measurement were done up to now.
	
*/
int getcurmeas() ;
//throw (std::domain_error);

/**
	@brief	Reset the counter of accumulated indices in local index storages
	
	Once the indices in the private module storages are not needed any more
	(averaged, displayed and saved to disk), the counter of current row in them
	should be reset to avoid the matrices overfill.

	@see write() average() init()
*/
void again() ;
//throw() ;

/**
	@brief	Deallocate the memory reserved for module index storages
	
	This function releases the memory which is allocated for index storages by
	init().
	
	This function should be called only to end with calculations at all, since
	the modifications of arrays may be done by the calls to init() with no
	intermediate done().

	\note
	This function must be kept in strict sync with init().
*/
void done() ;

/**
	@brief	Write the latest instant indices or count moments to the disk
	@param	f    	Disk file structure pointer
	@param	what	DSI: write the last obtained index set (both (D)SI and 
		DESI), MOM: write count moments
	@param	stamp	Reference string of Mode/UT to put in the first column
		(if \em !header) or its column name (if \em header)
	@param	suffix	Last-field string of a general purpose. Will be put after
		the last index value or last word of the header. By default - nothing to
		put.
	@param	header	Write header line instead of data
	
	Function writes the last computed indices or moments on the disk in one
	ASCII line. The first token in the line is the reference  \em stamp (which,
	e.g., contains the comment symbol "# ",  Normal/Generalized mode symbol
	"N/G " and some word with the UT hour).

	- if \em what==sc::DSI:
	
	If \em header is \b true, the header is created like follows:
	\verbatim
<stamp (name)> SI(A) SI(B) .. DSI(AB) ....... DESI3(A) ..
	\endverbatim
	where SI relate to normal indices, DSI - to differential indices, DESI3 are
	differential exposure indices with smoothing 3-element window. Header line 
	is kept in sync with what is written in data-writing mode (\em
	header==false)

	For writing the index values, the format IDXFMT is used. Numbers are 
	delimited with IDXDLM. The header-writing call is only possible after 
	the init() was called; data-writing is allowed after at least one
	calcidx() call. 
	
	The index set is written in one line, which has a structure corresponding
	to that given in the header: \em stamp, \em Nchannel normal indices,  \em
	napcomb(Nchannel)-Nchannel differential indices, and  finally \em Nchannel
	differential exposure indices (DESI).
	
	- if \em what==sc::MOM:
	
	If \em header is \b true, the header is created like follows:
	\verbatim
<stamp (name)> MEAN(A) MEAN(B)... SIG(A) SIG(B) .. COV(AB) ... COV1(A) ... COV2(A) ..
	\endverbatim
	where SIG relate to normalized dispersions and COV to normalized
	covariances of channels and	to normalized auto- and cross-covariance of the
	signal in channels  with lag 1 and 2. 

	\note The RAW mean counts are written (not non-linearity corrected). As
	about second moments (SIGs, COVs), they are written as they are. So, if the
	procedure was called in between the calls to sc::calcmom() and
	sc::calcidx(), then the purely raw count statistics are written. This is
	recommended. If after sc::calcidx(), then second moments will be corrected
	which may cause confusion while processing them again from the file.

	For writing the moment values, the format MOMFMT is used. Numbers are 
	delimited with MOMDLM. The header-writing call is only possible after 
	the init() was called; data-writing is allowed after at least one
	calcidx() call. 

	As a last field for both \em what-selections, the \em suffix string may be
	printed. In MASS Software it is an optional \em address record to link the
	instantaneous index values with the count series data in the \em count-file
	from which they were derived.

	- This function must be kept in strict sync with calcidx().
	- The write() must not be called if again() was already called  after the
	calcidx() call.
	- if macro NODUBHEAD defined, the header is output (\a header is \em true)
	only once for the given value of the file pointer \a f. Repetitive output
	of the header \em with \em the \em same \em stamp is suppressed.
	Suppressing may be reset with sc::isheader(): once forever - with no
	parameters, for a given file - with a file \a f parameter only.

	@see init() calcidx() again()
*/
void write(FILE * f, scwhat what, const char * stamp, const char * suffix = 0,
	bool header = false) ;
//throw(std::domain_error, std::invalid_argument, sc::ioproblem) ;

/**
	@brief	Is the header with a given stamp output in a given file
	@param	f	file pointer (0 to output all headers further)
	@param	stamp	Header prefix to check for being output in \a f
	
	This function says for a given file \a f whether the header with the given
	\a stamp was already output (using write() or writeavg() with the
	parameter \em header=true). 

	If called without \a stamp, it resets the collection and thus allows to 
	output the headers by write() and writeavg() once again.
	
	This flagger is called from sc::write() and sc::writeavg() each time they
	are envoked for header writing IF the macro NODUBHEAD defined.
*/
bool isheader(FILE * f, const char * stamp = 0);

/**
	@brief Import the string with count moments into the moment storages
	@param	line	string as recorded by sc::write() with sc::MOM option
		but \b without a stamp
	
	This utility scans the character string \a line and writes the parsed
	numbers into the elements of the module storages for mean counts and their
	normalized dispersions and auto/cross-covariances. This may be used to
	reproduce the scintillation indices from the count moments written in a
	previous program run to disk file. 

	The line must be identified (e.g. by stamp) and read from the disk into the
	character string without this stamp, i.e. the first word in  \a line must
	already be the first mean count value (i.e. mean in the first channel).
*/
void str2mom(const char * line);

/**
	@brief Average a matrix of values along the second dimension with row 
		selection
	@param	matrix	matrix[nrow][ncol] of values to average along second 
		dimension (i.e. to obtain an average row). Meanwhile, if ncol==1, 
		\em matrix is interpreted as a vector to be averaged into one number.
	@param	ncol	number of columns in \em matrix[][]
	@param	nrow	number of rows in \em matrix[][]
	@param	rowselect	row selection vector[\em nrow] (set NULL to select all)
	@param	select	selection value in \em rowselect (no matters if
		\em rowselect==NULL )
	@param	avg 	vector[\em ncol] of resulting average values (one value if
		\em ncol==1)
	@param	err2	vector[\em ncol] of resulting squared errors of avg.values 
		(one value if \em ncol==1)
	@param	lag  	maximal lag in correction of error for the correlation of
		values
	
	The rows are selected for averaging from the storage \em matrix by
	comparison of respective values in \em rowselect with \em select up in a
	range of the storage row numbers [0..nrow-1]. Average value is thus derived
	for each column for rows with index \em row	complying the \em
	rowselect[row]==select condition. 
	
	Errors of average values are computed corrected for correlation of values
	as given in Korn & Korn "Spravochnik po matematike" 4th edition, section
	19.8 with lag of auto-covariance up to \em lag: 

	\f[
	\sigma^2_{\overline x} = \frac{1}{N} \left[\sigma^2_x  +
		2\sum\limits_{k=1}^{lag} (1-\frac{k}{N}) \rho_k(x)\right]
	\f]
	where the \f$\rho_k\f$ is the covariance of values \em x with the 
	lag \em k and \f$\sigma^2_x\f$ is the dispersion of values
	\f[
		\rho_k(x) = \frac{1}{N-k}\sum\limits_{i=1}^{N-k}(x_i-\overline
		x)(x_{i+k}-\overline x), ~~~
		\sigma^2_x = \frac{1}{N-1}\sum\limits_{i=1}^{N}(x_i-\overline x)^2
	\f]
	
	If a given \em lag is more than \em nrow, the maximal lag is taken as 
	\em nrow-1.
	
	@see average
*/
void avgmatrix(double ** matrix, int ncol, int nrow, 
	signed char * rowselect, signed char select, 
	double * avg, double * err2, int lag) ;
//throw(std::bad_alloc, std::invalid_argument, sc::dataproblem) ;

/**
	@brief	Average scintillation indices and compute their \em squared errors
	@param	isgen	flag that the generalized mode was used
	@param	lag  	Maximal counts lag to account for the index correlation in
	the error of mean (MAXLAG may be set)
	
	This utility is called normally after \em accumulation \em time span during
	which rows of the index matrices are computed getcurmeas() times times. It
	is aimed to obtain some \em mean value of each index or, in other words, to
	implement the averaging of the index storage along its time-dimension.
	Also, the squares of uncertainties of these indices are  assessed.

	The computed average indices, average shifted indices, and DESIs (both
	normal and shifted) are put in the global vectors, to be accessed by
	getavgidx() and geterridx(). Note, that the errors are computed and
	stored \em squared \em absolute errors, meanwhile the latter function
	returns the \em relative index value.

	The utility avgmatrix() is used for averaging all the storages.

	\note 
	To the moment, the average index itself is a simple mean with no 
	rejection; it can in principle be implemented as a mean with rejection 
	or as a median.

*/
void average(bool isgen, int lag) ;
//throw(std::bad_alloc, std::invalid_argument, sc::dataproblem) ;

/**
	@brief	Write the \em average (accumulation-time related) indices or channel
		fluxes on the disk
	@param	f	Disk file structure pointer
	@param	what	DSI:  write the average indices, AFLUX: write the average
		channel fluxes, AFLUXP: write the average channel fluxes with
		non-Poisson values
	@param	isgen	"false": write non-shifted values with errors, "true": write 
		shifted	values with errors
	@param	stamp	Reference string to put in the first column
	@param	suffix	Last-field string of a general purpose. Will be put after
		the last number in a line or after the last word of the header. 
		By default - nothing to	put.
	@param	header	Write header line instead of data
	
	Function writes the average values computed by average() and their errors
	on the disk in one ASCII line. The first token in the line is the
	reference \em stamp (line prefix and UT normally).

	For writing the index values and their errors, the format IDXFMT is  used;
	numbers are delimited with IDXDLM. For writing the fluxes and their errors,
	the format FLXFMT is  used; numbers are delimited with FLXDLM; the
	non-Poisson parameter \em P has a fixed format "\%6.4f".

	The set of values is taken from	shifted or non-shifted parts of storages,
	according to the parameter \em isgen.

	The values are written in one line, which has the following structure:
	
	- if \em what==DSI:
	
	\em stamp, average normal and differential indices for all channels and
	their combinations, each followed by its error and  average differential
	exposure indices (DESI) each followed by its error. 

	- if \em what==AFLUX:

	\em stamp, the average fluxes each followed by its error, for all channels.

	- if \em what==AFLUXP:
	
	the same as for AFLUX, but each flux error is followed by the channel
	non-poissonity parameter from the channel structure (see chan_init()). It is
	supposed that this parameter is updated by statflux() with non-default \em 
	nonpois parameter before.
	
	As a last field, the \em suffix string may be printed, similarly to write().
	This option is added to make it possible to save the individual MASS counts
	in test modes, e.g. the Detector Test. The suffix may thus contain the 
	optional \em address to link the obtained data with the count series data 
	in the \em count-file plus the number of base-time series which results were
	averaged and which count buffers are written starting from this \em address.

	\note
	This function must be kept in strict sync with average().
	- if macro NODUBHEAD defined, the header is output (\a header is \em true)
	only once for the given value of the file pointer \a f. Repetitive output
	of the header \em with \em the \em same \em stamp is suppressed.
	Suppressing may be reset with sc::isheader(): once forever - with no
	parameters, for a given file - with a file \a f parameter only.

	@see init() 
*/
void writeavg(FILE * f, scwhat what, bool isgen, const char * stamp, 
	const char * suffix = 0, bool header = false) ;
//throw(std::domain_error, std::invalid_argument, sc::ioproblem) ;

/** 
	@brief	Get the instant index value
	@param	meas	Measurement (row number of the index matrix)
	@param	what	0 for usual index, 1 for DESI
	@param	i	index of channel or combination of channels [0..napcomb()-1]
	@return	the index or some bad number on error
	
	One may use getcurmeas()-1 for the parameter \em meas, to return the
	last obtained measurement result. Indices are ordered in the returned
	vector as specified in apername() documentation; the length of vector is
	napcomb(Nchannel) or \em Nchannel for  what==1 (i.e. DESI).

*/
double getidx(int meas, scwhat what, int i) ;
//throw (std::domain_error,std::out_of_range,std::invalid_argument) ;

/**
	@brief	Get pointer to the array of instant scintillation indices
	@param	meas	Measurement (row number of the index matrix)
	@param	what	DSI: for usual index, DESI: for DESI
	@return	pointer to the array of indices or to some bad number 
		on error

	This function is needed for atm::calcint() 
	@see geter2idxptr()	
*/
double * getidxptr(int meas, scwhat what) ;

/** 
	@brief	Get the averaged index value
	@param	what	DSI: for usual index, DESI: for DESI
	@param	isgen	0 for array of non-shifted average indices; 1 for shifted
	indices
	@param	i	index of channel or combination of channels [0..napcomb()-1]
	@return	averaged index value or some bad number on error
	
	The call with \em isgen==true is only possible if generalized mode was used
	during the last accumulation time.
	
	@see getidx()
	
*/
double getavgidx(scwhat what, bool isgen, int i) ;
//throw (std::domain_error,std::invalid_argument) ;

/**
	@brief	Get pointer to the array of average scintillation indices
	@param	what	DSI: for usual index, DESI: for DESI
	@param	isgen	"false" for array of non-shifted average indices; "true" for 
		shifted indices
	@return	pointer to the array of the average indices or to some bad number 
	on error

	This function is needed for atm::calcn2().
	@see geter2idxptr()	
*/
double * getavgidxptr(scwhat what, bool isgen) ;

/** 
	@brief	Get the \em relative error of the average index
	@param	what	DSI: for usual index, DESI: for DESI
	@param	isgen	"false" for array of non-shifted average indices; "true" for 
		shifted indices
	@param	i	index of channel or combination of channels [0..napcomb()-1]
	@return	relative error of the average index or some bad number on error
	
	If the stored value of the index squared error is positive, then the ratio
	of the square root of this squared error to the index value is returned;
	otherwise 0 is returned. 

	\note
	The squared error may, in principle, be negative due to the wrongly 
	accounted correlation of values in avgmatrix()

	@see getavgidx()	
*/
double geterridx(scwhat what, bool isgen, int i) ;
//throw (std::domain_error,std::invalid_argument) ;

/**
	@brief	Get pointer to the array of \em squared errors of average 
		scintillation indices
	@param	what	DSI: for usual index, DESI: for DESI
	@param	isgen	"false" for array of non-shifted average indices; "true" for 
		shifted indices
	@return	pointer to the array of the errors of average indices or to some bad 
	number on error
	
	This function is needed for atm::calcn2(). 
	
	\attention
	The internally stored array of average index errors contains the squared
	errors of indices computed by avgmatrix(). These are the values returned
	here and needed for the Cn2-profile restoration. Meanwhile, the result of
	geterridx() is a \em relative error of index.
	
	@see getavgidx()	
*/
double * geter2idxptr(scwhat what, bool isgen) ;

/** 
	@brief 	Get the mean count in the channel in [counts per ms]
	@param	iscor	"false": return the raw mean; "true": return the 
		non-linearity corrected mean
	@param	i	index of channel [0..Nchan-1]
	@param	meas	Measurement number in [0..sc::getcurmeas()-1], for
		iscor==true only. Last value is returned by default.
	@return	mean count in channel during the last base time [counts per ms]

*/
double getmean(bool iscor, int i, int meas = -1) ;
//throw (std::domain_error) ;

/** 
	@brief 	Get the average flux in the channel in [counts per ms]
	@param	isgen	"false": non-shifted average counts; "true": shifted 
		average counts
	@param	i	index of channel [0..Nchan-1]
	@return	average flux in channel [counts per ms]

	Average fluxes are obtained in average().

	See notes to getidx().
*/
double getavgflux(bool isgen, int i) ;
//throw (std::domain_error) ;

/** 
	@brief 	Get the \em relative error of the average flux in channel
	@param	isgen	"false": non-shifted average counts; "true": shifted 
		average counts
	@param	i	index of channel [0..Nchan-1]
	@return	relative error of average flux in channel

	Average fluxes are obtained in average().

	The ratio of the square root of the flux squared error to the flux
	value is returned.

	See notes to geterridx().
*/
double geterrflux(bool isgen, int i) ;
//throw (std::domain_error) ;

/**
	@brief	Get the normalized second moment of counts in channel
	@param	lag	0: simple dispersion (index<Nchannel) or covariance of counts
	(for returned array index more or equal to Nchannel); 1,2: the same with 
	lag=1,2 counts
	@param	i	index of channel or combination of channels [0..napcomb()-1]
	@return	the moment value (normalized dispersion or auto/cross-covariance)

*/
double getsig(int lag, int i) ;
//throw (std::domain_error,std::invalid_argument) ;

/**
	@brief	Compute the expected scintillation indices for the results of
		the MASS statistical test
	@param	ndata	Number of counts in the channel buffers (equal for
		all channels)
	@param	microexp	Exposure time of each count (\em micro-exposure) to scale
		the non-linearity parameter and background in calculations (see
		chan_init())
		
	This is the second utility (after calcidx()) which deals with the channel
	count series.

	MASS statistical test is performed using the specially modulated flux from
	the control LED. The current through the control LED is modulated step-like
	synchronously with the counting time diagram with a period of four
	micro-exposures. During each first and third takts of this modulation the
	current is at the middle level, whilst at the second and the fourth takts 
	it is increased	and decreased, respectively, relative to this middle level. 
	
	The task of this utility is to determine these middle, increased and
	decreased light levels and to compute the ("theoretical") expected
	scintillation indices (SI) and DESI which correspond to this modulation 
	(see calcidx()):

	\f[
		s^2_{th} = \frac{1.5 \sigma^2 - 0.5 \rho_1}{\overline{flux}^2},~~~~
		s^2(de)_{th} = \frac{2}{9}\left( 3 s^2_{th} + \frac{\rho_2-4\rho_1}{\overline{flux}^2}\right)
	\f]
	
	Here, for the specified way of modulation, the dispersion and
	auto-covariances with lags 1 and 2 are calculated as
	\f[
		\sigma^2 = \frac{2\Delta_0^2 + \Delta_+^2 + \Delta_-^2}{4},~~~
		\rho_1 = \frac{\Delta_0 (\Delta_+ + \Delta_-)}{2},~~~
		\rho_2 = \frac{\Delta^2_0 + \Delta_+ \Delta_-}{2},
	\f]
	and Deltas with subscripts "0", "+" and "-" are the deviations of the
	middle, increased and decreased light levels, respectively, from the
	average level.  Middle level may differ from average level if the positive
	and  the negative variations are different in magnitude.
	
	The averaged levels are corrected for non-linearity using the channel 
	parameters set by chan_init() and the supplied micro-exposure time \em
	microexp. No correction for background is made.

	The call of this utility implies that the module is initialized with init()
	the same way as for usual measurements with calcidx().	Since, normally,
	the calcidx() is also started on the \b same count series, the capacity \em
	nmeas of the index storage must be \b doubled while calling init() before
	the statistic test evaluation. It is assumed that \b no corrected data are
	present in the series (as if \em ncorr array in calcidx() contained zeros).
	So, damaged series should not be fed to stattest().

	The expected values of indices are put in the index storage in the same way
	as calcidx() does. Since the flux modulation is the same in all the
	channels of MASS, cross-channel differential indices measured during the
	test (by indices()) are set to zero. 
	
	\note
	The "mode" set for the computed expected indices in the index storage is set
	to sc::MODEGEN to allow the simple averaging of them similarly to the case
	of the generalized mode measurements. Thus, the averaging of statistic test
	results must be done with average() with the first parameter \b true .
	
*/
void stattest(int ndata, double microexp) ;
//count_t * data, int n, double nonlin, double * scind) ;
//throw (std::invalid_argument, sc::dataproblem) ;

/**
	@brief Recurrence computer of average fluxes and their \em squared \em
		errors from the count series in channels
	@param	ndata	Actual number of counts in the count series stored in the 
		channel buffers (equal for all channels)
	@param	ncorr	array of numbers of corrected counts in series (i.e. 
		replaced with mean of series if missing) 
	@param	microexp	Exposure time of each count (\em micro-exposure) to scale
		the non-linearity parameter and fluxes in calculations (see	chan_init())
	@param	k	sequential number of current channel series in the measurement 
		session (i.e. this call is \em k-th to measure the statistics stored in
		\em avg and \em er2)
	@param	isfilt	To account for the long-term modulation of the signal
		(clouds, etc.) in dispersion \em sig
	@param	nonpois	non-Poisson parameters array[0..Nchan-1], by default -
		not computed
	@param	avgflux	Average flux array[0..Nchan-1], by default - the local storage
		accessible with getavgflux()
	@param	er2flux	\em Squared \em absolute error of flux: array[0..Nchan-1], 
		by default - the local storage accessible with geterrflux()

	This is a third utility (after calcidx() and stattest()) which deals with
	the channel data series. It is suited for calculation of the average flux
	in a channel and its error. Counts are assumed to be \em non-correlated, so
	the squared error of average is \em k*ndata times less than the count
	dispersion. 
	
	The results are put in the global module storage of \em average fluxes (or
	other place if last two or three parameters are not defaulted) or \em updated there on
	calls with \em k>1 (see below). They are permanently available with
	functions getavgflux() and geterrflux() after the first call with \em k=1
	(unless the parameters \em avgflux and \em er2flux are specified
	explicitly). It is assumed that the storages for an average flux and its
	dispersion are allocated with init().

	For any call to the function, the average count and its \em dispersion
	for a current series are computed like follows (hereafter \em n stays for \em ndata parameter value):
	\f[
		\overline{data} = \frac{1}{n}\sum\limits_{i=0}^{n-1} data[i] ,
	\f]
	\f[
		\sigma_{data} = \frac{1}{n-1}\sum\limits_{i=0}^{n-1}
		(data[i]-\overline{data})^2
	\f]

	Here all the counts in channels are actually used corrected for non-linearity
	individually using the channel parameter \em deadtime initialized by 
	chan_init():
	\f[
		data[i] = data_{raw}[i] (1 + data_{raw}[i] \times deadtime / microexp)
	\f]
	\note
	The deadtime-parameter must be set to zero in chan_init() before making
	statflux() for determination of the	non-linearity of detectors itself.
	It should be noticed also that the  background level in channels is 
	ignored (assumed zero). Detector non-poissonity is evidently not
	taken into any account but is computed and updated in channels if \a
	nonpois is provided as non-null.

	This utility may be used either as a plain average and its error computer
	or as a tool for calculation of the statistics of a long series consisting
	of multiple data pieces (\em base-time series). In both instances, the
	first call to the utility must have \em k=1. If so, the average flux values
	and their errors in flux storages are assumed to contain no information
	about the  previous data series. Then they are assigned from the average
	count and its dispersion like follows:
	\f[
		avgflux = \frac{\overline{data}}{microexp}~~,~~~~
		er2flux = \frac{\sigma_{data}}{n\times microexp^2} ~~,
	\f]
	that means that we express the average flux and its error
	\f$\sqrt{er2flux}\f$ in [counts per unity of micro-exposure].

	When one wishes to obtain the statistics of a continued series of counts
	(base-times of the same accumulation time, i.e. with the same source of the
	signal), the values in \em avgflux and \em er2flux obtained in a
	previous $(k-1)$-th call to statflux() are \em back-converted into the
	\em non-normalized counts \em avgcount and their dispersion \em er2count:
	\f[
		avgcount = avgflux \times microexp~,~~
		er2count = er2flux \times n(k-1) \times microexp^2
	\f]

	and updated with a current $k$-th series statistics like follows:
	\f[
		avgcount_{new} = \frac{1}{k} \left( (k-1) avgcount + \overline{data} 
		\right)~~,
	\f]
	\f[
		er2count_{new} = \frac{1}{k} \left( (k-1) er2count + \sigma_{data} 
		\right) + Corr~~,
	\f]
	where the correction \f$ Corr=(k-1)\times(avgcount_{new}-avgcount)^2 \f$ of
	the	dispersion for the low-frequency  modulation of the average count is
	computed and added if \em isfilt==false. Note, that \em k must be
	incremented by user.

	If the parameter \em nonpois is not null, then the estimate of the
	non-poissonity of the detector is computed for each channel as:
	\f[
		nonpois = \frac{er2count}{avgcount}
	\f]
	Before exiting the function, the updated parameters are again converted into
	the flux units similarly to the formula given above:
	\f[
		avgflux = \frac{avgcount}{microexp}~~,~~~~
		er2flux = \frac{sigcount}{n\times k\times microexp^2 }.
	\f]

	\note	
	The output results are saved in the module average flux storages \b even if
	the \em avgflux[] and \em er2flux[] place-holders are explicit parameters. 
	This allows	to write the \b last \b obtained accumulated statistics with 
	writeavg() and access them with getavgflux(), geterrflux().

	\attention
	The confusion may occur on the meaning of the parameters saved in \em er2flux.
	This is the storage for the \em squared \em absolute error of average
	flux; meanwhile the value returned by geterrflux() is a \em relative 
	error of flux.

*/
void statflux(int ndata, int * ncorr, double microexp, int k, 
	bool isfilt=false, double * nonpois = 0, double * avgflux = 0, double *
	er2flux = 0) ;
//throw (std::invalid_argument) ;

/** Smoothing window width in DESI calculations. Used implicitly in
calcidx() ) */
const int DESIBIN = 3 ;

/** Format of the relative error of any value output in the file with a leading 
"\%s" to insert a delimiter. Length comes from the length of a longest heading*/
const char ERRFMT[] = "%s%8.3f" ;

/** Index writing printf-format in the file with a leading "\%s" to insert a
delimiter. Possible "-"-sign is accounted. */
const char IDXFMT[] = "%s%7.4f" ;
/** Index file delimiter */
const char IDXDLM[] = " " ;

/** Average flux per micro-exposure writing printf-format in the file with 
a leading "\%s" to insert a delimiter. MUST include the precision as '*' */
const char FLXFMT[] = "%s%6.*f" ;

/** Flux accuracy: maximal number of decimal digits (see fluxprec()) */
const int FLXPREC = 3;

/**
	@brief	Return the decimal digits number for the FLUX representation
	@param	flux	flux value
	@param	maxprec	maximal number of decimal digits
	@return	precision to use in *printf() functions with formats like "%6.*f"
*/
inline int fluxprec(double flux, int maxprec = sc::FLXPREC) 
{ 
	int prec = FLXPREC - int(log10(flux)) ;
	if (prec<0) prec = 0 ;
	else if (prec>FLXPREC) prec = FLXPREC ;
	return prec ;
}

/** Average flux delimiter in file */
const char FLXDLM[] = " " ;

/** non-Poissonity parameter writing printf-format in the file with a leading 
"\%s" to insert a delimiter */
const char PFMT[] = "%s%5.3f";

/** Moments (means, covariances) writing printf-format in the file with 
a leading "\%s" to insert a delimiter */
const char MOMFMT[] = "%s%#10.4g" ;
/** Moment file delimiter */
const char MOMDLM[] = " " ;

/** Not-shifted flag \em isgen in average() */
const int MODENORM = 0 ;
/** Shifted flag \em isgen in average() */
const int MODEGEN  = 1 ;
/** Initial mode: unknown in isgen[] */
const int MODENO = -1 ;

/** Maximal normal scintillation index stored in a storage (initial filtering) */
const double MAXSCIND = 2;
/** Minimal normal scintillation index stored in a storage (initial filtering) */
const double MINSCIND = 0;

/** Star escapes filtering by R=SI(largest aperture)/SI(smallest): critical
ratio of (R-<R>)/sigma(R) */
const double KAPPA = 4.0;
/** Star escapes filtering by R=SI(largest aperture)/SI(smallest): maximal
number of iterations. Set to zero to omit filtering at all. */
const int MAXFITER = 4;

/** Maximal lag in calculations of the error of mean for correlated values */
const int MAXLAG =0 ;

/** Macro to suppress the repetitive output of headers by sc::write(),
writeavg() */
#define NODUBHEADER
} 	// of namespace sc

#endif
