/****************************************************************************/
/** @file weif.cpp
                 MASS project: TURBINA module file                         
                             weif.cpp                                      
                    Description is given in weif.hxx                        */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition WEIFTEST in compilation for the module self-tests:
gcc -c weif.c .... -DWEIFTEST=1 */
//#define WEIFTEST
/* Insert similarly the definition WEIFDEBUG to print all the debug information 
for the module procedures */
//#define WEIFDEBUG

#include <math.h>
#include "weif.hxx"
#include "nrutil.h"
#include "nr.h"
#include <iostream>

using namespace std;
#ifdef WEIFTEST
/*------------------------------ TESTING PART --------------------------*/
#ifndef WEIFNTEST
#define WEIFNTEST 0
#endif
/* We test the module with the main() part included: see definition WEIFTEST */

/*----------- Comes normally from configuration file: cf_... vars ----------*/

/* Apertures: */
/* number of apertures in segmentor : */
int cf_naper = 4;
/* Physical size of segmentor annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.106, 0.197, 0.354};
double cf_aper_out[] = { 0.103, 0.190, 0.350, 0.669};
//double cf_aper_inn[] = { 0.000, 0.100};
//double cf_aper_out[] = { 0.100, 0.200};
/* Magnification of the system: */
double cf_magnif = 18.9	 ;

/* Altitude grid: */
double cf_z0 = wf::Z0, cf_zmax = wf::ZMAX, cf_dz = wf::DZ, cf_dzmin = wf::DZMIN;
/* Altitude grid: */
//double cf_z0 = 0.0, cf_zmax = 25.0, cf_dz = 0.5, cf_dzmin = 0.0;

/* Alternative grid: to interpolate to */
double int_z0 = 0, int_zmax = 20.0, int_dz = 0.02, int_dzmin = 0;

double cf_shift = 0.5 ;	/* Shift weight functions by cf_shift kilometers */

char wf_fwo3[] = "weight.int";   /* Output weight file: interpolated */
char wf_fwo4[] = "weight.sh-";   /* Output weight file: shifted below */
char wf_fwo5[] = "weight.sh+";   /* Output weight file: shifted above */

int main( int ac, char **av ) {

/* Parameters parsing: */
if (ac<2) {
	cerr << "\tMASS PROJECT: testing the TURBINA module WEIF.CPP \n";
	cerr << "\tUsage: "<<av[0]<<" <sed file> <response file> <weight file>\n";
	return -1;
}
/* Get a weight file name: */
const char * wf_fn = av[3];
const char * resp_fn = av[2];
const char * sed_fn = av[1];

/* Cycling usage of functions to detect memory allo/deallocation errors: */
int ntest=0;
do {

	/* Test and, if bad, compute the weight: */
	int result = wf::checalc( sed_fn, resp_fn, wf_fn,
		cf_naper, cf_magnif, cf_aper_inn, cf_aper_out,
		0,0,0,0,true,wf::progress);
	if (!result)
		cout << "Ok\n";
	else {
		cerr << "wf::checkalc result = " << result << " : ";
		cerr << nr::ermessage() << '\n';	
	}

	if (!WEIFNTEST) break;	// simple weight stand-alone computer mode
	
	/* Read back the tested/computed weight:*/
	wf_t wf;
	int err = wf.read(wf_fn) ;
	if (err < 0) {
		// can hardly happen: just tested with checalc()!
		cerr << "Read error: " << nr::ermessage() << ". Cannot be!\n";
		return err;
	} 

	/* Computed or successfully read: Interpolation with shifts: */
	for (int i=-1; i<2; i++) {
		char * whichfile ;
		wf_t wf_int;
		if (!i) {
			whichfile = wf_fwo3;
			cout << "Put another z-grid:" ;
			wf_int.setzgrid(int_z0,int_zmax,int_dz,int_dzmin);
		} else {
			whichfile = (i<0) ? wf_fwo4 : wf_fwo5 ;
			cout << "Put the same z-grid with shift " << cf_shift*i << "km:";
			wf_int.setzgrid(cf_z0,cf_zmax,cf_dz,cf_dzmin);
		}
		cout << "\nInterpolate weights of old wf into z-grid of new wf :";
#ifdef POLINT
		double interr =0 ;
		int iz_interr = 0, w_interr = 0 ;
		wf_int.interpolate(&wf, cf_shift*i,&interr, &iz_interr, &w_interr) ;
#else
		wf_int.interpolate(&wf, cf_shift*i) ;
#endif
		if (nr::erget()) {
			cerr << "Error : " << nr::ermessage() << '\n' ;
			return nr::erget();
		}
	#ifdef POLINT
	#ifdef INTRELERR
		cout << "Ok. Max.rel. err = " << interr
	#else	
		cout << "Ok. Max.err = " <<	interr/wf::WSCALE << " * ("<<wf::WSCALE<<')'
	#endif
		<< " at z=" << wf_int.getalt(iz_interr) << "km (" << iz_interr << 
		"-th step) in " << wf_int.name(w_interr) << " weight\n" ;
	#else
		cout << "Ok.\n";
	#endif

		cout << "\nWriting the interpolated weight file " << whichfile << ":\n";
		wf_int.write(whichfile) ;
		wf_int.setaper(0);
	}

	cout << "Ok.\n";
	cout << "====================== " << ntest << " =========================\n";
	ntest++;
} while (ntest<WEIFNTEST);

return 0;

}
#undef FILENAMESZ
#endif /* ifdef WEIFTEST end */
/*----------------------- TESTING PART END --------------------------------*/
/* local functions for pure elementary calculations of weights: */
namespace wf {

int zgrid_set(double ** zgrid, double z0, double zmax, double dz, double dzmin);
void make_fft(double * lambda, double * edist, int nedist) ;
double sfunc(double) ;
double sdum(double f) ;
double flama(double x) ;
double weight(double z, double d, 
              double eps1, double eps2, double eps3, double eps4) ;

void getspechar(double * lambda, double * edist, int nedist, double * leff,
double * lblue, double * lred) ;

/* Splines: define the first derivative of weights at z=0 equal to 0: */
const double DER0  = 0.0 ;
/* signal to set the "natural" upper boundary cond.: 2nd weight der.=0*/
const double DERNAT  = 1.0E31 ;
/* Conversion of [mkm] to [cm]: */
const double MKM2CM	 = 1E-4 ;

}

/***************************************************************************/
/*                  Public functions implementation                        */
/***************************************************************************/

/*------------------------- wf::progress -----------------------------------*/

void wf::progress(int stage) {
	/* Example function - progress indicator: */
	if (stage<100) 
		printf("\r%3d%% ", stage);
	else		 
		printf("\r%3d%% \n", stage);
	fflush(stdout);
}

/* Make the constants available without qualifiers in wf_t: */
using namespace wf ;
//using namespace sc ;

/*------------------------- wf_t::constructor -----------------------------------*/

wf_t::wf_t() {
//	printf("wf_t Constructor\n") ;
	/* Assign NULL to all pointers: "not allocated" : */
	zgrid = eps_inn = eps_out = 0 ;	/* 1-D arrays */
	ws = 0 ;		/* 2-D array */
	/* set zero in all parameters: */
	nz = naper = 0 ;
	z0 = zmax = dz = dzmin = diam = leff = lblue = lred = 0;
	staticapert = false ;
}

/*------------------------- clear --------------------------------------*/

void wf_t::clear() {

	if (ws && nz) {
		/* Release weight matrix memory: */
		if (ws) nr::free_dmatrix(ws, 0, getnw()-1, 0, nz-1) ;
		/* Assign NULL : "not allocated" : */
		ws = 0 ;
	}
	if (zgrid && nz) {
		/* Release zgrid array memory: */
		nr::free_dvector(zgrid, 0, nz-1) ;
		/* Assign NULL : "not allocated" : */
		zgrid = 0 ;
		nz = 0 ;
		z0 = zmax = dz = dzmin = 0 ;
	}
}

/*------------------------- setaper ------------------------------------*/

void wf_t::setaper(int nap, double dia, double * eps_in, 
		double * eps_ou ) {
/* Parameter names are made shorter than member names */
	
	if (!nap) {
		/* Release aperture arrays if allocated: */
		clear() ;	/* Release arrays if allocated */
		if ((naper>0) && (eps_inn!=NULL) && (eps_out!=NULL)) {
			if (!staticapert) {
				nr::free_dvector(eps_inn, 0, naper-1) ;
				nr::free_dvector(eps_out, 0, naper-1) ;
			}
			eps_inn = eps_out = NULL ;
			naper = 0; diam = 0 ;
		}
		return ;
	}
	
	if (!eps_in || !eps_ou) {
		nr::nrerror("WEIF: setaper(): NULL size-arrays supplied", nr::ERNUL);
		return ;
	}

	if ((dia<=0) || (dia>wf::MAX_DIAM)) {
		nr::nrerror("WEIF: setaper(): invalid diameter/magnification",nr::ERPAR);
		return ;
	}

	/* Recursive call for clearing the previous data: */
	if (eps_inn || eps_out)	setaper(0) ;
	
	diam = dia ;

	if (nap<0) {
		/* Copy size-pointers to members instead of allocation and copying: */		
		naper = abs(nap) ;
		eps_inn = eps_in ;
		eps_out = eps_ou ;
		staticapert = true ;
	} else {
		/* Allocate new arrays and copy the information: */
		naper = nap ;
		eps_inn = nr::dvector(0,nap-1) ;
		eps_out = nr::dvector(0,nap-1) ;
	
		for (int i=0; i<nap; i++) {
			eps_inn[i] = eps_in[i] ;
			eps_out[i] = eps_ou[i] ;
		}
		staticapert = false ;
	}
}

/*--------------------------- setzgrid -----------------------------------*/

void wf_t::setzgrid(double uz0, double uzmax, double udz, double udzmin) {
	int unz = wf::zgrid_set(&zgrid, uz0, uzmax, udz, udzmin) ;
	if (unz > 0) {
		/* No Error occured in zgrid generation, otherwise nz = error code */
		/* Success: */
		nz = unz ;
		z0 = uz0 ;
		zmax = uzmax ;
		dz = udz ;
		dzmin = udzmin ;
	} else 
		nr::nrerror("WEIF: setzgrid(): bad parameters", nr::ERZGR);
}

/*--------------------------- write ---------------------------------*/

void wf_t::write(const char * filename) const {
	if (!zgrid || !ws) {
		nr::nrerror("WEIF: write() empty weight object", nr::ERPAR) ;
//		cout << nz << "=Nz\n";
		return ;
	}

	FILE * f = fopen(filename,"wt") ;
	if (!f) {
		string msg = string("WEIF: write(): failed to open file ") + filename ;
		nr::nrerror(msg.c_str(), nr::ERFIO) ;
		return ;
	}
	
	/* Write the header: */
	/* # N <Naper> */
	/* Check that first write attempt is successful: */
	if (!(fprintf(f,"# N %d\n", naper))) {
		string msg = string("WEIF: write(): failed to write file ")+filename;
		nr::nrerror(msg.c_str(), nr::ERFIO) ;
		return ;
	}
	
	/* # <AperName> <InnerSize> <OuterSize> : */
	for(int i = 0; i< naper; i++ ) 
		fprintf( f,"# %s %6.3f %6.3f\n", name(i), 
			diam * eps_inn[i], diam * eps_out[i]);
			
	/* # S <Effective wavelength[mkm]> <50% blue point> <50% red point> */
	fprintf(f,"# S %9.4f %9.4f %9.4f\n",leff,lblue,lred);

	/* # Z <grid_z0> <zmax> <dz> <dzmin> */
	fprintf(f,"# Z %7.4f %7.4f %7.4f %7.4f\n",z0,zmax,dz,dzmin);
		
	/* # W <Weight writing scale> */
	fprintf(f,"# W %5.0E\n",wf::WSCALE);

	/* ## <header> */
	fprintf(f,"##Altit. ");
	int nw = getnw() ;
	for (int i = 0; i < nw; i++) {
		fprintf(f," W(%s)/%5.0E",name(i),wf::WSCALE);
	}
	fprintf(f,"\n");

	/* Write the data: */

	for (int i = 0; i < nz; i++) {       
		fprintf(f,"%7.3f  ", zgrid[i] );
		for (int j = 0; j < nw; j++) {
			if ((fprintf(f," %10.4E", ws[j][i]/wf::WSCALE ))==0) {
				string msg = string("WEIF: write: failed to write file ")
					+filename ;
				nr::nrerror(msg.c_str(), nr::ERFIO) ;
				return ;
			}
		}
		fprintf(f,"\n");
	}
	fclose(f);
}

/*--------------------------- readhdr -------------------------------*/
/**
	@name	readhdr()
	@memo	Read the weight file header values into the numeric members
	@param	filename	the name of file to read
	@param	weightscale	scaling factor of weight matrix data writing
	@return	0 if reading Ok, <0 if error

	This supplementary function reads the header of the weight file. It is
	used  in other public functions of the module and made private since it has
	no independent value.

	The wf_t members are assigned zero before reading and thus remain zero if
	not found in a header.
	
	If header is read successfully, then, apart from scalar fields, the arrays
	\em eps_inn[] and \em eps_out[] are allocated and initialized with aperture
	relative sizes. Other array pointers are not changed.
	
	If the error is returned, the wf_t has no allocated arrays.
		
	@see checkfile() read()
	
*/
int wf_t::readhdr(const char * filename, double * weightscale) {
	int toread = 0;	/* counter of aperture data to read */

	FILE * f = fopen(filename,"rt") ;

	if (!f) {
		string msg = string("WEIF: readhdr() failed to open file: ")+filename;
		nr::nrerror(msg.c_str(), nr::ERFIO);
		return nr::ERFIO;
	}
	if (naper) setaper(0) ;	/* If something's inside - clean it */
	
	char line[wf::LTEXT+1]; 	/* text buffer to read from the header */
	while (fgets(line,wf::LTEXT,f)) {
#ifdef WEIFDEBUG
printf("Read: %s",line);
#endif
		/* Put the id pointer into the first token found in s: */
		char * id = strtok(line, wf::FDELIM) ;
#ifdef WEIFDEBUG
printf("id=%s\n",id);		
#endif
		if (id  				/* not empty line */
			&& (!strcmp(id,"#"))) {	/* first token is '#' */
			id = strtok(NULL,wf::FDELIM) ;	/* extract key */
		} else {
			if (strcmp(id,"##")) 
				/* not comment either: stop reading */
				break ;
			else
				/* Comment probably */
				id = 0 ;
		}
#ifdef WEIFDEBUG
printf("id2=%s\n",id);		
#endif
		if (id) {
			char key[2] = {' ',char(0)};
			strncpy(key,id,1) ;	/* Save key */
			/* Extract first value (may be last): */
			id = strtok(NULL, wf::FDELIM) ;
#ifdef WEIFDEBUG
printf("key=%s, val1=%s\n",key,id);
#endif			
			switch (key[0]) {
				case 'N': 
					if (!naper) {
						toread = atoi(id) ;
						if (toread > 0) {
							naper = toread ;
							eps_inn=nr::dvector(0,naper-1) ;
							eps_out=nr::dvector(0,naper-1) ;
							for (int i=0; i<naper; i++) 
								eps_inn[i] = eps_out[i] = 0 ;
						} else {
							toread = 0 ;
						}
					} 
					/* else - ignore second declaration of N */
				break;

				case 'Z': 
				if (id) {
					z0 = atof(id) ;
					if ((id = strtok(NULL, wf::FDELIM))) {
						zmax = atof(id) ;
						if ((id = strtok(NULL, wf::FDELIM))) {
							dz = atof(id) ;
							if ((id = strtok(NULL, wf::FDELIM))) {
								dzmin = atof(id) ;
							}
						}
					}
				}
				break;

				case 'S': 
				if (id) {
					leff = atof(id);					
					if ((id = strtok(NULL, wf::FDELIM))) {
						lblue = atof(id) ;
						if ((id = strtok(NULL, wf::FDELIM))) {
							lred = atof(id) ;
						}
					}
				}
				break;
				
				case 'W': 
				if (weightscale)
					*weightscale = atof(id);
				break;
				
				default:
				/* Could it be the aperture designation? */
				int i = (int)key[0] - (int)'A';	/* index guess */
#ifdef WEIFDEBUG
printf("i=%d key=%s\n",i,key);
#endif
				if ((i>=0) && (i<naper) && 	/* valid for declared naper */
					(toread>0) && /* naper's declared, not all apertures read*/
					(eps_inn[i]==eps_out[i]) &&	/* still zero */
					(strlen(key)==1)	/* Single character */
					) {
					/* Still missing aperture specifications follow: */
#ifdef WEIFDEBUG
printf("Aperture %s: ",key);
#endif
					eps_inn[i] = atof(id) ;
					if ((id = strtok(NULL,wf::FDELIM))) {
						eps_out[i] = atof(id) ;
#ifdef WEIFDEBUG
printf("%f - %f\n",(float)eps_inn[i],eps_out[i]);
#endif
						toread-- ;
					}
				} 
			} 
		} 
	} ;
	fclose(f);

#ifdef WEIFDEBUG
printf("\nFile end. Toread=%d, naper=%d, z0=%5.2f,zmax=%5.2f, dz=%f dzmin=%f\n",
	toread,naper,z0,zmax,dz,dzmin);
#endif
	/* Below toread will play the role of ifok-flag: */
	if ((z0<zmax) && (dz!=0) && (leff>0) && (lblue>0) &&
		(lred>lblue) && (naper>0) && (toread==0)) {
		/* Success: normalize and assign the aperture data and return Ok */
#ifdef WEIFDEBUG
printf("z-grid Ok, naper>0\n");
#endif
		/* Find maximum aperture size: */
		double maxaper = 0;
		for (int i=0; i<naper; i++) {
			if (maxaper<eps_inn[i]) maxaper = eps_inn[i] ;
			if (maxaper<eps_out[i]) maxaper = eps_out[i] ;
			if (eps_out[i]<eps_inn[i]) maxaper = 0 ;	/* error */
		}
		/* Normalize apertures to this maximal size: */
		if ((maxaper>0) && (maxaper<=wf::MAX_DIAM)) {
			/* Ok: there are non-zero apertures: */
			for (int i=0; i<naper; i++) {
				eps_inn[i] /= maxaper ;
				eps_out[i] /= maxaper ;
			}
			diam = maxaper ;
			return 0 ;	/* return on success */
		}
	}
	/* If we're here, at least one parameter value has no sense */
	if (naper>0) {
		/* Release useless info: */
		setaper(0) ;
	}
	string msg = string("WEIF: readhdr(): bad header of ")+filename ;
	nr::nrerror(msg.c_str(), nr::ERFIO) ;	/* not ERPAR sinse the file may be bad, not usage of
						function */
	return nr::ERFIO ;
}

/*--------------------------- checkfile ------------------------------*/

int wf_t::checkfile(const char * filename) const {

	wf_t file_wf ;

#ifdef WEIFDEBUG
printf(" WEIF: checkfile(): err(in)=%d \n",nr::erget()); 
#endif

	/* check if the file header is ok (weightscale is not interesting): */	
	int err = file_wf.readhdr(filename, NULL) ;
	if (err) return err ;	/* file_wf has not no allocated arrays if error */

#ifdef WEIFDEBUG
printf("Read Ok. Compare parameters...\n");
#endif
	/* File is Ok. Let's compare its parameters: */

#ifdef WEIFDEBUG
printf("Compare z-grids:\n");
printf("fwf.zmax=%f,wf->zmax=%f\n",file_wf.zmax,zmax);
printf("fwf.z0=%f,wf->z0=%f\n",file_wf.z0,z0);
printf("fwf.dz=%f,wf->dz=%f\n",file_wf.dz,dz);
printf("fwf.dzmin=%f,wf->dzmin=%f\n",file_wf.dzmin,dzmin);
#endif

	double fd = (file_wf.dzmin>0 ? file_wf.dzmin : file_wf.dz) ;
	double diff = (dzmin>0 ? dzmin : dz) ;
	if ( (file_wf.z0 > z0) || 
		 (file_wf.zmax < zmax) ||
		 (file_wf.dz > dz*wf::EPSDZ) || 
		 (fd > (diff*wf::EPSDZ)) )
		err = nr::ERZGR ;
	
	if (!err && (file_wf.naper != naper)) 
		err = nr::ERGEO ;

	/* Compare apertures geometry: */
#ifdef WEIFDEBUG
printf("wf->naper=%d, wf->diam=%5.2f,wf->eps_inn[3]=%f\n",naper,
diam,eps_inn[3]);
printf("fwf.naper=%d, fwf.diam=%5.2f,fwf.eps_inn[3]=%f\n",file_wf.naper,
file_wf.diam,file_wf.eps_inn[3]);
#endif

	if (!err)
		for (int i=0; i<naper; i++) {
			fd = eps_inn[i]*diam ;
			diff = file_wf.eps_inn[i]*file_wf.diam - fd ;
			if (fd>0) diff /= fd ;
		
			if (fabs(diff) > wf::EPSD) {
				err = nr::ERGEO ;
				break ;
			}

			fd = eps_out[i]*diam ;
			diff = file_wf.eps_out[i]*file_wf.diam - fd ;
			if (fd>0) diff /= fd ;

			if (fabs(diff) > wf::EPSD) {
				err = nr::ERGEO ;
				break ;
			}
		}
	
#ifdef WEIFDEBUG
printf("Check leff, lblue and lred and z-grids:\n");
printf("fwf.leff=%f,wf->leff=%f\n",file_wf.leff,leff);
printf("fwf.fwhm=%f,wf->fwhm=%f\n",file_wf.lred-file_wf.lblue,lred-lblue);
printf("fwf.zmax=%f,wf->zmax=%f\n",file_wf.zmax,zmax);
printf("fwf.z0=%f,wf->z0=%f\n",file_wf.z0,z0);
printf("fwf.dz=%f,wf->dz=%f\n",file_wf.dz,dz);
printf("fwf.dzmin=%f,wf->dzmin=%f\n",file_wf.dzmin,dzmin);
#endif
	if (!err) {
		if ( (fabs(file_wf.leff - leff) > wf::EPSLEFF) ||
		     (fabs(file_wf.lblue - lblue) > wf::EPSLEFF) ||
	    	 (fabs(file_wf.lred - lred) > wf::EPSLEFF))
			 err = nr::ERSPE ;
	}
	
	if (err) {
		string msg = string("WEIF: the header of ") + filename + 
			" does not fit: " + nr::ercodemessage(err) ;
		nr::nrerror(msg.c_str(), err) ;
	}

	if (!err) 
		/* Read weight */
		err = file_wf.read(filename) ;

	/* destructor of file_wf will be invoked on return */
#ifdef WEIFDEBUG
printf(" WEIF: checkfile(): err(out)=%d \n",nr::erget()); 
#endif

	return err ;
}

/*--------------------------- read -----------------------------------*/
int wf_t::read(const char * filename) {
	const double MAXDZMAX = 0.1 ;	/* max. diff. between header zmax and z[nz-1] */
	double weightscale = 0;

	int err = readhdr(filename, &weightscale) ;
	if (err<0) return err ;	/* message is already set: see nr::erget() */
	
	/* All the fields except for zgrid and ws are initialized now */
	
	/* Count zgrid steps: */
	FILE * f = fopen(filename,"rt") ;
	if (!f) {
		nr::nrerror("WEIF: read(): cannot reopen file",nr::ERFIO);
		return nr::ERFIO ;
	}
		
	/* Allocate the buffer for reading lines: */
	int nnw = sc::napcomb(naper) ;
	int buflen = (wf::WVALEN+1)*(nnw+1)+2 ;
	char * buff = new char[buflen] ;
	/*< maximal length of line */
	
	int nnz=0;
	while (fgets(buff,buflen,f)) {
		int l = strlen(buff) ;	/* Crude estimation of expected line length */
		if (char * id = strtok(buff, wf::FDELIM)) {
			if ((id[0]!='#') && (l>nnw)) {
				/* it is not comment and there are items: data. */
				nnz++;
			}
		}
	}
	fclose(f);

	/* Reopen the file and read: */
	if (!(f = fopen(filename,"rt"))) {
		nr::nrerror("WEIF: read(): cannot reopen file",nr::ERFIO);
		return nr::ERFIO ;
	}

	/* Allocate memory: */
	ws = nr::dmatrix(0,nnw-1,0,nnz-1) ;
	zgrid = nr::dvector(0,nnz-1) ;

	nnz = 0;
	while (fgets(buff,buflen,f)) { 
		int l = strlen(buff) ;
		char * id = strtok(buff, wf::FDELIM) ;
		if (id) {
			if ((id[0]!='#') && (l>nnw)) {
				/* it is not comment : try to read data now: */
				/* altitude: */
				char * endptr ;
				zgrid[nnz] = strtod(id, &endptr) ;
				if (!endptr || (strlen(endptr))) id = NULL ;
				for (int i=0; i<nnw; i++) {
					if (!id || (!(id = strtok(NULL, wf::FDELIM)))) {
						/* Reading error: */
						delete[] buff ;
						setaper(0);
						char ch[8];
						sprintf(ch, "%5.2f",zgrid[nnz]) ;
						string msg = string("WEIF: read(): problem for z=")+ch ;
						nr::nrerror(msg.c_str(), nr::ERFIO) ;
						setaper(0) ;
						return nr::ERFIO ;
					} else {
						ws[i][nnz] = strtod(id,&endptr) * weightscale;
						if (!endptr || (strlen(endptr))) id = NULL ;
					}
				}
				nnz++;
			}
		}
	}
	nz = nnz;
	fclose(f);
	delete[] buff ;

	/* Check that the file is not cut and weights (e.g. A) are not shifted */
	if ((zgrid[nnz-1]<zmax-MAXDZMAX) || ((zgrid[0]==0)&&(ws[0][0]!=0))){
		/* File is cut or weights are shifted! */
		string msg = string("WEIF: read(): cut or z-shifted file ")+filename ;
		nr::nrerror(msg.c_str(),nr::ERFIO) ;
		setaper(0);
		return nr::ERFIO ;
	} else
		return 0 ;
}

/*------------------------------- interpolate -----------------------------*/

inline int IMIN(int a, int b) { return (a < b ? a : b) ; }
inline int IMAX(int a, int b) { return (a > b ? a : b) ; }

void wf_t::interpolate(wf_t * refwf, double shift, 
	double * maxerr, int * iz_maxerr, int * iw_maxerr) {

	/* minimal number of nodes in old grid to be still able to interpolate */
	const  int MINNZ = 2 ;

	/* Check entries and parameters: new_wf zgrid should be enclosed in z-range 
	of the old wf grid! */
	if (!refwf || !refwf->ws || !refwf->nz || !refwf->getnw()) {
		nr::nrerror("WEIF: interpolate() with empty reference wf",nr::ERNUL) ;
		return ;
	}
	if (!nz) {
		nr::nrerror("WEIF: interpolate() with no zgrid set", nr::ERNUL);
		return ;
	}
	if (ws) {
		nr::free_dmatrix(ws,0,getnw()-1,0,nz-1) ;	/* Clean old weights in matrix and zgrid */
	}

	if ((refwf->nz<MINNZ) || (refwf->z0>z0) ||
	    (refwf->zmax<zmax)) {
		nr::nrerror("WEIF: interpolate() from narrower zgrid", nr::ERZGR);
		return ;
	}
	/* Allocate the new weight matrix with new z-dimention and reference nw: */
	int nw = refwf->getnw() ;
	ws = nr::dmatrix(0,nw-1,0,nz-1) ;
	/* Interpolate sequentially all weights: */
#ifdef POLINT
	/* Either polynomial or rational function interpolation: */
	if (maxerr) {
		*iz_maxerr = *iw_maxerr = -1;
		*maxerr =  -1 ;
	}
	int nodes = IMIN( refwf->nz, INTPOW ) ;
	for (int w = 0; w<nw; w++) {
		int jlo = 0 ;
		for (int iz = 0; iz<nz; iz++) {
			double z = fabs(zgrid[iz] - shift) ;
			int k = 1;
			if (nodes<nz) {
				/* Search for proper boundaries of nodes: */
				nr::hunt(refwf->zgrid-1, refwf->nz, z, &jlo );
				/* reassign k: */
				k = IMIN(IMAX(jlo-(nodes-1)/2,1), refwf->nz+1-nodes) ;
			} 
			double neww = 0, dneww = 0 ;
			nr::polint (&(refwf->zgrid[k-2]), &(refwf->ws[w][k-2]), nodes, z, &neww, &dneww) ;
			ws[w][iz] = neww ;
#ifdef INTRELERR
			dneww = ((neww==0) ? 0.0 : dneww / neww ;	/* relative error */
#endif
			if (maxerr) 
				if (dneww > *maxerr) {
					/* new maximum error found: */
					*maxerr = dneww ;
					*iz_maxerr = iz ;
					*iw_maxerr = w ;
				}
		}
	}
#else
	/* Spline interpolation : */
	/* allocate the second-derivatives array: */
	double * y2 = nr::dvector(0,refwf->nz-1) ;
	for (int w = 0; w<nw; w++) {
		/* Zero value of weight (not shifted) => zero (theoretic) deriv.: */
		double start_deriv = (refwf->ws[w][0] != 0) ? wf::DERNAT : wf::DER0 ;	

		nr::spline(refwf->zgrid-1,refwf->ws[w]-1, refwf->nz, start_deriv, wf::DERNAT, y2-1);

		for (int iz = 0; iz<nz; iz++) {
			double z = fabs(zgrid[iz] - shift) ;
			nr::splint(refwf->zgrid-1, refwf->ws[w]-1, y2-1, refwf->nz, 
				z, &ws[w][iz]);
		}
	}
	nr::free_dvector(y2,0,refwf->nz-1) ;
#endif	

	/* Interpolation is successful: assign fields: */
	leff = refwf->leff ;
	lblue = refwf->lblue ;
	lred = refwf->lred ;
	setaper(refwf->naper, refwf->diam, refwf->eps_inn, refwf->eps_out) ;
}
 
/*--------------------------- getspechar --------------------------------*/
/**
	@brief	Compute some characteristic points on the spectrum
	@param	lambda 	wavelength grid
	@param	edist	respective spectral energy distribution
	@param	nedist	length of \em lambda and \em edist grids
	@param	leff	Returned effective wavelength in [units of \em lambda]
	@param	lblue	Returned wavelength of the 50\% level of SED at blue side 
			from the SED's peak in [units of \em lambda]
	@param	lred	ditto at the red side from the SED's peak
	
	Function searchs for the maximum of \em edist[] computing simultaneously
	the gravity center of \em edist to be returned as \em leff. Then, it goes
	to the left and to the right from the maximum until the level of 50\% level
	of the maximal one is crossed by the curve. The \em lblue and \em lred
	points are computed by the local linear interpolation between the adjacent
	points.

	@see wf_leff()
*/	
void wf::getspechar(double * lambda, double * edist, int nedist, double * leff,
	double * lblue, double * lred) {
 
	if (!lambda || !edist || !leff || !lblue || !lred) {
		nr::nrerror("WEIF: getspechar(): null argument", nr::ERNUL) ;
		return ;
	}

	*leff = 0 ;
	double max = 0 ; 
	double norm = 0 ;
	int lmax = 0 ;
	for (int l=0; l<nedist; l++) {
		*leff += lambda[l] * edist[l] ;
		norm += edist[l];
		if (max<edist[l]) {
			lmax = l;
			max = edist[l];
		}
	}

	if (norm>0)
		*leff /= norm;
	else {
		nr::nrerror("WEIF: getspechar(): zero SED norm", nr::ERSPE ) ;
		return ;
	}

	/* 50%-level: */
	max /= 2 ;

	/* Left wing: */
	double x = 0 ;
	for (int l=lmax; l>0; l--) {
		if (edist[l]<max) {
			/* Crossing point is found: */
			x = lambda[l] + (lambda[l+1]-lambda[l]) * 
				(max - edist[l])/(edist[l+1]-edist[l]) ;
			break ;
		}
	}
	if (x==0) {
		x = lambda[0] ;
		nr::nrerror("WEIF: getspechar(): no blue-side slope in SED", nr::ERSPE) ;
		/* we still can estimate other characteristics, hense no return */
	}
	*lblue = x ;
 
	/* Right wing: */
	x = 0 ;
	for (int l=lmax; l<nedist; l++) {
		if (edist[l]<max) {
			/* Crossing point is found: */
			x = lambda[l-1] + (lambda[l]-lambda[l-1]) * 
				(edist[l-1]-max)/(edist[l-1]-edist[l]) ;
			break ;
		}
	}
	if (x==0) {
		x = lambda[nedist-1] ;
		nr::nrerror("WEIF: getspechar(): no red-side slope in SED", nr::ERSPE) ;
	}
	*lred = x ;
}

/*--------------------------- leff --------------------------------*/

int wf_t::calcleff(double * lambda, double * edist, int nedist, int fix) {
#define SEDDIFFERS (+1)
 
	/* cleaning mode: */
	if (!nedist) {
		leff = lblue = lred = 0 ; 
		return 0 ;
	}

	double nleff = 0,nlblue = 0,nlred = 0 ;
	wf::getspechar(lambda,edist,nedist, &nleff, &nlblue, &nlred) ;
	if (nr::erget()) return SEDDIFFERS ;	/* bad spectrum == different sp. */

	if ((fabs(leff-nleff)>wf::EPSLEFF) ||
		(fabs(lblue-nlblue)>wf::EPSLEFF) ||
		(fabs(lred-nlred)>wf::EPSLEFF)) {
		if (fix) {
			leff = nleff ;
			lblue = nlblue ;
			lred = nlred ;
		}
		return SEDDIFFERS ;
	} else 
		return 0 ;	/* Spectral parameters are equal within the given precision */
}

/*----------------------------- getval -------------------------------*/

double wf_t::getval(int iz, int iw) const {
	if ((iw<0) || (iw>=getnw()) || (iz<0) || (iz>=nz)) {
		nr::nrerror("WEIF: getval(): invalid weight or altitude index",nr::ERPAR) ;
		return 0 ;
	}
	return ws[iw][iz] ;
}

/*------------------------------ copy --------------------------------*/

void wf_t::copy(int lowiz, int hiiz, int iw, double * dest) const {

	if ((iw<0) || (iw>=getnw()) || (lowiz>hiiz) || (lowiz<0) || (hiiz>=nz)) { 
		nr::nrerror("WEIF: copy(): invalid weight index or altitude range",nr::ERPAR) ;
		return ;
	}
	if (!dest) { 
		nr::nrerror("WEIF: copy(): null destination array",nr::ERPAR) ;
		return ;
	}
	for (int i=lowiz; i<=hiiz; i++)
		dest[i] = ws[iw][i] ;
}

/*----------------------------- getalt ------------------------------*/

double wf_t::getalt(int iz) const {
	if ((iz<0) || (iz>=nz)) {
		nr::nrerror("WEIF: getalt(): invalid altitude index",nr::ERPAR) ;
		return 0 ;
	}
	return zgrid[iz] ;
}

/*==================== weight calculation section =====================*/

/* reopen namespace to put additional global variables there */
namespace wf {
/* Global private parameters needed to transfer the geometry of apertures into 
the function flama() to integrate for weight: */
double diameter, epsilon1, epsilon2, epsilon3, epsilon4, lambda0 ;
/*<< Last parameter is the SED origin in [mkm] (wavelength of the first point 
in a given SED) which is a shift of SED before doing its Fourier Tranform. */

/* Weight integration: global private arrays */
/* Complex FT of SED, respective frequency grid, its Re and Im parts,
their spline coefficients and length: */
double * edist_ft, * edist_freq, * re_edist_ft, * im_edist_ft, 
* re_edist_ftspln, * im_edist_ftspln;
int edist_ftnn ;

double altitude ;	/* Current altitude in [cm] */

}

/*------------------------------- make_fft --------------------------*/

/**
	@brief	Compute the FT of SED in units of [1/cm] 
	@param	lambda	wavelength grid
	@param	edist	light SED
	@param	nedist	w/l grid length
	@return	0 if Ok, <0 on error
	
	The Fast Fourrier Transform is used to compute the FT. 
	
	Before FFT, the SED is normalized by division of each value by the integral
	of SED. Then, the grid of SED is expanded four times with padding the obtained
	"tail" with zero values; SED values are divided by their argument
	wavelengths. Thus, the result of the FFT algorithm (four1()
	utility from NR is used) has a better sampling in the frequency domain
	which makes the spline interpolation of the spectrum more robust.
	
	Since the grid of SED starts at non-zero wavelength but immediately near the
	wavelength of the blue-end of the sensitivity curve of MASS, the FT is made
	on this origin-shifted spectrum. So, the correction for this shift is needed
	afterwards (see below).
	
	The units of the wavelengths are converted into [cm], the shift of the SED
	to its origin and the sampling of the wavelength grid are also expressed in
	[cm]. This ensures the correct normalization of the resulting FT and a
	proper correction of the real and imaginary parts of the FT for the made
	shift of SED to its origin. The latter correction is made in sfunc()
	analytically.

*/
void wf::make_fft(double * lambda, double * edist, int nedist) {

/* "Compute forward FFT": */
const int ISIGN = 1 ;
/* Number of stretching twice the F(lambda) grid to improve sampling of its FT:
 0 - no stretch, 1 - stretch twice, 2 - stretch 4-times etc. */
const int GRIDFACT = 2 ;

	/* The shift of the SED grid "lambda0" in [cm] (global wf::-var.) */
	wf::lambda0 = lambda[0]*MKM2CM ;

	/* Sampling of SED: */
	double dlambda = lambda[1]-lambda[0] ;
	if (dlambda==0) {
		nr::nrerror("WEIF: make_fft(): lambda-grid has zero step", nr::ERSPE) ;
		return ;
	}
	/* Normalize SED: */
	double enorm = 0 ;
	for (int l=0 ; l<nedist; l++) enorm += edist[l] ;
	if (enorm<=0) {
		nr::nrerror("WEIF: make_fft(): SED has non-positive norm", nr::ERSPE) ;
		return ;
	}
	
#ifdef WEIFDEBUG
	printf("SED norm = %f\n",enorm);
#endif
	/* normalize SED */
	for (int l=0; l<nedist;l++) edist[l] /= enorm ;

	/* Make the least number which is a power of two and not less than nedist:*/
	/* The length "nn" of the data grid for FFT is meanwhile 2**GRIDFACT 
	times more than this least number. */
	edist_ftnn = 1 << GRIDFACT ;
	for (int l=nedist; l!=0; edist_ftnn<<=1,l>>=1) {};

	/* Allocation of memory: */
	/* Make the complex-type SED-function array with the length like 2**N */
	wf::edist_ft = nr::dvector(0,edist_ftnn*2-1) ;
	/* Make the double-type array for frequencies of the FT of SED to be
		computed: */
	wf::edist_freq = nr::dvector(0,edist_ftnn-1) ;
	/* Make double-type arrays for Re- and Im-parts of the FT of SED: */
	wf::re_edist_ft = nr::dvector(0,edist_ftnn/2-1) ;
	wf::im_edist_ft = nr::dvector(0,edist_ftnn/2-1) ;
	/* Make double-type arrays for spline coefficients of Re- and Im-parts
	 of the FT of SED: */
	wf::re_edist_ftspln = nr::dvector(0,edist_ftnn/2-1) ;
	wf::im_edist_ftspln = nr::dvector(0,edist_ftnn/2-1) ;

	/* Fill in the data array for FFT: */
	enorm = 0 ;
	for (int l=0; l<edist_ftnn; l++) {
		/* Real part of the complex SED divided by the wavelength [cm]: */
		edist_ft[2*l] = ((l<nedist) ? edist[l]/(lambda[l]*MKM2CM) : 0) ;	
		/* Imaginary part is zero: */
		edist_ft[2*l+1] = 0 ;
		enorm += edist_ft[2*l]*edist_ft[2*l] ;
	}
#ifdef WEIFDEBUG
	printf(" Norm of SED/lambda (sum of squared vals) = %10.5e\n",enorm) ;
#endif

	/* Compute the Fourier transform of the origin-shifted SED: */
	nr::four1(edist_ft-1,edist_ftnn,ISIGN) ;
	
	/* Split Re- and Im-parts in the part of FT with positive frequencies and
	normalize them: */
	for (int l=0; l<edist_ftnn/2; l++) {
		re_edist_ft[l] = edist_ft[2*l] ;
		im_edist_ft[l] = edist_ft[2*l+1] ;
		edist_freq [l] = l / (dlambda*MKM2CM) / edist_ftnn ;
	}
	
	/* TEST: the Parseval theorem */
#ifdef WEIFDEBUG
	enorm = 0 ;
	for (int l=0; l<edist_ftnn; l++) {	
		enorm += edist_ft[2*l] * edist_ft[2*l] + edist_ft[2*l+1] * edist_ft[2*l+1] ;
	}
	printf(" Norm of FT (sum of squared vals/N) = %10.5e\n",enorm/edist_ftnn) ;
#endif	
	/* Interpolate the positive frequency Re- and Im-parts of FFT(F(lambda)) by 
	splines: */
	nr::spline(edist_freq-1, re_edist_ft-1, edist_ftnn/2, wf::DERNAT, wf::DERNAT, 
		re_edist_ftspln-1); 
	nr::spline(edist_freq-1, im_edist_ft-1, edist_ftnn/2, wf::DERNAT, wf::DERNAT, 
		im_edist_ftspln-1); 

}

/*-------------------------------- calc ---------------------------------*/

void wf_t::calc(double * lambda, double * edist, int nedist, void (*progress)(int)) {

	/* Do not compute anything if error is present */
	if (nr::erget()) return ;

	if (!zgrid || nz<=0) {
		nr::nrerror("WEIF: calc() on empty weight structure",nr::ERNUL) ;
		return ;
	}
	if (progress) progress(0) ;
	
	/* Compute the spline coefficients for the FT of the shifted SED: */
	wf::make_fft(lambda, edist, nedist) ;
	if (nr::erget()) return ;
	
	/* Allocate the weight matrix if still missing */
	if (!ws) { 
		int nw = sc::napcomb(naper) ;
		ws = nr::dmatrix(0,nw-1,0,nz-1) ;
	}

	if (progress) progress((int)(10./nz+0.5)) ;

	for (int i=0; i<nz; i++) {
		int nnw = 0;	/* Counter of weighting functions */	
		/* Normal weighting functions: */
		for (int a = 0; a < naper; a++) {
			ws[nnw][i] = weight(zgrid[i], diam, eps_out[a], eps_inn[a], 0, 0);
			if (nr::erget()) return ;
			nnw++;
		} 

		if (progress) progress((int)(100*((i + 0.3)/nz)+0.5)) ;

		/* Differential weighting functions: */
		for (int a = 0; a < naper - 1; a++) {
			for (int b = a + 1; b < naper; b++) {
				ws[nnw][i] = weight(zgrid[i], diam, eps_out[a], eps_inn[a],
				                  eps_out[b], eps_inn[b]);
				if (nr::erget()) return ;
				nnw++;
			}
		}
		if (progress) progress((int)(100*((i + 1.0)/nz)+0.5)) ;
	}

	/* Free FFT of SED allocated by wf_make_fft() : */
	nr::free_dvector(edist_ft,0,2*edist_ftnn-1) ;
	nr::free_dvector(edist_freq,0,edist_ftnn-1); 
	nr::free_dvector(re_edist_ft,0,edist_ftnn/2-1); 
	nr::free_dvector(im_edist_ft,0,edist_ftnn/2-1); 
	nr::free_dvector(re_edist_ftspln,0,edist_ftnn/2-1); 
	nr::free_dvector(im_edist_ftspln,0,edist_ftnn/2-1); 
	
	/* Call the SED's effective wavelength and 50% points computer at last */
	const int fixleff = 1;
	calcleff(lambda, edist, nedist, fixleff) ;
}

/***************************************************************************/
/*                  Private functions implementation                       */
/***************************************************************************/

/*------------------------------ zgrid_set() -------------------------------*/

/**
	@brief	Generate the grid of values given its generation parameters 
	@param	zgrid	pointer double-type array 
	@param	z0	starting value
	@param	zmax	value to reach 
	@param	dz	value modifier
	@param	dzmin	minimal value step
	@return Number of steps in z, -6 if NOT NULL zgrid pointer, -3 if bad params

	This function implements the generation of a grid of abscissa values used
	by wf_t::setzgrid() to generate the altitude grid in weight function
	structure. See weif.hxx for description of an algorithm given in
	wf::setzgrid() documentation. 

	The function counts the steps that will be computed, allocates the
	double-type array (to which the first parameter must point) and fills in
	it. The array must be freed after usage (free_dvector()).

*/
int wf::zgrid_set(double ** zgrid, double z0, double zmax, double dz, double dzmin)
{				 	
	if (*zgrid != NULL) return nr::ERNNL ;
	
	int nz = 0 ;
	
	/* We do this cheap work twice: first - compute the size of an array,
	and then fill in it */
	for (;;) {
		double z = z0, prevz = 0 ;
		int i = 0;	/* Counter of altitude steps, finally returned as the grid
			length */
		do {
			if (nz > 0) {
				/* Second cycle: assign to grid: */
				if (z<=zmax)
					(*zgrid)[i] = z ;
				else
					(*zgrid)[i] = zmax ;	/* last node */
			}
			double step = 0 ;
			if (dzmin==0) {
				/* Equidistant grid: */
				step = dz;
			} else {
				/* Proportional grid: */
				double step0 = z * dz;
				step = ( dzmin>step0 ) ? (step=dzmin) : (step=step0);
			}
			if (step==0) return nr::ERZGR ; /* can be only if z-params're bad */

			prevz = z;
			/* The z for the next turn: */
			z += step;
			i++;
		} while ((prevz<zmax) && (i<=wf::MAXNZ)) ;
		
		if ((i>wf::MAXNZ) || (i<=0)) return nr::ERZGR ;
		
		if (nz == 0) {
			/* first turn: allocate memory: */
			nz = i ;
			*zgrid = nr::dvector(0,nz-1) ;
		} else {
			return i ;
		}
	}
}
 
/*------------------------------- SFUNC ----------------------------------*/
/**
	@brief	Compute the imaginary part of the Fourier transform of the light SED
	@param	f	frequency in [1/cm] units 
	@return	Im(FT(SED))**2

	Private function used by flama().
		
	The results of the spline approximation of the real and imaginary parts  of
	FT of \f$\lambda_{min}\f$-shifted SED are used. Spline coefficients are
	already computed in make_fft() and put into \em re_edist_ftspln[] and
	\em im_edist_ftspln[] arrays. Re- and Im-parts  are taken from
	splines, corrected for the shift of SED to its origin (made while making
	FFT) by multiplication by \f$ exp(-2\pi i \lambda_{min}) \f$ and the square
	of the  imaginary part of the product is returned.

	Spline coefficients \em re/im_edist_ftspln[], computed FT grid \em
	re/im_edist_ft[] and the frequency grid \em edist_freq[] are the global
	(private) arrays with the length \em edist_ftnn.

	@see wf_t::calc()
*/
double wf::sfunc(double f) {

	/* High frequencies (beyond Nyquist) are cut off: */
	if (f>edist_freq[edist_ftnn/2-1]) return 0 ;

	/* Compute the real and imaginary parts of shifted F(lambda[mkm]): */
	double re_ft = 0, im_ft = 0 ;
	nr::splint(edist_freq-1, re_edist_ft-1, re_edist_ftspln-1, edist_ftnn/2, f,
	&re_ft);
	nr::splint(edist_freq-1, im_edist_ft-1, im_edist_ftspln-1, edist_ftnn/2, f,
	&im_ft);
	/* Correct the FT for shifting: */
	double re_shiftcor = cos(2*M_PI*f*lambda0) ;
	double im_shiftcor = sin(2*M_PI*f*lambda0) ;

	/* Multiply the numbers taking only the imaginary part: */
	double tmp = im_ft*re_shiftcor + re_ft*im_shiftcor ;
	return tmp*tmp ;
	
}

/*--------------------------- SDUM --------------------------------*/

/**
	@brief	Test-case of S-function: analytical for "quasi-gaussian" SED
	@param	f	frequency [1/cm]
	@return	S
	
	Private function which may replace sfunc() in flama() for test-calculations.

	In case of the SED being the "quasi-gaussian" function, the \em S-function
	is a purely analytical expression:
	\f[
	S = \sin^2(2\pi\lambda_0 f) exp(-4(\pi\sigma f^2)^2)/\lambda_0
	\f]
	for the SED which is a normalized \f$\lambda\f$-multiplied gaussian with 
	a \f$\sigma\f$-dispersion centered at \f$\lambda_0\f$.
	
	The substitution of sdum() instead of sfunc() in flama() may help to test
	the influence of \em "non-gaussianity" of the spectrum on the resulting
	shape of the weight function. The effect, in general, does not exceed a few
	percents (below 10 in any case) for the A0V star and the MASS response 
	function.
	
*/
double wf::sdum(double f) {
/* Put below, directly the FWHM and central wavelength of "quasi-gaussian" in
[mkm]: */
const double DELTA = 0.1 ;
const double LAMBDA0 = 0.5 ;
	double tmp = sin(2.0*M_PI*LAMBDA0*MKM2CM*f)/(LAMBDA0*MKM2CM) ;
	tmp *= tmp ;
	tmp *= exp(-7.11941*f*f*(DELTA*MKM2CM)*(DELTA*MKM2CM)) ;
	return tmp ;
}

/*------------------------------- FLAMA ----------------------------------*/

inline double SQR(double a) { return ((a)*(a)) ; }

/**
	@brief	Compute the under-integral expression for the weight integration
	@param	x	spatial frequency in inverse meters
	@return Under-integral expression (see doc)

	Private function used by weight().

	In polychromatic treatment, the weight is integrated from:
	\f[
	E = f^{-8/3} * S(z f^2 /2) * A(f) .
	\f]

	where the function \em S(f) is a \em squared integral of SED of incident 
	light with sine-modulation over wavelengths:
	
	\f[
	S(f) = \left [ \int\limits_{\lambda_{min}}^{\lambda_{max}}
		\sin(2\pi \lambda f) F(\lambda) d\lambda \right]^2
	\f]
	which is, in fact, the squared imaginary part of the Fourier Transform 
	of \f$ F(\lambda)/\lambda \f$ computed by sfunc().
	
	If SED is a \em delta-function centered at \f$ \lambda_0\f$, then the 
	expression turns into the known monochromatic one:

	\f[
	E = f^{-8/3} * \sin^2(z f^2 \lambda_0) * A(f) .
	\f]
   
	Here A(f) is a normalized aperture OTF for the given geometry of two
	annular apertures. For simplicity, the geometry parameters diameter,
	epsilon1..4 are  transferred via the global parameters; function \em
	S() is also the global (private) function sfunc().

*/
double wf::flama(double x) {

const double POWIND = -2.66666666667 ;

	double tmp = sfunc(altitude * x * x / 2);

	double apert = 1.0 ;	
	double apert1, apert2, apert3, apert4;
	if (wf::diameter>0) {
		apert  = x*wf::diameter*M_PI;	/* = 2pi*f*(r=D/2) */
		/* First aperture: */
		if (wf::epsilon1 == wf::epsilon2) 
			apert1 = 0.0;
		else {
			apert1 = epsilon1 * apert;
			apert1 = 2.0 * nr::bessj1(apert1) / apert;
			apert2 = epsilon2 * apert;
			apert2 = 2.0 * nr::bessj1(apert2) / apert;

			apert1 = (epsilon1 * apert1 - epsilon2 * apert2);
			apert1 = apert1 / ( SQR(epsilon1) - SQR(epsilon2) );
		}
		/* Second aperture: the same with indices 1->3, 2->4: */
		if (epsilon3 == epsilon4) 
			apert3 = 0.0;
		else {
			apert3 = epsilon3 * apert;
			apert3 = 2.0 * nr::bessj1(apert3) / apert;
			/* Normally, the circular apert. with zero central obscuration
			drops here in epsilon3>0, epsilon4=0: */
			if (epsilon4 == 0) {
				apert4 = 0;
			} else {
				apert4 = epsilon4 * apert;
				apert4 = 2.0 * nr::bessj1(apert4) / apert;
			}

			apert3 = (epsilon3 * apert3 - epsilon4 * apert4);
			apert3 = apert3 / ( SQR(epsilon3) - SQR(epsilon4) );
		}
		apert = SQR(apert1-apert3);
	} 
	tmp = tmp * pow(x,POWIND) * apert;
	return tmp ;
};

/*---------------------------  WEIGHT ----------------------------------*/

/**
	@brief	Compute the weight of a given \em z in a scintillation index
	@param	z	altitude [km]
	@param	d	reference size of an aperture [cm]
	@param	eps1	relative outer size of 1st (larger) aperture
	@param	eps2	relative inner size of 1st (larger) aperture
	@param	eps3	relative outer size of 2nd (smaller) aperture (may be 0)
	@param	eps4	relative inner size of 2nd (smaller) aperture (may be 0)
	@return Weight value in [m^(1/3)]

	Private	function used by wf_t::calc().
	
	Function returns the weight for an altitude \em z in the stellar
	scintillation. The FT of the incident light SED divided by \em lambda is
	assumed to be ready and accessible by the function sfunc(). The diameter of
	the largest entrance aperture is \em d and the radii of annular apertures
	are obtained by its multiplication by \em eps1..4.

	The first (say, larger) aperture has the outer diameter \em d*eps1 and the
	inner \em d*eps2. The second (smaller) aperture has the outer diameter \em
	d*eps3 and the inner \em d*eps4.

	If only two of four epsilons differ from each other (non-zero ring width)
	then the weighting function value for the  \em normal scintillation index
	is  returned for a given aperture (for the larger one - if eps1>eps2>0, 
	eps3=eps4; for the smaller one - if vise versa).

	If three or four epsilons are different (eps1>eps2 and eps3>eps4) then the
	weighting finction value for the \em differential scintillation index is
	returned for a pair of apertures with sizes \em (d*eps1,d*eps2) -- \em
	(d*eps3,d*eps4).

	The polychromatic expression for the weight is used here, which uses the FT
	of the light SED to compute the so called \em S-function sfunc(). This
	function stays for the part of the under-integral expression of the weight
	which is computed by flama(). And the latter is already the function
	which is integrated by the qromo() integrator with the \em relative
	convergence criterion. It integrates the weight for the range of spatial
	frequencies from \em X1=1e-4 to \em X2=18.

	The weight for the zero altitude is artificially set to zero.
*/
double wf::weight(double z, double d, 
              double eps1, double eps2, double eps3, double eps4) {

/* Define the range of frequencies to integrate the aperture OTF A(f) 
(X1=0.0001, X2=15 correspond to L_0=30m and l_0=5 mm at 10 km at 0.5mkm): */
const double X1 = 0.0001 ;
const double X2 = 18.0 ;

/* Relative precision for integration with modified qromo(): */
const double EPS = -1.E-8 ;
/* CALIBR = (32*PI^3*9.69*1E-3)*(0.01^(5/3))*1E4; where 0.01^(5/3) comes from
integration in [cm] instead of [m]; 1E4 originates from conversion of [1/cm]^2 
into [1/m]^2 in integration of sfunc() */
const double CALIBR  = 44.62621363 ;
/* Conversion of [km] into [cm]: */
const double KM2CM  = 1E5 ; 

	/* Assign the global parameters: */
	wf::altitude = z * KM2CM;	/* Altitude in [cm] */
	wf::diameter = d ;			/* This version : D is not normalized */
	wf::epsilon1 = eps1;
	wf::epsilon2 = eps2;
	wf::epsilon3 = eps3;
	wf::epsilon4 = eps4;

/* CHECK: printout all the aperture functions if "return tmp" is replaced with
"return apert" in flama(): */
/*	if (z==0) for (w=X1; w<=X2; w*=1.2) {
		printf("%8.5f %10.4e\n",w,flama(w)); }	return 0;*/

	if (fabs(z)<fabs(EPS)) {
		return 0 ;
	} else {
		/* Integrate the weight in units of [1/cm]: */
		double w = nr::qromo(flama, X1, X2, nr::midpnt, EPS);
		/* Calibrate: */
		if (!nr::erget())
			return w * CALIBR;
		else
			return 0 ;
	}
}

/*-------------------------- getzshift ------------------------*/

double wf::getzshift(double foclen_feed, double foclen_conj) {
const double MM2KM = 1e-6 ;
	return (- foclen_feed * foclen_feed / foclen_conj) * MM2KM ;
}

/*---------------------------------- checalc --------------------------------*/

namespace wf {

/** Units of optical range wavelength: 0 for [mkm], 3 for [nm], 4 for [\AA] */
const int MKM_RANGE	= 0	; 


/* File-reader for spectra and response curve */
void read2col(const char * filename, double ** col1,
	double ** col2, int * nrow) ;

void conwgrid(double * wgrid, int n, int scale) ;

void sedcrv(double * slambda, double * sed, int nsed,
	double * rlambda, double * response, int nresponse);

}

int wf::checalc(const char * sedfile, const char * responsefile, 
	const char * wfile, 
	int naper, double diam, double * eps_inn, double * eps_out,
	double uz0, double uzmax, double udz, double udzmin, 
	bool calcit,
	void (*progress)(int)) {

	int result = nr::erget() ;
	if (result) return result;

	double * rlambda = 0, * response = 0;
	int nresponse = 0;
	double * slambda = 0, * sed = 0 ;
	int nsed = 0 ;

	/* Read the response function from the file: */
	read2col(responsefile,&rlambda,&response,&nresponse) ;
	/* Convert whatsoever (nm or angstrems) into microns: */
	conwgrid(rlambda, nresponse, MKM_RANGE) ;
	if ((result=nr::erget())) goto ENDLAB;	/* returns if read2col() or conwgrid() fail */

	/* Check that spectrum that we read from file is Ok: */
	read2col(sedfile, &slambda, &sed, &nsed) ;
	/* Convert wavelengths into [mkm] */
	conwgrid(slambda, nsed, MKM_RANGE) ;
	if ((result=nr::erget())) goto ENDLAB;	/* returns if read2col() or conwgrid() fail */

	/* Check that response curve wavelength grid complies the spectrum and
	multiply them: */
	sedcrv(slambda, sed, nsed, rlambda, response, nresponse) ;
	if ((result=nr::erget())) goto ENDLAB;
	{	
		wf_t twf;
	
		double z0 = uzmax>0? uz0 : wf::Z0;
		double zmax = uzmax>0? uzmax : wf::ZMAX;
		double dz = uzmax>0? udz : wf::DZ;
		double dzmin = uzmax>0? udzmin : wf::DZMIN;
		/* Set the altitude grid and the aperture geometry in basic wf */	
		twf.setzgrid(z0, zmax, dz, dzmin) ;

		twf.setaper(naper, diam, eps_inn, eps_out) ;

		/* Compute and store the current spectrum characteristics in wf: */
		twf.calcleff(rlambda, sed, nresponse, wf::LEFFFIX) ;

		/* Check if the weight file wfile fits the current state of wf: */
		result = twf.checkfile(wfile) ;
		if (calcit && result) {
			if (result<nr::ERFATAL)
				nr::erreset() ;/* reset non-fatal error */
	
			if (result!=nr::ERZGR) {
				/* Problem was not in the altitude grid: compute with
				default optimal grid : */
				twf.clear() ;
				twf.setzgrid(wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN) ;
			} /* else - Leave the user's grid intact */
			/* Since sed is shifted to the origin of response curve, Nsed =
				nresponse : */
			if (!nr::erget())
				twf.calc(rlambda, sed, nresponse, progress) ;

			if (!(result=nr::erget()))
				/* Write the new weight file: */
				twf.write(wfile) ;
		}
	}
	ENDLAB:
	if (sed) nr::free_dvector(sed,0,nsed-1) ;
	if (slambda) nr::free_dvector(slambda,0,nsed-1) ;
	if (response) nr::free_dvector(response,0,nresponse-1) ;
	if (rlambda) nr::free_dvector(rlambda,0,nresponse-1) ;
	return result;
}

/*------------------------------ read2col --------------------------------*/
/**
	@brief	Read the two columns of argument and function values from file
	@param	filename	file name
	@param	col1	pointer to the array of arguments (first column data)
	@param	col2	pointer to the array of function values (second column)
	@param	nrow	pointer to the place where to put the length of allocated 
		arrays of arguments and function values
	
	This utility reads the pairs of numeric values from the text-type file
	ignoring the lines beginning with comment symbol "#". Two non-allocated
	pointers must be supplied as NULLs on start. Function counts the non-comment
	data lines, allocates the memory for arrays of argument (first column) and
	function values (second column) with lengths equal to the number of the 
	counted data lines, and then reads the columns into them.
	
	The following checks are done on each non-comment line:
	- at least two numeric tokens are given in each line (argument and function)
	- argument is a monotonously rising function
	- argument values start from some positive number
	- function is non-negative 
	- function has a positive average
	The function is thus suited for reading the spectrum-like information.
	
	Errors possible are the problem of file opening/reading or line parsing. In
	case of error, the pointers are returned unallocated.
*/	
void wf::read2col(const char * filename, double ** col1,
	double ** col2, int * nrow) {

	const int DATALINE = 100 ;      /* longest feasible line */

	if (nr::erget()) return ;

	/* Protect from allocation on allocated pointer */
	if ((*col1!=NULL) || (*col2!=NULL)) {
		nr::nrerror("ATMOS: read2col() on non-NULL storages", nr::ERNNL) ;
		return ;
	}

	if (!filename) {
		nr::nrerror("ATMOS: read2col() with no file name", nr::ERPAR) ;
		return ;
	}
	
	/* Count col1 steps: */
	FILE * f = fopen(filename,"rt") ;
	string msg = "ATMOS: read2col() with file " ;
	msg += filename ;
	if (!f)	{
OPENERR:
		msg += ": open error" ;
		nr::nrerror(msg.c_str(), nr::ERFIO) ;
		return ;
	}
		
	int ndata = 0;
	char buff[DATALINE];
	while (fgets(buff,DATALINE,f)) {
		char * id = strtok(buff, wf::FDELIM) ; 
		if (id && (id[0]!='#')) 
			/* it is not a comment and there is at least one item */
			ndata++;
	}
	fclose(f);
	
	if (!ndata) {
NODATA:
		msg += ": contains no data" ;
		nr::nrerror(msg.c_str(), nr::ERNOD) ;
		return ;
	};
	
	/* Reopen the file : */
	f = fopen(filename,"rt") ;
	if (!f) goto OPENERR;

	/* Allocate memory and, if fails, close the file and return: */
	(*col1)=nr::dvector(0,ndata-1) ;
	(*col2)=nr::dvector(0,ndata-1) ;
	(*nrow)=ndata ;

	int err = 0;
	int nline = 0 ;
	ndata = 0;
	double prevarg = 0, sum = 0 ;	/* previous argument and function \int storages */
	while (fgets(buff,DATALINE,f)) {
		nline++ ;
		char * id = strtok(buff, wf::FDELIM) ; 
		if (id && (id[0]!='#')) {
			/* it is not a comment : try to read data now: */

			/* argument: */
			char * endptr ;
			(*col1)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || !(id = strtok(NULL, wf::FDELIM)) || 
				((*col1)[ndata]<=prevarg)) {
				/* Argument token is not a number or 
				   previous argument is not less than current or
				   function value is absent: return line number */
				err = nline ;
				break ;
			}
			
			/* function: */
			(*col2)[ndata] = strtod(id, &endptr) ;
			if ((endptr==id) || ((*col2)[ndata]<0)) {
				/* Function token is not a number or
				   function is negative */
				err = nline ;
				break ;
			}
			sum += (*col2)[ndata] ;
			prevarg = (*col1)[ndata] ;
			ndata++;
		}
	}
	fclose(f);
	/* Check that the function is positive (not zero): */
	if (!err && (sum<=0)) goto NODATA ;
	if (err) {
		/* Reading or conversion or argument/function constraints error: */
		nr::free_dvector(*col1,0,*nrow-1);
		nr::free_dvector(*col2,0,*nrow-1);
		(*col1) = (*col2) = 0 ;
		sprintf(buff,": error in line %d",err) ;
		msg += buff ;
		nr::nrerror(msg.c_str(), nr::ERSPE) ;
		return ;
	} 
}

/*-------------------------------- conwgrid -------------------------------*/

/**
	@brief	Convert array of optical (!) wavelengths into given scale
	@param	wgrid	array of optical wavelengths in [angstrem] or [nm]
	@param	n	its length
	@param	scale	decimal logarithm of the wavelength of the infrared end of 
		the optical range to obtain
	
	This is an utility for conversion of the spectral curves (response function
	or spectrum) into the wavelength grid in a given units \em scale. It takes
	the value from the middle of the array and assumes that it must be in
	between 0.1 and 1 mkm if it's an optical range. Then it makes the conversion 
	by multiplication of the array value by an integer power of 10 which makes 
	the	middle array value lying in the range (0.1--1.0)*10**{scale}.
	
	\b Example
	
	Let the middle point of \em wgrid (it is \em wgrid[n/2]) be equal to 410 on
	start (i.e. the grid is in [nm]). \em Scale is given as "0", i.e. we want
	the spectrum to belong to the range 0.1--1.0, i.e. to be expressed in [mkm]
	on exit. The initial range is \em range=(int)lg(410)+1=3. So, the
	conversion coefficient will be \em 10**conv, where \em conv=scale-range=-3.
*/
void wf::conwgrid(double * wgrid, int n, int scale) {

	if (nr::erget()) return ;

	if ((wgrid==NULL) || (n<1)) {
		nr::nrerror("ATMOS: conwgrid() on null data", nr::ERNUL) ;
		return ;
	}
	
	double conv = scale - ((int)log10(wgrid[n/2])+1) ;
/* printf("Power conversion = %3.0f\n",conv);*/
	conv = pow(10.0,conv);
/*printf("Conversion = %7.1e\n",conv);*/
	for (int i=0; i<n; i++) wgrid[i] *= conv;
}


/*---------------------------- sedcrv ---------------------------------*/
/**
	@brief	Multiply the spectrum curve by the response curve with 
		optionally shifted origins
	@param	slambda	array of wavelengths of spectrum
	@param	sed  	array of spectrum; returned \b multiplied by
		response in first \a nresponse elements
	@param	nsed	length of \a sed[] and \a slambda[]
	@param	rlambda	array of wavelengths of response
	@param	response	array of response curve
	@param	nresponse	length of \a response[] and \a rlambda[]
	
	Function looks for the coincidence of some \em k-th element of \em 
	slambda[] grid with the first element of the response curve wavelength
	grid \em rlambda. This \em k is a shift (or difference) of origins of
	SED and response curves (normally, the spectrum is wider). 

	Then, first elements of spectrum - \em sed[i], i=0..nresponse - are
	rewritten with the product of SED and response curves 
	\em	sed[i+k]*response[i] which takes into account the different origins of
	these functions. Last \em nsed-nresponse elements of \em sed[] are replaced with
	zeros. Thus, on exit only first \em nresponse elements of \em sed[] make
	sense.
	
	@see calcwf() checksed()
*/
void wf::sedcrv(double * slambda, double * sed, int nsed,
	double * rlambda, double * response, int nresponse) {

	if (nr::erget()) return ;

	if (!slambda || !sed || (nsed<=0)) {
		nr::nrerror("WEIF: sedcrv(): null input data") ;
		return ;
	}
	
	/* Find the starting point of the response curve at SED wavelength 
		grid: */
	int k = -1 ;	/* Set "no match" */
	for (int j=0; j<=nsed-nresponse; j++) 
		if (rlambda[0]==slambda[j]) { 
			k = j ; break; 
		}

	/* Multiply the arrays, given their different origin */
	if (k>=0)
		for (int j=0; j<nresponse; j++)
			/* Check the match of grids: */
			if (fabs(slambda[j+k] - rlambda[j])>wf::EPSLEFF) {
				k = -1;
				break ;
			} else 
				/* k is not negative: no data loss by overwrite below */
				sed[j] = sed[j+k] * response[j] ;

	if (k<0) 
		nr::nrerror("WEIF: sedcrv(): no match of response and spectrum grids",
			nr::ERCRV) ;

}
