/*--------------------------------------------------------------------------*/
/** @file weif.hxx
                  MASS project: TURBINA module file                       
                       header file for weif.c                             
                                                                          
 This module contains the collection of utilities suited to compute the   
 altitude weighting functions for a given set of apertures for measurement
 of both normal and differential scintillation indices observed with      
 a certain bandpass*energy distribution of incident light.
 
 \attention Spectral Energy Distributions must be photon numbers related, i.e.
 in [photons/\AA], not in [erg/\AA]. Meanwhile, they still are denoted as SED
 for simplicity.
                                                                          
 This module is based on the trial program wfm2.c written by A.Tokovinin  
 and V. Kornilov (version Feb 26, 2001). The only bug found in that       
 program was incorrect memory freeing by the (modified) free_vector() function
 taken from Numerical Recipes (NR) package.                               
                                                                          
 The module contains the definition of the weighting function class wf_t
 and member functions for the weight calculation and input/output.
 The structure bears all the data and parameters which determine the weights
 and the respective altitude grid on which the weight matrix is computed.
 The parameters determine the z-grid, the number of altitude steps, the
 number of annular apertures with their sizes and effective wavelength. 

 \b Usage

 Prior to any operation, the constructed weight structure should be filled by
 wf_t::setaper(), wf_t::setzgrid() and wf_t::calcleff(). Being thus defined,
 then weight functions may be  either: 
 - read from the chosen file : the wf_t::checkfile() is used to to check the
 file  to coincide with the preset weight parameters (altitudes, wavelength, 
 apertures). If the file fits the parameters, the weight function may be  read
 from the file with by wf_t::read(). 
 - computed by wf_t::calc().

 If one gets known the new spectral energy distribution, the need to recompute
 the weight for this SED may be verified by wf_t::calcleff() with parameter \em
 fix==0. Also, the general checker/computer wf::checalc() is provided for
 the arbitrary input response function, SED and system geometry parameters.

 The parameters - the number of apertures, resulting weights (for apertures and
 their combinations) and number of altitude steps - can be accessed with
 wf_t::getnaper(), wf_t::getnw() and wf_t::getnz(), respectively. The \em name
 of the weight is obtained by the wf_t::name() function according to the column
 number in the weight matrix.

 Resulting weights may be shifted by altitude and/or recomputed for another
 altitude grid with wf_t::interpolate(). Weight matrix can be written in the
 disk file by wf_t::write() and released with wf_t::clear(). Full release of
 memory in weight structure is done with wf_t::setaper() with one zero
 parameter. The individual values of weight or altitude can be obtained with
 wf_t::getval() and  wf_t::getalt(); to copy some part of a certain weight
 function from the weight matrix in structure into a vector, use wf_t::copy().

 The testing main() is active in module compiled as a separate executable (set
 WEIFTEST macro to do so). If WEIFNTEST is set zero, the executable shall be a
 simple weight computer given the responce and spectrum functions. If it is
 positive, the full circle of calculations and interpolations is done the
 specified number of times (say, for performance checks and memory leaks
 detection).
 
 @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)               

 @version 1.5: Polychromatic precise weight, in C++                         
         
          1.6: Correction of a few non-significant handling bugs. New main()
		  for standalone weights calculations.
	  1.61: corrections in include-files for compatibility with gcc v3.3
*/
/* This header file is equipped with documentation made in comments in a
* format suitable fro production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _WEIF_H
#define _WEIF_H

#include "scind.hxx"
#include "nrutil.h"

/****************************************************************************/
/*                     Weight function structure                            */
/****************************************************************************/

/** 
This class contains all the data which describe the scintillation weighting 
functions of altitudes for all entrance apertures with which it is measured. 
The information on the apertures themselves and on the spectral characteristics
of the incident light is also present in members of this class.

The allocation of the memory for double-type arrays is made in several member
functions:
- setaper() - for allocation of geometry array parameters \em eps_inn[], 
\em eps_out[] ;
- setzgrid() - allocates the altitude grid \em zgrid[]; 
- calc() and read() - allocate the matrix of weights before its calculation or
reading, respectively.

Deallocation of all the memory allocated for arrays is done by setaper() with
one zero parameter (it is called also from the destructor).

The constructor initializes the arrays and members with zeros; in all functions
null pointer is a necessary demand for array to allow its allocation and, wise 
versa, if an array is not null, it is believed to be already allocated.
*/

class wf_t {
	
	double * zgrid ;			/**< array of altitudes: z-grid [km] */ 

	int nz ;					/**< number of z-steps (length of zgrid[]) */		

	double z0 ;  	/**< z-grid: starting altitude [km] */
	double zmax ;	/**< z-grid: maximal altitude [km] */
	double dz ;		/**< z-grid: modifier (absolute [km] or relative step) */
	double dzmin ; 	/**< z-grid: minimal step (0 for equidistant grid) [km] */
	
	double ** ws ;				/**< matrix of weights ws[aper][z] */

	int naper ;					/**< number of apertures */	
	/** reference size of apertures i.e. diameter for \em eps_out = 1 [cm] */
	double diam ;		

	double * eps_inn ;	/**< relative inner diameters of apertures (units of
					diam*eps_inn/out is [cm]) */
	double * eps_out ;	/**< relative outer diameters of apertures */

	double leff ;	/**< effective wavelength of weight func. [mkm] */
	
	double lblue ; 	/**< 50\% level wavelength at the blue side of SED [mkm] */
	double lred ;	/**< 50\% level wavelength at the red side of SED [mkm] */
	bool staticapert ;	/**< Flag to disable the copying the relative size 
				arrays but assign the pointers instead while doing setaper() */

	int readhdr(const char* file, double * weightscale) ;	
public:
	     
/****************************************************************************/
/*             Declarations of public member functions:                     */
/****************************************************************************/

/**	
	@brief	Initialize the object fields with zero/null values
*/
wf_t() ;

/**
	@brief	Release the z-grid and weight matrix and assigns them zero
	
	The signature of allocation of memory to \em zgrid and \em ws members of
	wf_t is non-zero z-dimention \em nz and non-zero respective pointers. Memory
	is released if only both \em nz and pointers are not zero.
	
*/
void clear() ;

/**
	@brief	Release all dynamic memory storages for deletion of object
	
	The feature of setaper() to release all the memory on call with a zero
	argument is used.
*/
~wf_t() { setaper(0) ; }

/**
	@brief	Assign the reference size and relative diameters for apertures
	@param	nap	number of annular apertures
	@param	diam	reference size or magnification factor
	@param	eps_in	array of inner diameters of apertures [cm]
	@param	eps_ou	array of outer diameters of apertures [cm]

	Procedure assigns the field \em naper to the number of apertures and
	allocates and fills in the arrays of their relative sizes. Input arrays \em
	eps_inn and \em eps_out are interpreted as:
	- physical instrumental sizes of apertures -- then diam plays the role of 
		the magnification of the system; 
	- or relative sizes (\f$eps\_...\le 1\f$) with respect to the outer diameter 
	of the largest aperture -- then diam is a size of a largest \em entrance 
	aperture. 
	
	In any case, only the values diam*eps_...  have a physical sense.

	\b Usage:
	
	\arg
	if nap<0, then the member arrays \em eps_inn and \em eps_out 
	are simply ASSIGNED to parameters \em eps_in and \em eps_ou,
	respectively. In this case they are not released afterwards (see naper==0
	case) -- the member flag \em staticapert is set "true" for this.
	
	\arg
	if nap>0, then input arrays are COPIED into newly \em allocated member 
	arrays \em eps_inn, \em eps_out. \em staticapert is set "false" to allow
	releasing of them in destructor.

	\arg
	If nap==0, then this function RELEASES all the memory allocated
	previously in dynamic memory for member arrays.
	
	@see	setzgrid()
*/
void setaper(int nap, double diam=0, double * eps_in=0, double * eps_ou=0) ;

/**
	@brief	Return the number of apertures set in structure
	@return	Number of apertures set in wf_t

*/
int	getnaper() const { return naper ; }

/**
	@brief	Return the number of apertures and their combinations
	@return	Number of weights which corresponds to the apertures number set
	    
	The function which returns the number of apertures and their combinations 
	with no difference in order is defined in scind.hxx: sc::napcomb().

*/
int	getnw() const { return sc::napcomb(naper) ; }

/**
	@brief	Give a name for respective weight 
	@param  index	sequential number of weight in weight matrix
	@return	pointer to the static const char* name, which is empty if index is 
		invalid
	    
	Name of the weight is a name of aperture or their combinations which
	corresponds to the given index. Thus, this function is just a synonym
	of sc::apername(). In case of a problem, the error message nr::ermessage()
	originates from the WEIF module.

*/
const char * name(int index) const {
	if ((index>=getnw()) || (index<0))
		/* Set the error which would be set otherwise by sc::apername */
		nr::nrerror("WEIF: name(): invalid weight index",nr::ERPAR);
	return sc::apername(naper,index) ;
}

/**
	@brief	Generate the altitude grid given its parameters 
	@param	z0	starting altitude [km]
	@param	zmax	altitude to reach [km]
	@param	dz	altitude modifier ([km] for equidistant grid, unitless for
	proportional)
	@param	dzmin	minimal altitude step [km]
	    
	Compute the grid of altitude steps for calculation of weighting
	functions like follows:
	
	The grid may be either equidistant (if \em dzmin ==0) or 
	proportional (\em dzmin !=0). The current restriction is 
	that the altitude modifier	should be positive.

	Equidistant grid is given as:
	@f[
 	z[i] = z0 + i * dz
	@f]
	where i is in [0,nz-1], nz conforms to the following condition: 
	@f[
	z[0] + dz  (nz-2) < zmax
	@f]
	and
	@f[
	z[0] + dz  (nz-1) \ge zmax
	@f]
	
	Proportional grid is constructed as follows:
	\verbatim
	z[0] = z0
	z[i] = z[i-1] + step
	step0 = z[i-1] * dz
	step = ( dzmin > step0 ) ? (step = dzmin) : (step = step0)
	\endverbatim
	where i is in [0,nz-1], nz conforms to the following condition: 
	@f[
 	z[nz-2] < zmax
	@f]
	and
	@f[
	z[nz-1]  \ge zmax .
	@f]
	In both types of grid, z[nz-1] is replaced with \em zmax.
	
	This function is an interface to the lower-level private function zgrid_set()
	which works directly with the array \em wf_t::zgrid.
	
	@see	setaper()	
*/
void setzgrid(double z0, double zmax, double dz, double dzmin);

/**
	@brief	Return the number of altitude steps in z-grid 
	@return	Number of z-steps set in object

*/
int getnz() const { return nz ; }

/**
	@brief	Writes weight matrix with parameters in disk ASCII file
	@param	filename	file name to write weight
	
	Procedure writes the matrix of weights which rows are preceded with the
	z-grid altitudes in the first column. The header is also created which
	contains all the parameters - the fields of wf_t.

	The format of the header line is following:

	\verbatim
	# <KEY> <VALUE1> [<VALUE2> [VALUE3 VALUE4]]
	\endverbatim

	where <KEY> is a single-letter designation of parameter and <VALUE> - its
	value(s). Keys are listed below with names of respective wf_t fields and
	description in brackets:

	\arg \c N --\em naper (number of apertures)
	\arg \c A -- \em D*eps_inn[0] \em D*eps_inn[0] (diameters of the 1st aperture)
	\arg \c B -- \em D*eps_inn[1] \em D*eps_inn[1] (diameters of the 2nd aperture)
	\arg ...
	\arg \c S -- \em leff \em lblue \em lred (effective wavelength and
	50\%-level points in [mkm])
	\arg \c Z -- \em z0 \em zmax \em dz \em dzmin (parameters for an altitude 
	grid generation)
	\arg \c W --\em weightscale (the scale-factor for the weight values 
	written in this file)
	
	After the header, the line goes which precedes the z-grid and wf-matrix
	values: 
	
	\verbatim
	##Altit. W(A)/1E+11 W(B)/1E+11 ... W(AB)/1E+11
	\endverbatim
	
	where 1E+11 is a number to which the values of weights are scaled in a
	file (\em weightscale).
	
	Any similar line which starts with "##" is interpreted by wf_t::read() as a
	comment.

	After the heading line "##Altit..." the z-grid and weight-matrix values go 
	being separated with spaces	for each altitude; thus -- one line per each 
	altitude.

*/
void write(const char * filename) const ;

/**
	@brief	Checks the weight file header values to conform the \em wf_t 
		parameters (fields)
	@param	filename	name of the file to read 
	@return	0 if parameters coincide, <0 if error or parameters do not coincide
		(the reason of incoincidence is reflected by following error codes: 
		nr::ERSPE, nr::ERGEO and nr::ERZGR).
	
	Procedure reads the header info in weight file (see description of the
	header in write() documentation) and checks whether the parameters given
	in the file header coincide (within some allowed errors) with respective
	fields of the supplied wf_t object. The latter can be, for example,
	initialized from some task-file (configuration file or something else). The
	aim is to get a clue to decide whether the weights for given parameters can
	be read from a given file or should be recomputed.
	
	The comparison is made until the first significant difference is found, in
	the following order: Z-grid, Apertures and Spectrum.
	
	The major source of the difference in weights is the spectrum of light and
	apertures geometry. Spectral characteristics in file (3 wavelengths, see 
	wf_t::calcleff()) must coincide to within EPSLEFF with given in \em wf; aperture
	diameters in wf_t and in file must differ not more than EPSD in a relative
	measure. 

	Another determinator is an altitude grid. The range boundaries in file 
	(\em z0 and	\em zmax) must cover from both sides the range in wf_t to be 
	allowed. Since the weights are normally well interpolated, the minimal
	spacing in altitude \em dzmin and altitude modifier \em dz in file do not
	have to coincide to that given in \em wf. Instead, they only have to be not
	more than EPSZ times larger than in \em wf_t. If the grid in the file is 
	thinner, it is accepted.

	If no error is found in header, the function read() is called on the
	checked file with a trial weight structure. After call this structure is
	cleared and resulting error (if any ) is returned.

	@see write() read()
*/	
int checkfile(const char * filename) const ;

/**
	@brief	Read the object -- weight matrix and its parameters -- from the 
		disk file
	@param	filename	name of file to read
	@return	0 if Ok, <0 if reading error
	
	Procedure reads the matrix of weights and its parameters from the disk file
	into the fields of object. Thus, parameters are modified
	compared to that assigned to the structure fields before. Before reading,
	the procedure will try to clear all the dynamic data if the pointers of 
	respective member arrays are not nulls.

	This function is meant to be used with files which were tested by
	checkfile() to contain exactly the  needed data. It does not check the
	validity of numbers in file. The weight matrix and z-grid data are converted
	from character representation to floating point numbers with standard
	function "strtod" and thus checked to be indeed numbers. Otherwise, the
	nr::ERFIO is returned and the altitude where the error in weight value 
	occured.
	
	\note
	The \em zgrid[] is read directly from the leftmost column of data section
	of the weight file and is NOT CHECKED to correspond to the given z-grid
	parameters \em z0,zmax,dz,dzmin parameters which are read from the header.

	@see	checkfile()
*/
int read(const char * filename) ;

/**
	@brief	Calculate the weight matrix for a given energy distribution
	@param	lambda	grid of wavelengths [mkm]
	@param	edist	incident energy distribution (given on lambda grid)
	@param	nedist	length of the \em lambda and \em edist grids
	@param	progress	function to call to visualize the work stage (if needed)
	    
	Given the parameters in numeric members of the object (aperture geometry)
	and input spectrum, the weights are computed on a filled \em zgrid (see
	setzgrid()).

	The weights are computed for the polychromatic case of incident light SED
	which involves the Fourier Transform of the SED divided by the wavelength.
	This FT is made by wf::make_fft() once before calculation of all weights
	and is represented in two global arrays as cubic spline coefficients. In
	order to have smoother shape, the SED is shifted to its origin (starting
	wavelength \f$\lambda_{min}\f$) to dump the high-frequency oscillations in
	FT. The back-correction for this shift is made analytically after the
	spline interpolation for the given frequency.

	Before calculations, the weight	matrix is allocated IF it was NULL.
	If it was not NULL, it is assumed that it was already allocated (e.g. in
	a previous call of this routine, say, with different en.dist.).
	
	Then, for all apertures, their combinations and for all altitudes in the
	z-grid, the function wf::weight() is called to make the weight integration
	itself.
	
	At the end, the SED characteristics computer wf_t::calcleff() is called to fill
	in the respective spectral fields. It appears that effective wavelength
	represents best the shape of weights if they are computed for the
	"quasi-gaussian" approximation of SED with its actual FWHM. Meanwhile, the
	deviations may be as high as 10\% from the exact shape, especially for
	differential indices of small apertures. 

	The utility calls the external function \em progress() (if supplied as
	non-null) to visualize the degree of the made work. This degree, in
	percents, is an integer-type argument changing from 0 in the beginning to
	100 at the end. Progress function is called twice per altitude step.

	\b Performance
	
	At the IBM P-III 667 MHz machine, the calculations of weights with 26
	proportional steps of altitude (up to 20 km) and for 4 apertures take \b 14
	seconds.
*/
void calc(double * lambda, double * edist, int nedist, void (*progress)(int));

/**
	@brief	Calculate the effective wavelength and 50\%-levels of the light 
		energy distribution and compares them with that set in member fields
	@param	lambda	grid of wavelengths [mkm]
	@param	edist	incident energy distribution (SED), given on \em lambda grid
	@param	nedist	number of nodes in \em lambda grid (points of en.dist.), 
		or 0 to	reset the spectral information in weight structure
	@param	fix  	If \em fix==1: assign the computed spectral characteristics
		to wf_t fields, or leave them otherwise (use wf::LEFFFIX or 
		wf::LEFFTEST)
	@return int(SED characteristics differ from that set in the	structure) or
		the (negative) error code
	
	The effective wavelength and the wavelengths at which the relative light
	energy is equal to 50\% of the maximal are computed by the local linear
	interpolation for the given incident light SED. It is assumed that an array
	of relative intensities edist[] takes into account the instrument and
	atmospheric transmission functions. The computed effective wavelength and
	50\% points are first compared to the values in the fields \em leff, \em
	lblue and \em lred, and then assigned to them, if \em fix==1 and (!) the
	difference \em is significant.

	The result of comparison is returned: 0 if the difference is not
	significant, i.e. all three calculated numbers are within +/- \em
	EPSLEFF interval from the field values, and 1 if at least one of them
	differs by more than \em EPSLEFF. All wavelengths are in [mkm].

	If the parameter \nedist is supplied as 0, the calculations and comparison
	are skipped and the fields \em leff, \em lblue and \em lred are reset to 
	zero. 0 is returned.

*/
int calcleff(double * lambda, double * edist, int nedist, int fix);

/**
	@brief	Interpolate the weights from the reference object to the z-grid 
		set in the object with a given optional z-shift
	@param	refwf	wf_t to get the weights from
	@param	shift	the shift of weight function along z-grid [km]
	@param	maxerr	pointer where to put the maximal weight error due to interp 
		(works in POLINT option of compilation; may be null).
	@param	iz_maxerr	pointer to put the z-grid index of this error (may be
		null)
	@param	iw_maxerr	pointer to put the weight index of this error  (may be
		null)
	    
	This function uses the NR recipe for interpolation of weight values from
	one z-grid into another. Input weight structure wf_t must be fully
	determined; output structure must only contain the computed new z-grid: i.e.
	only the setzgrid() function must be called for it before interpolation.
	Other parameters must be empty - they will be assigned from the reference 
	wf_t fields.

	The parameter \em shift allows one to obtain the shifted weighting
	functions suited to work with the indices measured with the shifted pupil of
	the system (defocusing of the instrument). The \em negative altitudes
	when probed during the interpolation of the positively shifted functions are
	replaced with their absolute values. When negatively shifted function is 
	stretched to reach the same \em zmax, the extrapolation case takes place.

	The rest parameters are three pointers to variables in which the user may
	receive the estimated maximal interpolation error of the weight, and the
	indices in z-grid and weight number where this maximal error was found. For
	example, for 4-aperture weight matrix (10 weights for each altitude),
	iz_maxerr = 2 and iw_maxerr = 5 will mean that maximal error \em maxerr was
	found at the third stage of the NEW altitude grid \em zgrid[2] and for the
	weight of AC-aperture combination.

*/
void interpolate(wf_t * refwf, double shift, 
	double * maxerr = 0, int * iz_maxerr = 0, int * iw_maxerr = 0);

/**
	Return the member \em leff value 
*/
double getleff() const { return leff ;}

/**
	Return the member \em lblue value 
*/
double getblue() const { return lblue ; }

/**
	Return the member \em lred value 
*/
double getred() const { return lred ; }

/**
	@brief	Get the weight value for given weight and altitude indices
	@param	iz	altitude grid index, must be in range [0..getnz(wf)-1]
	@param	iw	weight index, must be in range [0..getnw(wf)-1]
	@return	Weight value (non-negative) 

	@see copy()
*/
double getval(int iz, int iw) const ;

/**
	@brief	Copy (part of) a certain weight function to double-type array
	@param	lowiz	low limit of the altitude index (see getnz())
	@param	hiiz	upper limit of the altitude index (see getnz())
	@param	iw	weight index, must be in range [0..getnw(wf)-1]
	@param	dest	double-type array
	
	This function copies one of available rows of the weight matrix into the
	pre-allocated destination array.

	\attention 
	In case of unallocated \em dest vector, the segmentation fault
	failure occurs! It is only checked to be non-null.
*/
void copy(int lowiz, int hiiz, int iw, double * dest) const ;

/**
	@brief	Get the altitude grid value for a given altitude index
	@param	iz	altitude grid index, must be in range [0..getnz(wf)-1]
	@return	Altitude value (non-negative, in [km]) 
*/
double getalt(int iz) const ;

} ;	/* of the class wf_t */

/** 
	@brief 
	Additional to wf_t, this namespace wf contains two supplementary 
	functions - getzshift() and the sample progress indicator progress(). Also, 
	all the module constants are defined here.
*/
namespace wf {

/**
	@brief	Compute altitude shift in generalized mode from parameters of optics
	@param	foclen_feed	Focal length of feeding optics, [mm]
	@param	foclen_conj	Focal length of conjugating lens, [mm]
	@return	zshift	Altitude shift in generalized mode, [km]

	Function computes the shift from the ratio of squared focal length of
	feeding optics and focal length of conjugating lens. Normally, the positive
	lens (with a positive \a foclen_conj) is used for conjugation to negative
	altitudes i.e. working in a generalized mode. That's why the result is
	negative for positive parameters.
	
	\note
	May depend on defocus, to be modified.
*/
double getzshift(double foclen_feed, double foclen_conj) ;

/**
	@brief	Sample progress indicator for console
	@param	stage	Part of the "work-done" (0..100) [percents]
	@return	void
	
	Example progress indicator, to be replaced with a suitable one 
*/
void progress(int stage) ;

/**
	@brief	Check the weight file to correspond to given system and spectral
		parameters. May recalculate on mismatch
	@param	sedfile	Stellar spectrum energy distribution file
	@param	response	System response curve file
	@param	wfile	Weight file to check and/or put the results of calculations
	@param	naper	Number of entrance apertures in scintillation device
   	@param  diam    see wf_t::setaper()
   	@param  eps_inn see wf_t::setaper()
   	@param  eps_out see wf_t::setaper()
	@param	z0	altitude grid starting point (see wf_t::setzgrid())
	@param	zmax	maximal altitude in weight matrix
	@param	dz	altitude step
	@param	dzmin	minimal step for propostional grid (0 for equidistant)
	@param	calcit	False: only check the file, True: recalculate if does not
		fit
	@param  progress	function to call to visualize the work stage (if needed)
	@result	the comparison result returned by wf_t::checkfile() or error code
		
	This function makes use of all the functionalities presented by the class
	wf_t to compute the weight functions set for a system of a given geometry
	and a certain spectrum of incident light (composed by the star's SED and the
	system response function).
	
	Via parameters, all the needed information on the geometry and spectrum is
	provided. The altitude grid, if left default (all values are 0), is taken
	as given by the parameters wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN.
	
	If some mismatch is found between the weight in file \a wfile and the
	configuration specified by the parameters, the result "false" is returned
	in case \a calcit=false.
	The particular reason of a mismatch is provided by the call to
	nr::erget(), nr::ermessage(). Don't forget to reset then the error before
	continuing with nr::erreset(). 
	
	If a mismatch is found and \a calcit is "true", the weights are recomputed 
	in the file \a wfile. "True" is returned normally (the result of
	wf_t::calc() actually).
	
	\note In case of recalculation, the "optimal" grid of altitudes specified by
	the parameters wf::Z0, wf::ZMAX, wf::DZ, wf::DZMIN is taken for calculations
	unless the mismatch between the file and parameters was indeed the grids
	mismatch. In latter case, the \a z0, \a zmax, \a dz and \a dzmin parameters
	are taken instead.
*/
int checalc(const char * sedfile, const char * response, const char * wfile, 
	int naper, double diam, double * eps_inn, double * eps_out,
	double z0 = 0, double zmax = 0, double dz = 0, double dzmin = 0, 
	bool calcit = false, void (*progress)(int) = 0) ;


/***************************************************************************/
/*                  Constant definitions                                   */
/***************************************************************************/

/** Maximal size of aperture [cm] (arbitrary but plausible) */
const double MAX_DIAM = 1000 ;

/** Maximal number of altitude steps */
const int MAXNZ	= 10000 ;

/** Writing weights to the file : the Scale factor to divide the values by */
/* ATTENTION! Should always be purely logarithmic al: 1E10 or 1E11 or... */
const double WSCALE	= 1E11 ;

/** Longest plausible weight value record in a file (normal format: %10.4E) */
const int WVALEN  = 15 ;

/** Usual value for low boundary of altitude range */
const double Z0 = 0 ;

/** Maximal feasible value for upper boundary of altitude range. Since
"altitude" in weight functions is a length along line-of-sight in atmospheric
calculations, this ZMAX=40 [km] stays for an altitude of 20 km at airmass=2. */
const double ZMAX = 40 ;

/** Optimal altitude modifier for z-grid creation in proportional mode */
const double DZ = 0.2 ;

/** Optimal minimal altitude step for z-grid creation in proportional mode */
const double DZMIN  = 0.04 ;

/** Relative tolerance in aperture sizes to accept the weight file */
const double EPSD = 2E-2 ;

/** Tolerance level in effective wavelength and SED's 50\% points [mkm]. This
value corresponds to change of these wavelengths which causes the change of
weights not more than 1--2\%. */
const double EPSLEFF = 0.003 ;

/** Flag for wf_t::calcleff() to fix the computed spectral characteristics after
comparison */
enum LEFFTYPE {
LEFFTEST=0, /**< do not fix calculated Lblue, Lred, Leff in wf_t, only compare */
LEFFFIX=1   /**< fix wavelengths in wf_t \em after comparison */
} ;

/** Relative tolerance in the altitude grid spacing: maximal acceptable ratio of
\em dz and \em dzmin in file and in the comparison structure, >1 */
const double EPSDZ = 1.5 ;

/** Length of lines in a header part of the weight file */
const int LTEXT = 80 ;

/* Weight interpolation: type of the function - polynomial or rational func. */
/* Use polint() recipe from NR: */
//#define POLINT 	
/* ATTENTION! If no POLINT set, the 
  SPLINE interpolation is used. As far as we know, its the BEST choice,
  at least with INTPOW=6. */

#ifdef POLINT
/** Maximal power of interpolation polynomial (in case of defined POLINT) */
const int INTPOW  = 6 ;

/** Locate maximal relative error in weight interpolation rather than absolute: */
/*#define INTRELERR*/
#endif

/* Weight file delimiters: */
const char FDELIM[] = " \t\n\r" ;

/* Error codes and messages are all collected in SCIND */

}	/* of namespace wf */ 

#endif
