/*
	MASS project module: SUMMARY
	See the interface description in the header file summary.h.
*/
#include "summary.h"
#include "atmos.hxx"
#include "scind.hxx"
#include <stdio.h>
#include <map>
#include <vector>
#include "nr.h"
#include <algorithm>
#include <iostream>
#include <fstream>
#include <strstream>

//added 22.09.2003 to access standard C++ types string and map:
using namespace std;

#ifdef SUMMARYTEST
int main(int argc, const char * argv[]) {
	if (argc<3) {
		cout << "SUMMARY TEST: make/append summary-file summary.lst and a dump sumdump.bin\n";
		cout << "Usage: "<<argv[0]<<" <date as DDMMYY> <Nmeas to simulate>\n";
		exit(1);
	}
	char * endptr = 0;
	long date = strtol(argv[1],&endptr,10);
	const int NCHAN = 4;
	int NMEAS = strtol(argv[2],&endptr,10) ;
	if (NMEAS<0) NMEAS = 20;
	sum::init("summary.lst",date,NCHAN,"sumdump.bin");
	if (nr::erget()) {
		cerr << nr::ermessage() << '\n';
		if (nr::erget()!=nr::ERFIO)
			return -1 ;
		else
			nr::erreset();
	}
	for (int meas=0; meas<NMEAS; meas++) {
		double UT = sum::getn() / 15.0 ;
		const double vmag = 1.1 ;
		const double b_v = 0.57 ;
		double airmass = 1.0 + meas / 100.0 ;
		bool isgen = ((meas+1)%10==0) ;
		sum::add(UT,sum::massmag(vmag,b_v),airmass,isgen) ;
	}
	cout << "Enter par.name:";
	string name;
	cin >> name;
	double refl = sum::getref(name,sum::LASTMEAN);
	double reff = sum::getref(name,sum::FIRSTMEAN);
	double refs = sum::getref(name,sum::SEASONMEAN);
	const double * data = sum::getdata(name);
	const double * absc = sum::getdata("UT");
	if (nr::erget()) {
		cerr << nr::ermessage() << '\n';
		nr::erreset();
	} else {
		cout << "Ref(L)=" << refl <<'\n';
		cout << "Ref(F)=" << reff <<'\n';
		cout << "Ref(S)=" << refs <<'\n';
		int ndat = sum::getn();
		if (name=="CN2" || name=="ZCN2") {
			int nmeas = NMEAS-1;
			cout << "Meas. number [0-" << sum::getn()-1 <<"]: " ;
			cin >> nmeas;
			if (nmeas<0 || nmeas>=sum::getn()) nmeas = sum::getn()-1;
			data = sum::getdata(name,nmeas);
			absc = sum::getdata("ZCN2",nmeas);
			ndat = (int)sum::getdata("NCN2")[nmeas];
		}
		for (int m=0; m<ndat; m++)
			cout << absc[m] << ":  " << data[m] << '\n';
	}
	sum::done();
	if (nr::erget())
		cerr << nr::ermessage() << '\n';
}
#endif

namespace sum {
	/** Types of data stored continuously during the measurements; not saved in
	a summary-file are the last items starting from NCN2 */
	enum partype {UT, AIRM, FLUX, SI, DESI, AINT, NCN2, ZCN2, CN2, UNKNOWN} ;

	// information which goes first in a summary file line:
	const char DATENAME[] = "DATE	" ; // 3 spaces for matching the date rec.:
	const char DATEFMT[] = "%02d-%02d-%02d" ;
	const char NOBSNAME[] = "Nob" ;	// number of normal mode measurements
	const char NGENNAME[] = "Ng " ; // number of normal mode measurements
	const char NOBSFMT[] = "%s%3d" ;	// number of total/gen. mode measurements
	const char UTNAME[] = "UT" ;
	const char UTFMT[] = "%s%02d:%02d" ;   // HH:MM, with a leading delimiter
	const char AIRMFMT[] = "%s%4.2f" ;	// 1.23 with a leading delimiter

	int nchannel = 0 ;	// number of registered MASS channels
	int nparam = 0 ;	// total number of parameters registered
	int nobs = 0 ;
	int ngen = 0 ;
	long curdate = 0 ;		// current date as YYMMDD
	double curtime = 0; // current fractional UT hour
	double curairmass = 1;	// current airmass
	string sumfile; // Summary file name
	string dumpfile; // Dump-file name
	string header;	// summary file header
	double flux_conversion = 1; // 0-mag/current star fluxes ratio

/**
	@brief	Get "standard" name for a parameter type
*/
inline const char * partype_name(partype t) {
	// Header line names of columns (without refixes/suffixes): 
	switch (t) {
		case UT:   return UTNAME ;	  // suffixes: _ST, _EN
		case AIRM: return "AIRM" ;
		case FLUX: return "FLUX" ;	// suffixes: _A .. _D
		case SI:   return "SI" ;	// suffixes: _A.._D,_AB.._CD
		case DESI: return "DESI" ;	// suffixes: _A .. _D
		case AINT: return "" ;		// to be determined externally
		case NCN2: return "NCN2" ;
		case ZCN2: return "ZCN2" ;
		case CN2:  return "CN2" ;
		case UNKNOWN: return "";
	}
	return "";
}

// ================= PARAM_T CLASS DECLARATION =================
	struct param_t {
		static const int startlength = 50 ;	// starting length of a data array
		partype type ;	// kind of result
		string format ;	  // for I/O operations with stdio
		string name ;	// name of parameter to give in a summary file heading
		union { // specification
			int i ; // for fluxes and indices
			atm::ordint apar ;	// for atmospheric parameters
		} which ;
		bool isgen ;	// generalized mode parameter
		int ndata ; // number of filled entries 
		int length ;	// number of allocated space
		double ref[sum::LASTREF] ;
		int nref ;	// number of reference results averaged in SEASONMEAN
		union {
			double * p ;	// for non-[Z]CN2 types: all measurements as one row
			double ** pp ;	// for [Z]CN2-types: profile [altitudes] pointers
		} data ;
		/** 
			@brief: Constructor: zero the fields
		*/
		param_t() ;
		/**
			@brief	Associate the structure with a particular output parameter
			@param	t	Type of parameter
			@param	i	Detailed specification within the type (index number
				or atmospheric integral parameter kind)
			@param	isgen	Shifted pupil parameter flag

			Assigns the output C-format and allocates the data array with a 
			starting length.
		*/
		void assign(partype t, int i = 0, bool isgen = false) ;
		/**
			@brief Access the necessary function to get a new data value
		*/
		void add (bool isgen) ;
		/**
			@brief Get the pointer to data array
			@param	nres	Number of result for type== \c [Z]CN2
			@return data array or the needed profile/altitudes array
		*/
		double * getdata(int nres = -1) ;
		/**
			@brief	Compute and return the median over the data array
			\attention	The data array is modified! (sorted).
			This function must be used after dumping of the series.
		*/
		double getmedian() ;
		/**
			@brief	(Re)allocate the data array
		*/
		void alloc(int length) ;
		/**
			@brief: Destructor: release memory
		*/
		~param_t() ;
	} ;

// ================= PARAM_T CLASS IMPLEMENTATION =================

inline void debug(const char * s, bool addpref = true) {
#ifdef SUMMARYDEBUG
	if (addpref)
		printf("SUMMARY: ") ;
	printf("%s",s);
#endif
}

std::map<string,param_t*> parmap ;
std::vector<param_t> paramset ;

// Constructor:
param_t::param_t() {
	type = UNKNOWN ;
	ndata = length = nref = 0 ;
	data.p = 0 ;
	isgen = false ;
	format = "" ;
	for (int i=0; i<sum::LASTREF; i++) ref[i] = sum::NODATA ;
	debug("Constructor\n");
}

inline void ctoupper(char& c) {c = toupper(c);}

// Associate with a certain parameter
void param_t::assign(partype t, int i, bool shifted) {
	type = t ;
	name = partype_name(type) ;
	isgen = shifted ;
	switch (t) {
	case UT: 
		format = UTFMT ; 
	break ;
	case AIRM: 
		format = AIRMFMT ; 
	break ;
	case FLUX: 
	{
		format = sc::FLXFMT ; 
		which.i = i; 
		string aps = string("_") + sc::apername(sum::nchannel,i);
		string shortname = string("F")+aps;
		// put the name like this in a map, in addition to the final version
		// put at the end (for all types):
		sum::parmap[shortname] = this ;
		name += aps ;
	}
	break ;
	case SI: case DESI:
	{
		format = sc::IDXFMT ; 
		which.i = i ;
		string aps = string("_") + sc::apername(sum::nchannel,i) ;
		if (isgen) aps += "S" ; // generalized mode parameter
		// now "aps" is like _A[B][s]
		name += aps ;
		if (i>=sum::nchannel) {
			// put the name like this in a map, in addition to the final version
			// put at the end (for all types):
			sum::parmap[name] = this ;
			// differential index name:
			name = string("D")+name ;
		}
	}
	break ;
	case AINT: 
		format = atm::getfmt(atm::ordint(i)) ; 
		which.apar = atm::ordint(i) ; 
		name = atm::getname(which.apar);
		// convert into upper case:
		for_each(name.begin(),name.end(),ctoupper);
		if (name[0]=='W') {
			// alternative name of whole-atmoshere parameters:
			string shortname = name.substr(1);
			sum::parmap[shortname] = this ;
		}

	break ;
	case CN2: case NCN2: case ZCN2: 
		format = "" ; 
	break;
	default:
		nr::nrerror("SUMMARY: param_t::assign(): unknown type",nr::ERPAR);
		return ;
	}
	// Allocate the data array with a starting length:
//	cout << "SUMMARY: assign "<<name<<'\n';
	alloc(startlength) ;

	// Put a name in the map of names:
	sum::parmap[name.c_str()] = this ;
}

// Add the datum
void param_t::add (bool shifted) {
	if (length<=ndata) alloc(length+startlength) ;
	if ((type==FLUX) || (type==SI) || (type==DESI))
		if (!shifted && isgen) {
			data.p[ndata++] = sum::NODATA ;
			return ;
		}
	switch (type) {
		case UT:
			data.p[ndata++] = curtime; return ;
		case AIRM:
			data.p[ndata++] = curairmass; return ;
		case FLUX: 
				data.p[ndata++] = 
#ifdef SUMMARYTEST
				type*100+which.i*10+ndata ;
#else
				sc::getavgflux(isgen,which.i) * flux_conversion ; 
#endif
			return ;
		case SI: 
				data.p[ndata++] = 
#ifdef SUMMARYTEST
				0.0001*(type*100+which.i*10+isgen*5+ndata) ;
#else
				sc::getavgidx(sc::DSI,isgen,which.i) ; 
#endif
			return ;
		case DESI: 
				data.p[ndata++] = 
#ifdef SUMMARYTEST
				0.0001*(type*100+which.i*10+isgen*5+ndata) ;
#else
				sc::getavgidx(sc::DESI,isgen,which.i) ; 
#endif
			return ;
		case AINT: 
			data.p[ndata++] = 
#ifdef SUMMARYTEST
				type*100+which.apar*10+ndata ;
			if (format=="%s%5.2f") data.p[ndata-1]/=100;
#else
				atm::getval(which.apar); 
#endif
			return ;
		case NCN2: 
			data.p[ndata++] = 
#ifdef SUMMARYTEST
				5+shifted ; 
#else
				atm::getncn2() ; 
#endif
			return ;
		case ZCN2: case CN2: 
		{
#ifdef SUMMARYTEST
			const int NLAYER = 6 ;
			int n = NLAYER - !shifted ; 
#else
			int n = atm::getncn2() ; 
#endif
			if (n>0) {
				data.pp[ndata] = new double[n] ;
				if (type==CN2)
#ifdef SUMMARYTEST
				{
					const double prof0[NLAYER] = {1.4e-13,5.6e-14,8.6e-16,1.1e-14,2.3e-14,7.0e-14} ;
					static int idum = -13; // random seed
					const int corrtime = 10 ;	// measurements correlation	length
					const double statnoise = 0.1 * sqrt(corrtime);	// noise of CN2 values
					for (int i=0; i<n; i++) {
						double noise = exp(nr::gasdev(&idum)*statnoise) ;
						double newcn2 = prof0[i+!shifted] * noise;
						int start = ndata+1>=corrtime ? ndata-corrtime+1 : 0 ;
						int navg = 1;
						for (int j = start; j<ndata; j++) {
							newcn2 += data.pp[j][i];
							navg++;
						}
						data.pp[ndata][i] = newcn2/navg ;
					}
				} else
					for (int i=0; i<n; i++) 
						data.pp[ndata][i] = pow(2,i-shifted)-0.5*(i==0 && shifted);
#else
					for (int i=0; i<n; i++) data.pp[ndata][i] = atm::getcn2(i) ;
				else
					for (int i=0; i<n; i++) data.pp[ndata][i] = atm::getzcn2(i);
				debug("(Z)CN2 added\n",false);
#endif
			} else {
				debug("(Z)CN2=0 added\n",false);
				data.pp[ndata] = 0 ;
			}
			ndata ++ ;
			return ;	// otherwise the pointer data.pp[ndata] is rewritten
		}
		default: 
			nr::nrerror("SUMMARY: add(): unknown parameter type");
	}
}

// (Re)allocate memory:
void param_t::alloc(int len) {
	if (length>=len) return ;
	if ((type==CN2) || (type==ZCN2)) {
		double ** pp = new (double*)[len] ;
		for (int i=0; i<ndata; i++) pp[i] = data.pp[i] ;
		for (int i=ndata; i<len; i++) pp[i] = 0 ;
		delete[] data.pp ;
		data.pp = pp ;
//		debug("Allocate pp\n");
	} else {
		double * p = new double[len] ;
		if (ndata) 
			for (int i=0; i<ndata; i++) p[i] = data.p[i] ;
		delete[] data.p ;
		data.p = p ;
//		debug("Allocate p\n");
	}
	length = len ;
}

double * param_t::getdata(int nres) {
	if ((type==CN2) || (type==ZCN2)) {
		int i = (nres>=0 && nres<ndata) ? nres : ndata - 1 ;
		return data.pp[i] ;
	} else 
		return data.p ;
}
	
double param_t::getmedian() {
	if ((ndata>0) && (int(type)<int(NCN2))) {
		if (ndata>1)
			nr::sort(ndata,data.p-1);
		// locate NODATA at one of the tails of the sorted sequence:
		int ibeg = 0 ;
		while ((data.p[ibeg]==sum::NODATA) && (ibeg<ndata-1)) ibeg++;
		int iend = ndata-1 ;
		while ((data.p[iend]==sum::NODATA) && (iend>=ibeg)) iend--;
		if (iend<ibeg) return sum::NODATA ;
		int half = (ibeg+iend) / 2 ;
		if (((iend-ibeg+1)%2) || (half==iend))
			return data.p[half] ;
		else
			return 0.5*(data.p[half]+data.p[half+1]) ;
	} else
		return sum::NODATA ;
}

// Destructor
param_t::~param_t() {
//	cout << "Type = " << int(type) << '\n';
	if ((type==CN2) || (type==ZCN2)) {
		if (!data.pp) return ;
		for (int i=0; i<ndata; i++) 
			if (data.pp[i]) delete[] data.pp[i] ;
		delete[] data.pp ;
//		debug("destructor pp\n");
	} else {
		if (!data.p) return ;
		delete[] data.p ;
//		debug("destructor p\n");
	}
}

void printpair(const pair<string, param_t*>& r);

int createset(int nchan);

const int descrlen = 4; // length of a dump-file descriptor (in long-int numbers)

}	// of reopened namespace sum

void sum::printpair(const pair<string, param_t*>& r) {
	cout << "Type: " << r.second->type << ' ' << r.first << '\n';
}

/**
	@brief	Create the set of series with properly assigned parameters
	@param	nchan	number of MASS channels
	@return 0 on success, nr::erget() on error (nr::nrerror() is set)
*/
inline int sum::createset(int nchan) {
	const int NSHIFTED = 2 ;	// number of apertures to save/store the shifted indices
	nparam = 0 ;
	nchannel = nchan ;
	// total number of elements:
	int nelement = int(UNKNOWN) + 
		// repetitions in FLUX, DESI and shifted DESI
		(nchannel-1)*2 + NSHIFTED +
		// repetitions in SI:
		sc::napcomb(nchannel)+NSHIFTED+1 - 1 +
		// repetitions in AINT:
		atm::LAST-1 ;
	// allocate vector:
	paramset.resize(nelement) ;
	
	for (int p = 0; p<int(UNKNOWN); p++) {
		partype type = partype(p) ;
		switch (type) {
			case UT: case AIRM: case NCN2: case ZCN2: case CN2:
				if (nparam>=int(paramset.size())) {
					nr::nrerror("SUMMARY: createset(): wrong parameter set length",
						nr::ERPAR);
					return nr::ERPAR;
				}
				// assign it:
				paramset[nparam].assign(type);
				nparam++;
				break ;
			case FLUX: case DESI: case SI: 
				{
				// number of parameters of a current type:
				int nap = (type==SI) ? sc::napcomb(nchannel) : nchannel ;
				if (nparam+nap>int(paramset.size())) {
					nr::nrerror("SUMMARY: init(): wrong parameter set length",
						nr::ERPAR);
					return nr::ERPAR;
				}
				for (int ap=0; ap<nap; ap++) {
						
					paramset[nparam].assign(type,ap) ;
					nparam++;

					// for indices, add parameters for 2 shifted innermost aps.:
					if ((type!=FLUX) && 
						((ap<NSHIFTED) || (ap==nchannel))) {

						paramset[nparam].assign(type,ap,true) ;
						nparam++;
					}
				}
				}
				break ;
			case AINT:
				{
				if (nparam+atm::LAST>int(paramset.size())) {
					nr::nrerror("SUMMARY: init(): wrong parameter set length",
						nr::ERPAR);
					return nr::ERPAR;
				}
				for (int i=0; i<atm::LAST; i++) {
					paramset[nparam].assign(type,i) ;
					nparam++;
				}
				}
				break ;
			default: 
				nr::nrerror("SUMMARY: createset(): unexpected parameter type",
					nr::ERPAR);
				return nr::ERPAR;
		}
	}
	// create the header line for a summary file:
	header = string("#")+DATENAME+SUMDLM+UTNAME+"_ST"+SUMDLM+UTNAME+"_EN";
	header += string(SUMDLM)+NOBSNAME+SUMDLM+NGENNAME+SUMDLM;
	for (int i = UT+1; i<nparam; i++) {
		if (paramset[i].type>=NCN2) break;
		// create extra-spaces to fill the length of a generated number
		// according to its format:
		size_t lenpos = paramset[i].format.rfind('%')+1;
		char * endptr ;
		int numlen = strtol(&(paramset[i].format.c_str()[lenpos]), &endptr,10) ;
		int numspace = numlen - paramset[i].name.length() ;
//		cout << "nspace="<<numspace<<' ';
		if (numspace<0) {
			int nold = numlen<10?1:2;	// number of digits in an old length
			// expand the format and zero number of spaces:
			numlen -= numspace ;	// numspace is <0, hense - enlarge numlen
//			cout << "=>"<< numlen<<' ' ;
			char cnumlen[3];
			sprintf(cnumlen,"%d",numlen);
			paramset[i].format.replace(lenpos,nold,cnumlen);
			numspace = 0;
		}
		string spaces = " ";
		spaces.replace(0,1,numspace,' ');
//		cout << paramset[i].name << " : " << paramset[i].format << ": " << numspace << "spaces =	|"<<spaces<<"|\n";
		header += paramset[i].name + spaces + SUMDLM ;
	}
//	cout << header << '\n';
	return 0;
}

void sum::init(const char * sumfilename, long date, int nchan, 
	const char * dumpfilename) {

	curdate = date ;

	debug("init():\n");
	// Create the parameter series: the sequence of them in paramset[]
	// is a sequence of writing the new summary file
	if (sum::createset(nchan)) return ;
//	std::for_each(parmap.begin(),parmap.end(), printpair);
	
	// Read the summary file:
	sumfile = sumfilename ? sumfilename : "" ;
	std::ifstream fsum(sumfilename) ;	// open input stream
	if (fsum) {	// IO error check 
		// open successfully. Read the header:
		string fline;
		getline(fsum, fline) ;
		// compare the "standard" part of a header (there might be something user's at the end)
		int lencmp = header.find_last_not_of(SUMDLM) ;	// index of last significant char
		if (fline.substr(0,lencmp)==header.substr(0,lencmp)) {
			// Ok: the standard part of a header is valid. Start reading the results:
			string sdate, sut_st, sut_en;
			int nn, ng ;
			debug(" summary-file...",false);

			while (!fsum.eof() && getline(fsum,fline) ) {
				istrstream ist(fline.c_str()) ;	// string input stream
				ist >> sdate >> sut_st >> sut_en >> nn >> ng ;
				char * endptr ;
				long indate = (sdate.size()==8) ? strtol(
					(sdate.substr(0,2)+sdate.substr(3,2)+sdate.substr(6,2)).c_str()
					, &endptr, 10) : 0 ;
//cout << "Date/UT1/UT2/Nn/Ng= "<<sdate<<'='<<indate<<' '<<sut_st<<' '<<sut_en<<' '<<nn<<' '<<ng<<'\n';				
				if (indate!=curdate && nn>1) {
					// input all the results now:
					for (int i=UT+1; i<nparam; i++) {
						if (paramset[i].type>=NCN2) break;
						double res;
						ist >> res ; 
						if (res!=sum::NODATA) {
							paramset[i].ref[LASTMEAN] = res;
							int n = paramset[i].nref ;
							if (!n && paramset[i].ref[FIRSTMEAN]==sum::NODATA) {
								// first result (may be the manual input in 1st line with date=0
								paramset[i].ref[FIRSTMEAN] = res;
							}
							if (indate) {
								// not a manual input in 1st line: may be accumulated in SEASONMEAN:
								paramset[i].ref[SEASONMEAN] = 
									(paramset[i].ref[SEASONMEAN]*n + res)/(n+1);
								paramset[i].nref++ ;
							}
						}
					}
				} 
			}
		} else {
			// headers are not compatible:
			nr::nrerror("SUMMARY: header is obsolete. No reference data	retrieved",
				nr::ERNOD);
		}
	} else {
		// File does not exist?
		if (sumfilename) 
			nr::nrerror("SUMMARY: cannot read summary-file", nr::ERFIO);
		// else: no file needed
	}

	// Read the dump-file:
	FILE * f = 0;
	if (dumpfilename) {
		f = fopen(dumpfilename,"rb") ;
		dumpfile = dumpfilename ;
	} else
		dumpfile = "";
	if (f) {
		long descr[descrlen] ;
		size_t size = sizeof(long);
		if (!fread(descr,size,descrlen,f)) {
			nr::nrerror("SUMMARY: error while reading the dump-file",nr::ERFIO);
		} else {		 
//			cout << "READ: "<<descr[0] << '-' << descr[1] << '-' << descr[2] << '-' <<descr[3] << '\n';
			if ((descr[0]==curdate) && (descr[1]==nparam) && (descr[2]>0)) {

				debug(" dump-file...\n",false);
				// dump-file is valid and not out-dated: import the data now:
				size_t size = sizeof(double);
				// memory amount to allocate for each series: some startlength-blocks 
				// to get all the data read 
				int space = (descr[2]/param_t::startlength+1)*param_t::startlength ;
//cout << "space="<<space<<'\n';				
				int ifailed = -1 ;	// index of a failed series (series: >=0)
				for (int p=0; p<nparam; p++) {
					paramset[p].alloc(space) ;
					paramset[p].ndata = descr[2] ;
					if ((paramset[p].type==ZCN2) || (paramset[p].type==CN2)) {
						for (int m=0; m<descr[2];m++) {
							double * data = 
								paramset[p].data.pp[m] = new double[descr[3]] ;
							// read (z)cn2[]:
							if (!fread(data,size,descr[3],f) && descr[3]) {
								ifailed = p;
								break ;// exit this loop to be caught below	again
							}
						}
					} else {
						double * data = paramset[p].getdata() ;
						// read series:
						if (!fread(data,size,descr[2],f))
							ifailed = p;
					}
					if (ifailed>=0) {
						nr::nrerror("SUMMARY: dump-file is corrupted; data not restored",
							nr::ERFIO) ;
						break ;
					}
				}
				if (ifailed>=0) {
					// reset the read data: deallocate (Z)CN2[] and	ndata=>0:
					for (int p=0; p<=ifailed; p++) {
						if ((paramset[p].type==ZCN2) || (paramset[p].type==CN2)) {
							int m=0;
							while (m<=descr[3] && paramset[p].data.pp[m]) {
								delete[] paramset[p].data.pp[m] ;
								m++;
							}
						} 
						paramset[p].ndata=0;
					}
				} else {
					// compute the number of shifted and non-shifted measurements:
					double * data = parmap["SI_A"]->getdata();
					nobs = 0;
					for (int i=0; i<getn(); i++) if (data[i]!=sum::NODATA) nobs++;
					data = parmap["SI_AS"]->getdata();
					ngen = 0;
					for (int i=0; i<getn(); i++) if (data[i]!=sum::NODATA) ngen++;
//					cout << "NOBS="<<nobs<<" NGEN="<<ngen<<'\n';
				}
			}
		}
		fclose(f);
	}
	debug("Ok\n",false);

}

void sum::done() {

	debug("done():");

	// write the dump-file: while the series are not spoiled by getmedian():
	FILE * f ;
	if (dumpfile.length() && (nobs>0) && (f = fopen(dumpfile.c_str(),"wb"))) {
		const int descrlen = 4;
		long descr[descrlen] ;
		descr[0] = curdate;
		descr[1] = nparam;
		descr[2] = getn();
		descr[3] = 0;
		for (int m=0; m<getn(); m++) {
			int n = (int)((parmap["NCN2"]->getdata())[m]);
			if (n>descr[3]) descr[3] = n ;
		}
//		cout << descr[0] << '-' << descr[1] << '-' << descr[2] << '-' <<descr[3] << '\n';
		size_t size = sizeof(long);
		if (!fwrite(descr,size,descrlen,f)) {
			nr::nrerror("SUMMARY: error while writing the dump-file",nr::ERFIO);
		} else {
			debug("dump-file:\n");
			size_t size = sizeof(double);
			for (int p=0; p<descr[1]; p++) {
				debug((paramset[p].name+'\n').c_str());
				partype type = paramset[p].type ;
				if ((type==ZCN2) || (type==CN2)) {
					for (int m=0; m<descr[2];m++) {
						double * data = paramset[p].getdata(m) ;
						if (data) 
							fwrite(data,size,descr[3],f);// write (z)cn2[]
						else
							fwrite(descr,size,descr[3],f);// write foo
					}
				} else {
					double * data = paramset[p].getdata() ;
					fwrite(data,size,descr[2],f);// write series
				}

			}
			fclose(f);
		}
	}

	// write the summary file: first open as appendable text file and then
	// reopen as binary to add one line: at the end if no current date line is
	// is found and just after the current date record if found. 
	if (sumfile.length() && (nobs>0)) {
		debug("summary-file():");
		if (!(f = fopen(sumfile.c_str(),"a+t"))) {
			nr::nrerror((string("SUMMARY: cannot write the file ")+
				sumfile).c_str(), nr::ERFIO);
			return ;
		}
		
		// produce the date record:
		const int maxdatelen = 20;	// accounts for crazy dates also
		char sdate[maxdatelen+1] ;
		int d = curdate/10000;
		int m = (curdate-d*10000)/100;
		int y = curdate%100;
		sprintf(sdate,DATEFMT,d,m,y);
		int datelen = strlen(sdate);
		bool founddate = false ;	// flag of the current date record found
		if (!ftell(f)) {
			// new file:
			fprintf(f,"%s\n",header.c_str());
		} 
		// search for the current-date string:
		rewind(f);
		char buf[maxdatelen+1];
		while (fgets(buf,maxdatelen,f) && 
			!(founddate=(strncmp(buf,sdate,datelen)==0))) {}
		long curpos = ftell(f);
		if (founddate) curpos -= (maxdatelen-1) ;
//cout << buf <<" <-- " << curpos << "\n";

		// reopen for inserting/appending the current line:
		if (!freopen(sumfile.c_str(),"r+",f)) {
			nr::nrerror("SUMMARY: cannot append the results",nr::ERFIO);
			return ;
		}
		
		// create the results line:
		string toadd;
		char sres[500];	// buffer for sprintf
		toadd += sdate ;

		// start time:
		double utime = (paramset[0].getdata())[0];
		int h = int(utime);
		m = int((utime-h)*60);
		sprintf(sres,UTFMT,SUMDLM,h,m);
		toadd += sres ;

		utime = (paramset[0].getdata())[getn()-1];
		h = int(utime);
		m = int((utime-h)*60);
		sprintf(sres,UTFMT,SUMDLM,h,m);
		toadd += sres ;
		
		// number of non-shifted and shifted pupil observations:
		sprintf(sres,NOBSFMT,SUMDLM,nobs);
		toadd += sres ;
		sprintf(sres,NOBSFMT,SUMDLM,ngen);
		toadd += sres ;
		// write all data:
		for (int p=0;p<nparam;p++) {
			if (paramset[p].type==UT) continue ;	// already written
			if (paramset[p].type==NCN2) break;
			debug((paramset[p].name+'\n').c_str());
			// compute the median:
			double median = paramset[p].getmedian() ;
			const char * fmt = paramset[p].format.c_str() ;
			if (paramset[p].type==FLUX) {
				int prec = sc::fluxprec(median) ;
				sprintf(sres,fmt,SUMDLM,prec,median) ;				
			} else 
				sprintf(sres,fmt,SUMDLM,median) ;
			toadd += sres ;
		}
		toadd += "\n";
		fseek(f,curpos,SEEK_SET);
		long a = ftell(f);
//		cout << "Before writing: "<<a<<'\n';
		fwrite(toadd.c_str(),sizeof(char),toadd.size(),f);
		a = ftell(f);
//		cout << "After writing " << toadd.size() << " bytes: "<<a<<'\n';
		fclose(f);
	} 
	nparam = nchannel = nobs = ngen = 0 ;
	sumfile = "";
	parmap.clear();
	paramset.clear();
	debug("Ok.\n");
}

void sum::add(double UTh, double mag, double airmass, bool isgen) {
	debug("add()\n");
	curtime = UTh ;
	flux_conversion = pow(10.,0.4*mag);
	curairmass = airmass ;
	for (int p = 0; p<nparam; p++) paramset[p].add(isgen) ;
	nobs ++ ;
	if (isgen) ngen++ ;
}

double sum::getref(const string& pname, sum::reftype rtype) {
	param_t * p = parmap[pname] ;
	if (!p) {
		nr::nrerror((string("SUMMARY: get reference on non-existent parameter ")
			+pname).c_str(),nr::ERNOD) ;
		return sum::NODATA ;
	} else {
		return p->ref[rtype] ;
	}
}

const double * sum::getdata(const string& pname, int nres) {
	param_t * p = parmap[pname] ;
	if (!p) {
		nr::nrerror((string("SUMMARY: get data on non-existent parameter ")
			+pname).c_str(),nr::ERNOD) ;
		return parmap["UT"]->getdata() ;
	} else {
		return p->getdata(nres) ;
	}
}

int sum::getn() {
	if (paramset.size() && (paramset[0].type==UT))
		return paramset[0].ndata ;
	else
		return 0 ;
}
