/**		@file summary.h
@brief	Collection of scintillation indices and atmospheric parameters results
from scientific modules for their graphic representation and the summary-file
handling.

	The module is aimed to collect the output results (averages over \em 
	Accumulation Time) of scintillation	 measurements and atmospheric 
	calculations during the night and update the so-called \em summary-file	 of
	MASS with their night-time median parameters. In this file, the set of 
	these averages is given as a single line per program start (thus, 
	normally, per night) with some reference and statistic supplementary 
	information (the evening date, UT time span covered, number of obtained 
	and medianed results). The format of a summary-file is described in MASS 
	User Guide.

	An access to the result storages and references with sum::getdata() and 
	sum::getref() and while parsing the header of a summary-file in 
	sum::init() is done via their character names (e.g. "SI_A" for an A-
	channel scintillation index). The relation between the names (both given 
	in the head-line of a summary-file and supplied as a parameter in  getdata()
	etc.) and the respective result structures is installed via an	internal
	\em associative \em map.

	In addition, the results of previous nights are read from this summary-file
	to have \em reference while graphically presenting the series of the
	current night results. The continuously accumulated series of results serve
	both for the end-of-night median averaging in order to update the summary
	file and for providing the arrays of points for graphic representation
	during the night.

	In the hierarchy of TURBINA modules, this module thus stays above the
	scientific modules ATMOS and SCIND. It "knows" which scintillation and
	atmospheric parameters to access and how to write them in file or read from
	file (the same formats are used as in the \em mass-file).

	\b Usage
	
	Before starting the measurements, the module has to be initialized by
	sum::init() with a name of a summary-file and a number of the MASS device
	channels. The summary-file is read (if exists) and the \em reference value
	for each parameter is stored - the ones from the first and last
	summary-file lines ("First mean" and the "Last mean") and an average
	through all records in a file ("Season mean"). These reference values
	are accessible wit sum::getref(). Also, the "dump-file" name of
	accumulated series may be supplied	to restore the content of the series if
	TURBINA was restarted in the same night.

	Upon completion of each Accumulation Time of scientific measurements
	(Normal or Generalized mode), the function sum::add() has to be invoked to
	add the parameters in tails of their series. After this, the series
	graphics	may be updated using sum::getdata(); their length are taken from
	sum::getn().

	Before exit of TURBINA, the function sum::done() must be called. If some
	scientific measurements were performed, it computes the night-time median 
	parameters from the series.  The summary file is created with a header line
	if it did not exist; the name is taken as that supplied to sum::init(). If
	this is the first start in the current line, a new line of results is added
	in a summary-file; if the record with the current date is already present
	in the file, it is overwritten. 

	\attention
	Since the summary-file is overwritten in case of a repetitive TURBINA run,
	a loss of information is possible (cause is, e.g., overfill of a numeric
	format leading to discrepant sizes of the old and new records). So, the
	summary-file is better to save somewhere in the beginning of a new night.

	If a "dump-file" name was supplied to sum::init(), the series are written in
	a binary format in this file. Before quitting, the series storages are
	released.

	@author N.Shatsky (kolja@sai.msu.ru)
	@version	1.0 Creation
*/
#ifndef _SUMMARY_H
#define _SUMMARY_H
#include <string>
/**
	@brief
	Handling the scientific measurement results for the night-time summary and
	graphics
*/
namespace sum {

/** The value reserved for the unknown parameter value. Returned on failure; 
don't enter the averaging if encountered in a set. */
const double NODATA = 0 ;

/** Types of the reference information available: */
enum reftype { 
	FIRSTMEAN, /**< The value of the parameter taken from the \em first record 
	in the summary-file (e.g. manually written site averages) */
	LASTMEAN, /**< The value of the parameter taken from the \em last record in 
	the summary-file */
	SEASONMEAN, /**< The \em mean value of the parameter through all the lines
	in the summary-file (zero-date lines, e.g. with a word instead of date, 
	are averaged) */
	LASTREF /* reference enumeration "guard" */
} ;

/**
	@brief	Read the summary file, make references and initialize the
		series storages
	@param	sumfile MASS summary file name
	@param	date	Current date represented as DDMMYY
	@param	nchan	MASS channels number
	@param	dumpfile binary file name to restore the series content from
	
	The \a sumfile is read (if exists) with checking the compatibility of its
	header line with the data set expected. If not readable, the message is
	returned with error code nr::ERFIO. If the header is not compatible,
	the file content is rejected and no reference is made for parameters; error
	message is set with code nr::ERNOD. 
	
	If file is Ok, all the result lines are read and the mean value of all 
	columns is computed as a \em Season \em mean values. The first and last
	line	contents are saved separately.	These kinds of the reference data
	are available with getref(). First results line does not participate in
	averaging if it contains the invalid	(zero) date. Thus, the manual
	provision of a first line for the sum::FIRSTMEAN	reference type (e.g.
	the current or comparison site averages) must be done with	such a zero
	date to avoid biasing of a real season statistics.

	The necessary space is allocated for the result series to be accumulated
	with add(). If a \a dumpfile is supplied, the attempt to read this
	binary file is made. If successful and the date specified in the file descriptor
	coincides with the \a date, the series storages are filled from the file 
	and will thus contain the data from
	the previous runs of TURBINA in the same night, to be graphically displayed
	using getdata(). If not, the series are empty on startup; their starting 
	length (returned by getn()) is zero.

*/
void init(const char * sumfile, long date, int nchan, 
	const char * dumpfile = 0) ;

/**
	@brief	Add the results of the last accumulation time in the series tails
	@param	UTh		Fractional UT hour
	@param	mag		MASS-band magnitude of a target star (for the flux 
		conversion, see massmag())
	@param	airmass airmass of a target star (for summary writing only)
	@param	isgen	Generalized mode flag (availability of the shifted pupil
		results)

	The function adds all relevant parameter values to the tails of the
	result series using sc::getavgflux(), sc::getavgidx(), atm::getval(),
	atm::getzcn2() and atm::getcn2(). Non-computed parameters are saved as
	sum::NODATA (e.g. shifted indices in the Normal mode). 

	The magnitude \a mag is used to recompute the stellar flux for the zero-
	magnitude star. This magnitude is assumed to be converted into the MASS
	instrumental magnitude using the stellar B-V color index and the color
	equation of the MASS photometric system. It can be done with massmag(). The
	aim is to trace the atmospheric transparency of the site (having also the
	median \a airmass in the same summary-file) in parallel with the turbulence
	study.

	The shifted-pupil stellar fluxes are not accumulated and not saved in a
	summary file.

*/
void add(double UTh, double mag, double airmass, bool isgen) ;

/**
	@brief	Get the reference value for a given parameter
	@param	pname	The parameter name
	@param	rtype	Type of a reference: sum::FIRSTMEAN, sum::LASTMEAN, 
		sum::SEASONMEAN
	@return Reference value, \b or sum::NODATA if no reference is available

	The name \a pname, if supplied as is for the first time, is parsed to get
	know which of scintillation or atmopsheric parameters it refers to. If
	successfully parsed, the \em reference value of a specified type \a rtype
	is returned and the result of parsing (address of a reference storage) is
	stored in an associative array.

	The reserved names are
	\arg \c F_A..F_D - fluxes in channels A..D converted to zero magnitude 
		(\c FLUX_* are also understood)
	\arg \c SI_A..SI_D,SI_AB..SI_CD - scintillation indices in channels and 
		combinations (differential indices). Alternative names like \c DSI_AB 
		may also be used
	\arg \c DESI_A..DESI_D - differential exposure indices
	\arg \c SI_AS,SI_BS,(D)SI_ABS,DESI_AS - the selected indices set obtained 
		with a pupil shift in a Generalized mode
	\arg \c SEE,FSEE,HEFF,FHEFF,ISOPL,M0,M2,TAU - integral atmospheric 
		parameters (equivalent of atm::ordint enumeration). 
	The names are all in upper case and \b case-sensitive.

	The reference value is not available for the airmass, UT, and for 
	the Cn2 profile data (Cn2 is not stored in a summary-file).
	
	On error (non-identified parameter), the error nr::ERNOD is set and the
	value sum::NODATA is returned.
*/
double getref(const std::string& pname, sum::reftype rtype);

/**
	@brief	Get the pointer to the series storage for a given parameter
	@param	pname	The parameter name
	@param	nres	The result index for \a pname== \c ZCN2 and \c CN2; latest 
		result is returned by default
	@return pointer to the result series or to \a nres-th \em (z)cn2[] array 
		for \a pname== \c ZCN2 and \c CN2

	The name of the parameter is parsed as described for getref() and,
	if successfully done, the pointer to respective series storage is returned.
	In addition to the list of parameter names in getref(), the following names
	are allowed:
	\arg \c UT - the fractional UT for the graphics abscissa values
	\arg \c NCN2 - the number of restored turbulence profile layers
	\arg \c ZCN2 - the turbulence layer altitudes for the result #\a nres.
	\arg \c CN2 - the turbulence layer strengths for the result #\a nres.
	
	Note, that unlike other parameters, for pname== \c[Z]CN2 the layer
	altitudes and profile strengths arrays are returned, for the single (last by
	default) result, NOT the array of all the results obtained during the night.
	
	On error (non-identified parameter), the error nr::ERNOD is set and the
	pointer to the UT array is returned.
*/
const double * getdata(const std::string& pname, int nres = -1) ;
	
/**
	@brief	Get the number of results accumulated in series storages
	@return Current length of series (equal to all, taken from UT series)
*/
int getn() ;

/**
	@brief	Median average the result series and update/create the summary-file
		and the dump-file

	If the summary file \a sumfile, specified in init() call, does not exist,
	create it with a header-line with names of parameters and leading
	comment-char '#'. Then, if some scientific measurements were performed,
	compute the night-time median  parameters from the series and add a new
	line in a summary-file. This line is written after the last line in a file,
	or, if the line with the same date is found, on the place of this obsolete
	record.

	If a "dump-file" name \a dumpfile was supplied to init(), the series are
	(over)written in some binary format in this file. Before quitting, the
	series storages are released.

	The format of a dump-file is following. This binary file consists of a 
	series of records of double-type numbers, except for the first record. 
	The first record is a descriptor; it consists of four \b long \b integer numbers
	\verbatim
	<date DDMMYY> <Number_of_series> <length_of_each_series> <maximal N_layers>
	\endverbatim 
	
	Series include the one for fractional UT, one for airmass etc. 
	Then follow \c <Number of series> records - 
	\verbatim
	<parameter_(start_UT)> ... <parameter_(end_UT)>
	\endverbatim
	i.e. the contents of the series storages (arrays), each of the length \c
	<length_of_each_series>. 

	Thus, all the internally stored information of the module is written in
	the dump-file. This format is not fixed (not "official") since the
	dump-file is a "black-box" devoted for internal use in the SUMMARY
	module only.

*/
void done() ;

/** MASS color equation coefficient from linear regression of a synthetic
	relation Mag_{mass}-V vs (B-V)_0:  Mag_{mass} = V + MASSCE * (B-V) */
const double MASSCE = 0.5905 ;

/**
	@brief	Convert the Johnson V-magnitude into MASS magnitude
	@param	vmag	V-band magnitude
	@param	bv		B-V color
	@param	coleq	Linear color equation coefficient
	
	This function converts the standard V-magnitude of a star into the
	MASS-band magnitude using the linear color equation with a coefficient \a
	coleq. By default, this coefficient is taken from the module constant
	sum::MASSCE. For a star having (B-V)==0 (A0), the MASS magnitude is assumed to
	be equal to V-magnitude.

*/
inline double massmag(double vmag, double bv, double coleq = sum::MASSCE) 
{
	return vmag + bv * coleq ;
}
	
}	// end of namespace sum

const char SUMDLM[] = " ";	///< delimiter of a summary file

#endif
