/*--------------------------------------------------------------------------*/
/** @file tree.h
@brief  Class implementing the CFG tree organization: principles of its 
building, reading, searching info etc.

                  MASS project: TURBINA module file
                       header file for tree.cpp



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/
#ifndef TREE_H
#define TREE_H
#include <qobject.h>
#include <qstring.h>

class CFGTree;
class CFGBranch;
class CFGLeaf;

extern CFGLeaf leaf_empty;
extern CFGBranch branch_empty;

class Link {
private:
	Link* _next;
	char _id;	// 'l' or 'b' - leaf or branch
	void* _data;	//  ,     CFGLeaf or CFGBranch

public:
	Link(char id, void* data, CFGBranch* branch) {initLink(id,data,branch);}	
	Link(char id, void* data, CFGTree* tree) {initLink(id,data,tree);}		
	virtual ~Link(void);

	void initLink(char id, void* data, CFGBranch* branch);	
	void initLink(char id, void* data, CFGTree* tree);

	const char* BranchName(void);
	const char* getLeafKey(void);
	char* getLeafData(void);
	char getId(void)		{ return _id; }
	CFGBranch* getBranch(void) {if(_id=='b') return((CFGBranch*)_data); return NULL; }
	void setBranch(CFGBranch* branch) {if(branch) {_data=branch; _id='b';}}

	CFGLeaf* getLeaf(void) {if(_id == 'l') return((CFGLeaf*)_data); return NULL;}
	Link* getNext(void) { return _next; }
	void setNext(Link* next) { _next = next; }
};

class CFGLeaf {

private:
	QString _key;
	QString _data;
	bool _changed;
	QString validator_type;
	bool _enable_change;
	bool _for_root; // change enabled only for mass_root
	bool _count_disable; // disable or not v techenie nakopleniya
	QString _help;

public:
	CFGLeaf(const char* key,
		CFGBranch* branch,
		bool for_root = true,
		bool enable_change = true,
		bool count_disable = true) {initLeaf(key,for_root,enable_change,count_disable);
									if(branch) new Link('l',this,branch);
									}
	CFGLeaf(const char* key,
		QString& validator_type_,	// validator
		CFGBranch* branch,
		bool for_root = true,
		bool enable_change =true,
		bool count_disable = true) {initLeaf(key,validator_type_,for_root,enable_change,count_disable);
									if(branch) new Link('l',this,branch);
									}
	virtual ~CFGLeaf();


	void initLeaf(const char* key,
		bool for_root = true,
		bool enable_change = true,
		bool count_disable = true);
	void initLeaf(const char* key,
		QString& validator_type_,	// validator
		bool for_root = true,
		bool enable_change =true,
		bool count_disable = true);

	bool isChanged(void) { return _changed; }
	void setChanged(bool chng) { _changed = chng; }
	bool isChangeEnabled(void) { return _enable_change; }
	bool isForRoot(void) { return _for_root; }
	bool isCountDisabled(void) { return _count_disable; }
	const char* getKey(void) { return _key.data(); }
	QString* getData(void) { return &_data; }
	QString* getHelp(void) { return &_help; }
	void setHelp(const char*);
	int replaceData(const QString& data);
	int replaceData(const char* data);
	const char* getValidatorType() { return validator_type.data(); }
};

class CFGBranch {

private:
	QString _name;
	Link *first_link; //    ,    link
	Link *last_link;
	Link *current_link;
	QString _help;
	int n_leaf;		//         
	int n_subbranch; //     

public:
	CFGBranch(const char* name, CFGTree* tree) {initBranch(name);
									if(tree) {
										new Link('b',this,tree);
									}
								}
	CFGBranch(const char* name, CFGBranch* branch) {initBranch(name);
									if(branch) {
										new Link('b',this,branch);
									}
								}
	virtual ~CFGBranch(void);

	void initBranch(const char* name);

	void addLink(Link*);
	QString* getHelp(void) { return &_help; }
	void setHelp(const char*);
	Link* findBranchLink(const char* branch_name);
	CFGBranch* findBranch(const char*);
	CFGLeaf* findLeaf(const char*); //     
	QString* findLeaf_str(const char*); //       
	char* findLeaf_ch(const char*);
	int replaceLeaf(const char*,const char*);
	int replaceChanged(const char*, bool isChanged);
	int setLeafHelp(const char*,const char*);
	void toFirst(void) { current_link = first_link; }
	void toNext(void) { current_link = current_link->getNext(); }
	Link* getCurrentLink(void) { return current_link; }
	const char* getName(void) { return _name.data(); }
	int getNLeaf(void) { return n_leaf; }
	int getNSubBranch(void) { return n_subbranch; }
	void printBranch(CFGBranch* branch);
};                        

class CFGTree : public QObject {
	Q_OBJECT
private:
	Link* first_link;
	Link* current_link;
	Link* last_link;
	int _changed;

public:
	CFGTree(void);
	virtual ~CFGTree(void);
	void addLink(Link*);
	Link* findBranchLink(const char* branch_name);
	CFGBranch* findBranch(const char*);
	void toFirst(void) { current_link = first_link; } 
	void toNext(void) { current_link = current_link->getNext(); }
	Link* getCurrentLink(void) { return current_link; }
	const char* CurrentName(void) { return current_link->getBranch()->getName(); }
	void printTree(CFGTree* tree);
	int readCfg (const char* fncfg, const char* fnstaus=NULL);
	int saveToFile(const char*);
	void emitSignalChanged(void);
	int isChanged(void) {return(_changed); }
	void setSaved(void) {_changed = 0;}
	void setChanged(int chng) {_changed = chng;}
	void resetTree(void);
signals:
	void changed(void);
};

#endif
