#include <qtooltip.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qpainter.h>
#include <string>
#include "widgets.h"
#include "names.h"
//#include "exchange.h"
#include "scenariothr.h"
#include "star.h"
#include "tools.h"
#include "messages.h"
#include "cfg.h"

extern DevManager	devmanager;
extern Star		star;
extern ExportTable	table;
extern FILE		*out_file;

static const char* starText			= "Observed Star";
static const char* haText			= "Hour Angle";
static const char* raText			= "RA";
static const char* decText			= "Declination";
static const char* magText			= "V-magnitude";
static const char* spText			= "Spectrum";
static const char* duplText			= "Duplicity";

static const int SEPARATOR = 5;

int getHeight(int fontHeight) {
	return fontHeight*5 / 3;
}
QJKMainView::QJKMainView(QWidget* parent,const char* name)
				:QWidget(parent, name) {
	jdebug("QJKMainView: start Constructor");
	table = new QJKTable(this, "MainTable", MAX_ROW, MAX_COL);
	CHECK_PTR(table);

	textView = new QLineEdit(this, "MainTextView");
	CHECK_PTR(textView);
	textView->setReadOnly(true);
	textView->setAlignment(AlignHCenter);

	setChildrenGeometry();
	jdebug("QJKMainView: end Constructor");
}

QJKMainView::~QJKMainView() {
	jdebug("Destructor of QJKMainView");
}
                        
void QJKMainView::fillTable(void) {
	table->fillTable();
}

void QJKMainView::showTextLine(void) {
	jdebug("TextView setText: %s",(ExportText::text).data());
	textView->setText(ExportText::text);
}

void QJKMainView::setChildrenGeometry(void) {
	int _height = getHeight(fontMetrics().height());
	jdebug("%d %d %d",
		width(),height(),_height);
	table->setGeometry(0,0,width(),
		height() - _height - SEPARATOR);
	textView->setGeometry(0,table->y()+table->height()+SEPARATOR,
			width(), _height);
	table->setFrameStyle(QFrame::Sunken);
}

void QJKMainView::resizeEvent(QResizeEvent* e) {
	jdebug("QJKMainView: ResizeEvent!");
	setChildrenGeometry();
}

void QJKMainView::fitToFont(void) {
	table->fitToFont();
}

QJKTable::QJKTable(QWidget* parent, const char* name,
			int numRows, int numCols,
			bool grid)
				:QtTableView(parent, name) {
	setFocusPolicy(NoFocus);
	setBackgroundMode(PaletteBase);
	setNumCols(numCols);
	setNumRows(numRows);

	setCellGeometry();

	setTableFlags(
		Tbl_clipCellPainting	// avoid drawing outside cell
		);

	contents = new QString[numRows * numCols]; //make room for contents


//	isColumnSizeChanged = false;
	showGrid = grid;
}

QJKTable::~QJKTable() {
	jdebug("destructor of Table");
	delete[] contents;
	jdebug("end of destructor of Table");
}

void QJKTable::setCellContent(int row, int col, const QString& text) {
	contents[ indexOf(row,col)] = text;
	updateCell(row,col);	// show new content
}

void QJKTable::paintCell(QPainter* p, int row, int col) {
	int w = cellWidth(col);	// width of cell in pixels
	int h = cellHeight(row);	// height of cell in pixels

	if(showGrid) {
		int x2 = w - 1;
		int y2 = h - 1;
	// Draw our part of cell frame.
		p->drawLine(x2,0,x2,y2);	// draw vertical line on right
		p->drawLine(0, y2, x2, y2);	// raw horiz. line at bottom
	}

// Draw cell content (text)	
	p->drawText(0,0,w,h,AlignCenter,contents[indexOf(row,col)]);

}

int QJKTable::indexOf(int row, int col) const {
	return(row * numCols() + col);
}

void QJKTable::setCellGeometry(void) {
	int cellWidth = (int)((double)width() / ((double)numCols()+0.25));
	int cellHeight = fontMetrics().height();
	setCellWidth(cellWidth );
	setCellHeight(cellHeight);
}

void QJKTable::resizeEvent(QResizeEvent* e) {
	setCellGeometry();
}

void QJKTable::fillTable(void ) {

	int startRow = table.startRow;
	int endRow = table.startRow + table.nRows;
	int startCol = table.startCol;
	int endCol = table.startCol + table.nCols;
	if(startRow < 0) startRow = 0;
	if(startCol < 0) startCol = 0;
	if(endRow > numRows()) endRow = numRows();
	if(endCol > numCols()) endCol = numCols();

	jdebug("fillTable: %d %d %d %d",startRow,endRow,
			startCol,endCol);
	if(table.clean) {
		table.clean = false;
		for(int row=0; row<numRows(); row++) {
			if((row<startRow)
				|| (row>=endRow)) table.cleanRow(row);
			for(int col=0; col<numCols(); col++) {
				setCellContent(row,col,"");
			}
		}
	}
	for(int row=startRow; row<endRow; row++) {
		for(int col=startCol; col<endCol; col++) {
			setCellContent(row,col,table.getData(row,col));
		}
	}
}

void QJKTable::fitToFont(void) {
	setCellGeometry();
}

HelpView::HelpView(const QString& helpFileName,
		QWidget* parent, const char* name)
				:QDialog(parent,name,true/*modal*/) {
	setCaption("Help");
	int border = fontMetrics().height();

	layout = new QGridLayout(this,5,1,border);
	CHECK_PTR(layout);

	label = new QLabel("Help",this);
	CHECK_PTR(label);
	layout->addWidget(label,0,0,AlignCenter);

	textView = new QTextView(this,"helpTextView");
	CHECK_PTR(textView);
	int winWidth = qApp->desktop()->width()/2;
	int winHeight = qApp->desktop()->height()/2;
	textView->setMinimumSize(winWidth,winHeight);
	layout->addMultiCellWidget(textView,1,3,0,0,AlignCenter);

	closebutton = new QPushButton("Close",this);
	CHECK_PTR(closebutton);
	closebutton->setFixedSize(closebutton->sizeHint());
	connect(closebutton, SIGNAL(clicked()),
				this, SLOT(accept()));
	layout->addWidget(closebutton,4,0,AlignCenter);
	setSizeGripEnabled(true);

	loadHelp(helpFileName);
}

void HelpView::loadHelp(const QString& fileName) {
	QFile f(fileName);
	f.open(IO_ReadOnly);
	QTextStream t(&f);
	bool first = true;
	while(!t.eof()) {
		QString str = t.readLine();
		if(str.isEmpty()) str=".";
		if(first) {
			textView->setText(str);
			first = false;
		}
		else textView->append(str);

	}	
	f.close();
}

HelpView::~HelpView() {
	jdebug("destructor of TextView");
	delete textView;
	delete label;
	delete closebutton;
	delete layout;
}

/*
QJKTextView::QJKTextView(QWidget* parent, const char* name)
				:QTextView(parent, name) {
	setVScrollBarMode(QScrollView::AlwaysOff);
	setHScrollBarMode(QScrollView::AlwaysOff);
}

QJKTextView::~QJKTextView() {
	jdebug("destructor of TextView");
}

void QJKTextView::addLine(const QString& _text) {
	text = _text;
	setText(text);
}
*/

QJKPushButton::QJKPushButton(const char* text, 
			QWidget* parent, const char* name)
		: QPushButton(text, parent, name) {
}

QJKPushButton::~QJKPushButton() {
}

void QJKPushButton::focusInEvent(QFocusEvent* e) {
	setDefault(true);
	QPushButton::focusInEvent(e);
}

void QJKPushButton::focusOutEvent(QFocusEvent* e) {
	setDefault(false);
	QPushButton::focusOutEvent(e);
}

DigitalClock::DigitalClock(QWidget *parent, const char *name)
				: QLCDNumber(parent, name) {
	setNumDigits(8);
	setSegmentStyle(Flat);
	showTime();
	fitToFont();
}

DigitalClock::~DigitalClock() {
}

void DigitalClock::fitToFont(void) {
	int w = fontMetrics().
		width("WW:WW:WW");
	int h = fontMetrics().height();
	setFixedSize(w,h);
//	setFixedSize(sizeHint().width(),h);
}

void DigitalClock::showTime() {
	QString s = QTime::currentTime().toString().left(8);
	if(s[0] == '0') s[0] = ' ';
	display(s);   // set LCD number/text
}

TalkDialog::TalkDialog(void)	
		: QDialog(0,"TalkDialog",true /*modal*/) {
	setCaption("Talk");

	int border = fontMetrics().height();
	layout = new QVBoxLayout(this,border);
	CHECK_PTR(layout);

	label = new QLabel("Talk",this);
	CHECK_PTR(label);
	label->setFixedSize(label->sizeHint());
	layout->addWidget(label,0,AlignCenter);

	combo = new QComboBox(true,this);
	CHECK_PTR(combo);
	int _width = fontMetrics().width("wwwwwwwwwwwwwwwwww");
	int _height = getHeight(fontMetrics().height());
	combo->setFixedSize(_width, _height);
	combo->setFocus();
	connect(combo,SIGNAL(activated(const QString&)),
		this,SLOT(slotTalk(const QString&)));

	layout->addWidget(combo);

	edit = new QMultiLineEdit(this);
	CHECK_PTR(edit);
	edit->setFixedSize(_width,3*_height);
	edit->setReadOnly(true);
	edit->setFocusPolicy(NoFocus);
	layout->addWidget(edit);
	
	closebutton = new QJKPushButton("Close",this);
	CHECK_PTR(closebutton);
	closebutton->setFixedSize(closebutton->sizeHint());
	connect(closebutton, SIGNAL(clicked()), 
				this, SLOT(accept()));

	layout->addWidget(closebutton);
}

TalkDialog::~TalkDialog() {
	jdebug("TalkDialog destructor...");
	delete layout;
	delete label;
	delete combo;
	delete edit;
	delete closebutton;
}

void TalkDialog::slotTalk(const QString& text) {
	std::string resp = exch::talk(text.data());
	QString qresp = resp.c_str();
	jdebug("slotTalk: response: |%s|", qresp.data());
	edit->setText(qresp);
}

AdjustDialog::AdjustDialog(void)	
		: QDialog(0,"AdjustDialog",true /*modal*/) {

	lastAperture = devmanager.currentAperture();
	setCaption("Apertures adjust");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,5,3,border);
	CHECK_PTR(layout);

	label = new QLabel("Apertures adjust:",this);
	CHECK_PTR(label);
	layout->addMultiCellWidget(label,0,0,0,2,AlignHCenter);

	combo_apert = new QComboBox(false,this);
	CHECK_PTR(combo_apert);

	int ns = devmanager.getNStops();
	for(int i=0; i<ns; i++) {
		combo_apert->insertItem(devmanager.getStopName(i));
	}	
	connect(combo_apert,SIGNAL(activated(const QString&)),
		this,SLOT(setAperture(const QString&)));
	layout->addMultiCellWidget(combo_apert,1,1,1,2,AlignHCenter);
	label_apert = new QLabel("Aperture:",this);
	CHECK_PTR(label_apert);
	layout->addWidget(label_apert,1,0,AlignLeft);

	label_correct = new QLabel("Correction:",this);
	CHECK_PTR(label_correct);
	layout->addWidget(label_correct,2,0,AlignLeft);
	result_window = new QLineEdit(this);
	CHECK_PTR(result_window);
	result_window->setReadOnly(true);
	result_window->setAlignment(AlignHCenter);
	int _width = fontMetrics().width("wwwwwww");
	int _height = getHeight(fontMetrics().height());
	result_window->setFixedSize(_width, _height);
	QString str;
	str.setNum(devmanager.getStopCorrectNew());
	result_window->setText(str);
	layout->addMultiCellWidget(result_window,2,2,1,2,AlignHCenter);

	move_plus = new QJKPushButton("+",this);
	CHECK_PTR(move_plus);
	move_plus->setAccel('+');
	connect(move_plus, SIGNAL(clicked()), 
		this, SLOT(movePlus()));
	layout->addWidget(move_plus,3,2);

	move_minus = new QJKPushButton("-",this);
	CHECK_PTR(move_minus);
	move_minus->setAccel('-');
	connect(move_minus, SIGNAL(clicked()), 
		this, SLOT(moveMinus()));
	layout->addWidget(move_minus,3,1);

	label_move = new QLabel("Move:",this);
	CHECK_PTR(label_move);
	layout->addWidget(label_move,3,0,AlignLeft);

	acceptbutton = new QJKPushButton("Accept",this);
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize(acceptbutton->sizeHint());
	connect(acceptbutton, SIGNAL(clicked()), 
				this, SLOT(accept()));
	layout->addWidget(acceptbutton,4,0);

	cancelbutton = new QJKPushButton("Cancel",this);
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize(cancelbutton->sizeHint());
	connect(cancelbutton, SIGNAL(clicked()), 
				this, SLOT(reject()));
	layout->addMultiCellWidget(cancelbutton,4,4,1,2,AlignHCenter);
}

void AdjustDialog::update(void) {
	devmanager.setAperture(lastAperture,1);
	setCurrentItem(lastAperture);
	jdebug("AdjusDialog::update setAperture %d",lastAperture);
}

void AdjustDialog::moveMinus(void) {
	QString str;

	devmanager.moveMinus(1);
	str.setNum(devmanager.getStopCorrectNew());
	result_window->setText(str);
}

void AdjustDialog::movePlus(void) {
	QString str;

	devmanager.movePlus(1);
	str.setNum(devmanager.getStopCorrectNew());
	result_window->setText(str);
}

void AdjustDialog::setCurrentItem(int _item) {
	QString str;
	jdebug("setCurrentItem item=%d",_item);

	if((_item >= 0) && (_item < combo_apert->count())) {
		if(combo_apert) combo_apert->setCurrentItem(_item);
		str.setNum(devmanager.getStopCorrectNew());
		jdebug("str=%s",str.data());
		if(result_window) result_window->setText(str);
	}
	jdebug("return from setCurrentItem");
}

void AdjustDialog::setAperture(const QString& apertName) {
	devmanager.setAperture(apertName,1);
	lastAperture = devmanager.currentAperture();
	setCurrentItem(lastAperture);
}

AdjustDialog::~AdjustDialog() {
	jdebug("AdjustDialog destructor...");
	delete layout;
	delete label;
	delete label_apert;
	delete label_move;
	delete label_correct;
	delete result_window;
	delete combo_apert;
	delete acceptbutton;
	delete cancelbutton;
	delete move_plus;
	delete move_minus;
}

ShowStarDialog::ShowStarDialog(int _showHA_)
		: QDialog(0,"ShowStarDialog",true) {
	_showHA = _showHA_;
	setCaption("Show star");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,3,6,border);
	CHECK_PTR(layout);

	labelName = new QLabel(star.getName(),this);
	CHECK_PTR(labelName);
	QToolTip::add(labelName,starText);
	layout->addMultiCellWidget(labelName,0,0,0,5,AlignHCenter);
	
	if(_showHA) {
		star.updateCurrent();
		labelHA = new QLabel(star.getHA(),this);
		CHECK_PTR(labelHA);
		QToolTip::add(labelHA,haText);
	} else {
		labelHA = new QLabel(star.getRA(),this);
		CHECK_PTR(labelHA);
		QToolTip::add(labelHA,raText);
	}
	labelHA->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelHA->setAlignment(AlignCenter);
	layout->addWidget(labelHA,1,0);

	labelDec = new QLabel(star.getDec(),this);
	CHECK_PTR(labelDec);
	QToolTip::add(labelDec,decText);
	labelDec->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelDec->setAlignment(AlignCenter);
	layout->addWidget(labelDec,1,1);

	labelMagV = new QLabel(star.getMagV(),this);
	CHECK_PTR(labelMagV);
	QToolTip::add(labelMagV,magText);
	labelMagV->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelMagV->setAlignment(AlignCenter);
	layout->addWidget(labelMagV,1,2);

	labelSp = new QLabel(star.getSpectrum(),this);
	CHECK_PTR(labelSp);
	QToolTip::add(labelSp,spText);
	labelSp->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelSp->setAlignment(AlignCenter);
	layout->addWidget(labelSp,1,3);

	labelDupl = new QLabel(star.getDuplicity(),this);
	CHECK_PTR(labelDupl);
	QToolTip::add(labelDupl,duplText);
	labelDupl->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelDupl->setAlignment(AlignCenter);
	layout->addMultiCellWidget(labelDupl,1,1,4,5,AlignHCenter);

	closeButton = new QJKPushButton("Close",this);
	CHECK_PTR(closeButton);
	layout->addWidget(closeButton,2,3);
	connect(closeButton, SIGNAL(clicked()), 
				this, SLOT(accept()));
}

ShowStarDialog::~ShowStarDialog() {
	delete layout;
	delete labelName;
	delete labelHA;
	delete labelDec;
	delete labelMagV;
	delete labelSp;
	delete labelDupl;
	delete closeButton;
}

void ShowStarDialog::update(int _showHA_) {
	_showHA = _showHA_;
	labelName->setText(star.getName());
	QToolTip::remove(labelHA);
	if(_showHA) {
		star.updateCurrent();
		labelHA->setText(star.getHA());
		QToolTip::add(labelHA,haText);
	} else {
		labelHA->setText(star.getRA());
		QToolTip::add(labelHA,raText);
	}
	labelDec->setText(star.getDec());
	labelMagV->setText(star.getMagV());
	labelSp->setText(star.getSpectrum());
	labelDupl->setText(star.getDuplicity());
}

void ShowStarDialog::showHA(void) {
	star.updateCurrent();
	labelHA->setText(star.getHA());
}

CommentDialog::CommentDialog(void)
		: QDialog(0,"CommentDialog",true) {
	setCaption("Comment");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,2,3,border);
	CHECK_PTR(layout);

	labelHeader = new QLabel("Add comment to MASS-file",this);
	CHECK_PTR(labelHeader);
	layout->addMultiCellWidget(labelHeader,0,0,0,1,AlignHCenter);

	edit = new QLineEdit(this);
	CHECK_PTR(edit);
	edit->setAlignment(AlignHCenter);
	layout->addMultiCellWidget(edit,1,1,0,1,AlignHCenter);

	acceptButton = new QPushButton("Accept",this);
	CHECK_PTR(acceptButton);
	acceptButton->setDefault(true);
	layout->addWidget(acceptButton,2,0);
	connect(acceptButton, SIGNAL(clicked()), 
				this, SLOT(acceptText()));
	cancelButton = new QPushButton("Cancel",this);
	CHECK_PTR(cancelButton);
	layout->addWidget(cancelButton,2,1);
	connect(cancelButton, SIGNAL(clicked()), 
				this, SLOT(reject()));
	fitToFont();
}

void CommentDialog::acceptText(void) {
	(*comment) = edit->text();
	accept();
}

void CommentDialog::update(QString* p_comment) {
	comment = p_comment;
	edit->clear();
	edit->setFocus();

}

CommentDialog::~CommentDialog() {
	delete layout;
	delete labelHeader;
	delete edit;
	delete acceptButton;
	delete cancelButton;
}

void CommentDialog::fitToFont(void) {
	int _width = fontMetrics().
		width("wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww");
	int _height = getHeight(fontMetrics().height());
	edit->setFixedSize(_width, _height);
	setFixedSize(sizeHint());
}

CenteringDialog::CenteringDialog(void)
		: QDialog(0,"Centering",true) {
	setCaption("Centering");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,2,6,border);
	CHECK_PTR(layout);

	labelHeader = new QLabel("Star position",this);
	CHECK_PTR(labelHeader);
	layout->addMultiCellWidget(labelHeader,0,0,0,1,AlignHCenter);

	const uint numDigits=6;
	lcdDx = new QLCDNumber(numDigits,this,"lcd_DX");
	CHECK_PTR(lcdDx);
	lcdDx->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	lcdDx->setSegmentStyle(QLCDNumber::Filled);
//	lcdDx->setSmallDecimalPoint(true);
	layout->addMultiCellWidget(lcdDx,1,4,0,0,AlignHCenter);

	lcdDy = new QLCDNumber(numDigits,this,"lcd_DY");
	CHECK_PTR(lcdDy);
	lcdDy->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	lcdDy->setSegmentStyle(QLCDNumber::Filled);
//	lcdDy->setSmallDecimalPoint(true);
	layout->addMultiCellWidget(lcdDy,1,4,1,1,AlignHCenter);

	startButton = new QJKPushButton("Start",this);
	CHECK_PTR(startButton);
	layout->addWidget(startButton,5,0);
	connect(startButton, SIGNAL(clicked()), 
				this, SLOT(start()));

	closeButton = new QJKPushButton("Finish",this);
	CHECK_PTR(closeButton);
	layout->addWidget(closeButton,5,1);
	connect(closeButton, SIGNAL(clicked()), 
				this, SLOT(finish()));
	fitToFont();
}

CenteringDialog::~CenteringDialog() {
	delete layout;
	delete labelHeader;
	delete lcdDx;
	delete lcdDy;
	delete startButton;
	delete closeButton;
}

void CenteringDialog::update(void) {
	delta.setX(0);
	delta.setY(0);
	QString str;
	str.sprintf("+00.00");
	lcdDx->display(str);
	str.sprintf("+00.00");
	lcdDy->display(str);
	closeButton->setFocus();

// sbros oshibki:
	StateExport::cleanError();

	timer = startTimer(DIALOG_POLLING_INT); // 0.5 sec
	_polling = true;
	start();
}

void CenteringDialog::start(void) {
	startButton->setEnabled(false);
	scenariothr::centering(&delta); 
}

void CenteringDialog::finish(void) {
	devmanager.stopCounting();
	killTimer(timer);
	accept();
}

void CenteringDialog::timerEvent(QTimerEvent *e) {
	if(_polling) {
		_polling = false;
		if(StateExport::correction()) {
			StateExport::cleanCorrectionReady();
			showResult();
			if(StateExport::error()) {
				StateExport::cleanError();
				showStrongMessage(StateExport::message());
				startButton->setEnabled(true);
			}
		}
		_polling = true;
	}
}

void CenteringDialog::showResult(void) {
	QString str;
	str.sprintf("%+5.2f",delta.x());
	lcdDx->display(str);
	str.sprintf("%+5.2f",delta.y());
	lcdDy->display(str);
}

void CenteringDialog::fitToFont(void) {
	int w = 5*fontMetrics().width("WW");
	int h = 4*fontMetrics().height();
	lcdDx->setFixedSize(w,h);
	lcdDy->setFixedSize(w,h);
	setFixedSize(sizeHint());
}

SelectStarDialog::SelectStarDialog(void)	
		: QDialog(0,"SelectStarDialog",true /*modal*/) {

	setCaption("Select star");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,7,2,border);
	CHECK_PTR(layout);

	label = new QLabel("Select star:",this);
	CHECK_PTR(label);
	label->setFixedSize(label->sizeHint());
	layout->addMultiCellWidget(label,0,0,0,1,AlignHCenter);

	viewStarList = new ViewStarList(&star, this);
	viewStarList->setFixedSize(viewStarList->sizeHint());
	layout->addMultiCellWidget(viewStarList,1,5,0,1,AlignHCenter);

	selectbutton = new QJKPushButton("Select",this);
	CHECK_PTR(selectbutton);
	selectbutton->setFixedSize(selectbutton->sizeHint());
	if(InfoExport::withSV || devmanager.isCounting() || InfoExport::isScenario) {
		selectbutton->setDisabled(true);
	}
	layout->addWidget(selectbutton,6,0);
	connect(selectbutton, SIGNAL(clicked()), 
				this, SLOT(select()));
	connect(viewStarList, SIGNAL(doubleClicked(QListViewItem*)),
				this, SLOT(select(QListViewItem*)));

	cancelbutton = new QJKPushButton("Cancel",this);
	CHECK_PTR(cancelbutton);
	layout->addWidget(cancelbutton,6,1);
	cancelbutton->setFixedSize(cancelbutton->sizeHint());
	connect(cancelbutton, SIGNAL(clicked()), 
				this, SLOT(reject()));
}

void SelectStarDialog::update(void) {
	if(InfoExport::withSV || devmanager.isCounting() || InfoExport::isScenario) {
		selectbutton->setDisabled(true);
	} else {
		selectbutton->setDisabled(false);
	}
	viewStarList->update(&star);
}

void SelectStarDialog::select(void) {
	if(!InfoExport::withSV && !devmanager.isCounting() && !InfoExport::isScenario) {
		int number = viewStarList->getCurrentStarNumber();
		star.setCurrentStar(number);
	}
	accept();
}

void SelectStarDialog::select(QListViewItem* it) {
	if(!InfoExport::withSV && !devmanager.isCounting() && !InfoExport::isScenario) {
		int number = viewStarList->getCurrentStarNumber();
		star.setCurrentStar(number);
	}
	accept();
}

SelectStarDialog::~SelectStarDialog() {
	jdebug("SelectStarDialog destructor...");
	delete label;
	delete layout;
	delete viewStarList;
	delete selectbutton;
	delete cancelbutton;
}

IllumDialog::IllumDialog(void)	
		: QDialog(0,"IllumDialog",true /*modal*/) {
	setCaption("Illumination level");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,4,2,border);
	CHECK_PTR(layout);

	label = new QLabel("Illumination level",this);
	CHECK_PTR(label);
	label->setFixedSize(label->sizeHint());
	layout->addMultiCellWidget(label,0,0,0,1,AlignHCenter);

	bright_window = new QLineEdit(this);
	CHECK_PTR(bright_window);
	bright_window->setReadOnly(true);
	bright_window->setAlignment(AlignHCenter);
	int _width = fontMetrics().width("wwwwwww");
	int _height = getHeight(fontMetrics().height());
	bright_window->setFixedSize(_width, _height);
//	setBrightness();
//	connect(&devmanager, SIGNAL(signalBrightness(int)), 
//		this, SLOT(slotSetBrightness(int)));
	layout->addMultiCellWidget(bright_window,1,1,0,1,AlignHCenter);

	bright_plus = new QJKPushButton("+",this);
	CHECK_PTR(bright_plus);
	bright_plus->setAccel('+');
	bright_plus->setFixedSize(bright_plus->sizeHint());
	connect(bright_plus, SIGNAL(clicked()), 
		this, SLOT(setBrightnessUp()));
//		&devmanager, SLOT(setBrightnessUp()));
	layout->addWidget(bright_plus,2,1);

	bright_minus = new QJKPushButton("-",this);
	CHECK_PTR(bright_minus);
	bright_minus->setAccel('-');
	bright_minus->setFixedSize(bright_minus->sizeHint());
	connect(bright_minus, SIGNAL(clicked()), 
		this, SLOT(setBrightnessDown()));
//		&devmanager, SLOT(setBrightnessDown()));
	layout->addWidget(bright_minus,2,0);

	closebutton = new QJKPushButton("Close",this);
	CHECK_PTR(closebutton);
	closebutton->setFixedSize(closebutton->sizeHint());
	connect(closebutton, SIGNAL(clicked()), 
				this, SLOT(accept()));
	layout->addMultiCellWidget(closebutton,3,3,0,1,AlignHCenter);
}

void IllumDialog::update(void) {
	setBrightness();
}

void IllumDialog::setBrightnessUp(void) {
	devmanager.setBrightnessUp();
	setBrightness();
}

void IllumDialog::setBrightnessDown(void) {
	devmanager.setBrightnessDown();
	setBrightness();
}

void IllumDialog::setBrightness(void) {
	QString str;
	str.sprintf(" %.2f",
		(double)devmanager.getCurrentBrightness() / 15.0);
	bright_window->setText(str);
}

IllumDialog::~IllumDialog() {
	jdebug("IllumDialog destructor...");
	delete layout;
	delete label;
	delete bright_window;
	delete bright_plus;
	delete bright_minus;
	delete closebutton;
}

LightDialog::LightDialog(void)	
		: QDialog(0,"LightDialog",true /*modal*/) {
	setCaption("Control Light level");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,3,2,border);
	CHECK_PTR(layout);

	label = new QLabel("Control Light level",this);
	CHECK_PTR(label);
	label->setFixedSize(label->sizeHint());
	layout->addMultiCellWidget(label,0,0,0,1,AlignHCenter);

	level_window = new QLineEdit(this);
	CHECK_PTR(level_window);
	level_window->setAlignment(AlignHCenter);
	int _width = fontMetrics().width("wwwwwww");
	int _height = getHeight(fontMetrics().height());
	level_window->setFixedSize(_width, _height);
	layout->addMultiCellWidget(level_window,1,1,0,1,AlignHCenter);

	acceptbutton = new QJKPushButton("Accept",this);
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize(acceptbutton->sizeHint());
	connect(acceptbutton, SIGNAL(clicked()), 
				this, SLOT(acceptLight()));
	layout->addWidget(acceptbutton,2,0,AlignHCenter);

	cancelbutton = new QJKPushButton("Cancel",this);
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize(cancelbutton->sizeHint());
	connect(cancelbutton, SIGNAL(clicked()), 
				this, SLOT(reject()));
	layout->addWidget(cancelbutton,2,1,AlignHCenter);
}

void LightDialog::acceptLight(void) {
	QString light_str = level_window->text();
	double light = strToDouble(&light_str,
					"", 0.0, 1.0);
	if(light == -1) {} //reject();
	else {
		devmanager.setLight(light,1,0);
		accept();
	}
}

void LightDialog::update(void) {
	setLevel();
}

void LightDialog::setLevel(void) {
	QString str;
	str.sprintf(lightFormat, devmanager.getLight());
	level_window->setText(str);
}

LightDialog::~LightDialog() {
	jdebug("LightDialog destructor...");
	delete layout;
	delete label;
	delete level_window;
	delete acceptbutton;
	delete cancelbutton;
}

HVDialog::HVDialog(void)	
		: QDialog(0,"HVDialog",true /*modal*/) {
	setCaption("High Voltage");

	int border = fontMetrics().height();

	layout = new QGridLayout(this,3,2,border);
	CHECK_PTR(layout);

	label = new QLabel("High Voltage",this);
	CHECK_PTR(label);
	label->setFixedSize(label->sizeHint());
	layout->addMultiCellWidget(label,0,0,0,1,AlignHCenter);

	level_window = new QLineEdit(this);
	CHECK_PTR(level_window);
	level_window->setAlignment(AlignHCenter);
	int _width = fontMetrics().width("wwwwwww");
	int _height = getHeight(fontMetrics().height());
	level_window->setFixedSize(_width, _height);
	layout->addMultiCellWidget(level_window,1,1,0,1,AlignHCenter);

	acceptbutton = new QJKPushButton("Accept",this);
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize(acceptbutton->sizeHint());
	connect(acceptbutton, SIGNAL(clicked()), 
				this, SLOT(acceptVoltage()));
	layout->addWidget(acceptbutton,2,0,AlignHCenter);

	cancelbutton = new QJKPushButton("Cancel",this);
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize(cancelbutton->sizeHint());
	connect(cancelbutton, SIGNAL(clicked()), 
				this, SLOT(reject()));
	layout->addWidget(cancelbutton,2,1,AlignHCenter);
}

void HVDialog::acceptVoltage(void) {
	QString hv_str = level_window->text();
	int hv = strToInt(&hv_str,
					"", 0, 1250);
	if(hv == -1) {} //reject();
	else {
		if( devmanager.setVoltage(hv, 1, false) == SUCCESS ) {
			QString newVoltage;
			newVoltage.sprintf(hvFormat,hv);
			setLeafData(cfgFile, "Operations",
				"Common","HighVoltage", newVoltage,
				out_file);
		}
		accept();
	}
}

void HVDialog::update(void) {
	setLevel();
}

void HVDialog::setLevel(void) {
	QString str;
	str.sprintf(hvFormat, devmanager.getVoltage());
	level_window->setText(str);
}

HVDialog::~HVDialog() {
	jdebug("HVDialog destructor...");
	delete layout;
	delete label;
	delete level_window;
	delete acceptbutton;
	delete cancelbutton;
}

BackCorrectionDialog::BackCorrectionDialog(void)
			: QDialog(0,"Background Correction",true) {
	setCaption("Background Correction");
	int border = fontMetrics().height();
	layout = new QGridLayout(this,N_CHAN+3,5,border);
	CHECK_PTR(layout);

	label = new QLabel("Background Correction",this);
	CHECK_PTR(label);
	layout->addMultiCellWidget(label,0,0,0,3,AlignHCenter);

	label_cf = new QLabel("Center/Side ratio",this);
	CHECK_PTR(label_cf);
	layout->addMultiCellWidget(label_cf,1,1,0,3,AlignHCenter);

	label_l = new QLabel("Left",this);
	CHECK_PTR(label_l);
	layout->addWidget(label_l,2,1,AlignHCenter);

	label_r = new QLabel("Right",this);
	CHECK_PTR(label_r);
	layout->addWidget(label_r,2,2,AlignHCenter);

	label_av = new QLabel("Average",this);
	CHECK_PTR(label_av);
	layout->addWidget(label_av,2,3,AlignHCenter);

	label_a = new QLabel("Chan.A",this);
	CHECK_PTR(label_a);
	layout->addWidget(label_a,3,0,AlignHCenter);

	label_b = new QLabel("Chan.B",this);
	CHECK_PTR(label_b);
	layout->addWidget(label_b,4,0,AlignHCenter);

	label_c = new QLabel("Chan.C",this);
	CHECK_PTR(label_c);
	layout->addWidget(label_c,5,0,AlignHCenter);

	label_d = new QLabel("Chan.D",this);
	CHECK_PTR(label_d);
	layout->addWidget(label_d,6,0,AlignHCenter);

	int _width = fontMetrics().width("wwww");
	int _height = getHeight(fontMetrics().height());
	for(int i=0; i<N_CHAN; i++) {
		edit_l[i] = new QLineEdit(this);
		CHECK_PTR(edit_l[i]);
		edit_l[i]->setReadOnly(true);
		edit_l[i]->setAlignment(AlignHCenter);
		edit_l[i]->setFixedSize(_width, _height);
		edit_l[i]->setFocusPolicy(QWidget::NoFocus);
		layout->addWidget(edit_l[i],3+i,1);

		edit_r[i] = new QLineEdit(this);
		CHECK_PTR(edit_r[i]);
		edit_r[i]->setReadOnly(true);
		edit_r[i]->setAlignment(AlignHCenter);
		edit_r[i]->setFixedSize(_width, _height);
		edit_r[i]->setFocusPolicy(QWidget::NoFocus);
		layout->addWidget(edit_r[i],3+i,2);

		edit_av[i] = new QLineEdit(this);
		CHECK_PTR(edit_av[i]);
		edit_av[i]->setReadOnly(true);
		edit_av[i]->setAlignment(AlignHCenter);
		edit_av[i]->setFixedSize(_width, _height);
		edit_av[i]->setFocusPolicy(QWidget::NoFocus);
		layout->addWidget(edit_av[i],3+i,3);
	}

	repeatbutton = new QJKPushButton("&Measure",this);
	CHECK_PTR(repeatbutton);
	repeatbutton->setFixedSize(repeatbutton->sizeHint());
	layout->addMultiCellWidget(repeatbutton,
				N_CHAN+3,N_CHAN+3,0,1,AlignHCenter);
	connect(repeatbutton, SIGNAL(clicked()),
				this, SLOT(correct()));

	stopbutton = new QJKPushButton("&Break",this);
	CHECK_PTR(stopbutton);
	stopbutton->setFixedSize(repeatbutton->sizeHint());
	layout->addMultiCellWidget(stopbutton,
				N_CHAN+3,N_CHAN+3,2,3,AlignHCenter);
	connect(stopbutton, SIGNAL(clicked()),
				&devmanager, SLOT(stopCounting()));

	acceptbutton = new QJKPushButton("&Accept",this);
	CHECK_PTR(acceptbutton);
	acceptbutton->setFixedSize(repeatbutton->sizeHint());
	layout->addMultiCellWidget(acceptbutton,
				N_CHAN+4,N_CHAN+4,0,1,AlignHCenter);
	connect(acceptbutton, SIGNAL(clicked()),
				this, SLOT(acceptCorrection()));

	cancelbutton = new QJKPushButton("&Cancel",this);
	CHECK_PTR(cancelbutton);
	cancelbutton->setFixedSize(repeatbutton->sizeHint());
	layout->addMultiCellWidget(cancelbutton,
				N_CHAN+4,N_CHAN+4,2,3,AlignHCenter);	
	connect(cancelbutton, SIGNAL(clicked()),
				this, SLOT(reject()));
	setButtonsEnabled(true);
}

void BackCorrectionDialog::update(void) {
	devmanager.getBackgroundCorrectionCoefficients(&coef);
	QString str;
	for(int i=0; i<N_CHAN; i++) {
		coef[i][0] = 0;
		edit_l[i]->clear();
		coef[i][1] = 0;
		edit_r[i]->clear();
		str.setNum(coef[i][2],'f',2);
		edit_av[i]->setText(str);
	}
	timer = startTimer(DIALOG_POLLING_INT); // 0.5 sec
}

void BackCorrectionDialog::finish(void) {
	killTimer(timer);
}

void BackCorrectionDialog::acceptCorrection(void) {
	devmanager.acceptBackgroundCorrection();
	accept();
}

void BackCorrectionDialog::timerEvent(QTimerEvent *e) {
	if(StateExport::correction()) {
		StateExport::cleanCorrectionReady();
		showResult();
	}
}

void BackCorrectionDialog::setButtonsEnabled(bool enable) {
	acceptbutton->setEnabled(enable);
	cancelbutton->setEnabled(enable);
	repeatbutton->setEnabled(enable);
	stopbutton->setEnabled(!enable);
	if(!enable) stopbutton->setFocus();
}

void BackCorrectionDialog::correct(void) {
	scenariothr::backgroundCorrection(&coef);
}

void BackCorrectionDialog::showResult(void) {
	jdebug("Result:");
	QString str;
	if(coef) {
		for(int i=0; i<N_CHAN; i++) {
			edit_l[i]->setText(str.setNum(coef[i][0],'f',2));
			edit_r[i]->setText(str.setNum(coef[i][1],'f',2));
			edit_av[i]->setText(str.setNum(coef[i][2],'f',2));
		}
	}
}

BackCorrectionDialog::~BackCorrectionDialog() {
	delete label;
	delete label_cf;
	delete label_l;
	delete label_r;
	delete label_av;
	delete label_a;
	delete label_b;
	delete label_c;
	delete label_d;
	for(int i=0; i<N_CHAN; i++) {
		delete edit_l[i];
		delete edit_r[i];
		delete edit_av[i];
	}
	delete layout;
	delete acceptbutton;
	delete repeatbutton;
	delete cancelbutton;
	delete stopbutton;
}

CorrectionDialog::CorrectionDialog(const QString name,
				const QString win1name,
				const QString win2name,
				int what)
			: QDialog(0,name,true) {
	_what = what;
	setCaption(name);
	int border = fontMetrics().height();
	layout = new QGridLayout(this,5,2,border);
	CHECK_PTR(layout);

	label = new QLabel(name,this);
	CHECK_PTR(label);
	layout->addMultiCellWidget(label,0,0,0,1,AlignHCenter);

	label_x = new QLabel(win1name+":",this);
	CHECK_PTR(label_x);
	layout->addWidget(label_x,1,0,AlignHCenter);

	label_y = new QLabel(win2name+":",this);
	CHECK_PTR(label_y);
	layout->addWidget(label_y,2,0,AlignHCenter);

	edit_x = new QLineEdit(this);
	CHECK_PTR(edit_x);
	edit_x->setReadOnly(true);
	edit_x->setAlignment(AlignHCenter);
	edit_x->setFocusPolicy(QWidget::NoFocus);
	layout->addWidget(edit_x,1,1);

	edit_y = new QLineEdit(this);
	CHECK_PTR(edit_y);
	edit_y->setReadOnly(true);
	edit_y->setAlignment(AlignHCenter);
	edit_y->setFocusPolicy(QWidget::NoFocus);
	layout->addWidget(edit_y,2,1);

	repeatbutton = new QJKPushButton("&Measure",this);
	CHECK_PTR(repeatbutton);
	layout->addWidget(repeatbutton,3,0);
	connect(repeatbutton, SIGNAL(clicked()),
				this, SLOT(correct()));

	stopbutton = new QJKPushButton("&Break",this);
	CHECK_PTR(stopbutton);
	layout->addWidget(stopbutton,3,1);
	connect(stopbutton, SIGNAL(clicked()),
				&devmanager, SLOT(stopCounting()));

	acceptbutton = new QJKPushButton("&Accept",this);
	CHECK_PTR(acceptbutton);
	layout->addWidget(acceptbutton,4,0);
	connect(acceptbutton, SIGNAL(clicked()),
				this, SLOT(acceptCorrection()));

	cancelbutton = new QJKPushButton("&Cancel",this);
	CHECK_PTR(cancelbutton);
	layout->addWidget(cancelbutton,4,1);
	connect(cancelbutton, SIGNAL(clicked()),
				this, SLOT(reject()));
	setButtonsEnabled(true);
	fitToFont();
}

void CorrectionDialog::fitToFont(void) {
	int w = fontMetrics().width("wwwwww");
	int h = getHeight(fontMetrics().height());
	edit_x->setFixedSize(w,h);
	edit_y->setFixedSize(w,h);
	repeatbutton->setFixedSize(repeatbutton->sizeHint());
	stopbutton->setFixedSize(repeatbutton->sizeHint());
	acceptbutton->setFixedSize(repeatbutton->sizeHint());
	cancelbutton->setFixedSize(repeatbutton->sizeHint());
}

void CorrectionDialog::update(void) {
	delta.setX(0);
	delta.setY(0);
	QString str;
	edit_x->setText(str.setNum(delta.x(),'f',2));
	edit_y->setText(str.setNum(delta.y(),'f',2));
	_polling = true;
	timer = startTimer(DIALOG_POLLING_INT); // 0.5 sec
}

void CorrectionDialog::finish(void) {
	killTimer(timer);
}

void CorrectionDialog::acceptCorrection(void) {
	if(_what == CENTER) 
		devmanager.acceptCenterCorrection();
	else
		devmanager.acceptTrackCalibration();
	accept();
}

void CorrectionDialog::timerEvent(QTimerEvent *e) {
	if(_polling) {
		_polling = false;
		if(StateExport::correction()) {
			StateExport::cleanCorrectionReady();
			showResult();
			if(StateExport::error()) {
				StateExport::cleanError();
				showStrongMessage(StateExport::message());
			}
		}
		_polling = true;
	}
}

void CorrectionDialog::setButtonsEnabled(bool enable) {
	acceptbutton->setEnabled(enable);
	cancelbutton->setEnabled(enable);
	repeatbutton->setEnabled(enable);
	stopbutton->setEnabled(!enable);
	if(!enable) stopbutton->setFocus();
}

void CorrectionDialog::correct(void) {
	jdebug("slot repeat!");
	if(_what == CENTER)
		scenariothr::centerCorrection(&delta);
	else
		scenariothr::trackCalibration(&delta);
}

void CorrectionDialog::showResult(void) {
	jdebug("Result:");
	QString str;
	edit_x->setText(str.setNum(delta.x(),'f',2));
	edit_y->setText(str.setNum(delta.y(),'f',2));
}

CorrectionDialog::~CorrectionDialog() {
	delete label;
	delete label_x;
	delete label_y;
	delete edit_x;
	delete edit_y;
	delete layout;
	delete acceptbutton;
	delete repeatbutton;
	delete cancelbutton;
	delete stopbutton;
}
